
/*======================================================================*
 *  Module .............btdiag
 *  File ...............main.c
 *  Author .............Traveler Hauptman
 *  Creation Date ......18 Mar 2003
 *                                                                      *
 *  ******************************************************************  *
 *                                                                      *
 *  NOTES:
 *
 *  REVISION HISTORY:
 *  18 Mar 2003 - TH
 *    File created
 *  25 Mar 2003 - BZ
 *    File documented
 *  18 May 2004 - BZ
 *    Reworked interface
 *                                                                      *
 *======================================================================*/

/** \file main.c
    Provides a dignostic interface to a set of motor controllers.

 */


/*==============================*
 * INCLUDES - System Files      *
 *==============================*/
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <syslog.h>
#include <stddef.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <semaphore.h>
#include <malloc.h>
#include <process.h>
#include <time.h>
#include <inttypes.h>

/* QNX specific header files */
#include <sys/sched.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <sys/netmgr.h>
#include <sys/syspage.h>

/*==============================*
 * INCLUDES - Project Files     *
 *==============================*/

#include "newlibbt.h"
#include "btdriver.h"

#include "globals.h"
#include "control_loop.h"
#include "btcan.h"
#include "btsystem.h"
#include "btwam.h"
#include "SimpleControl.h"
#include "playlist.h"
#include "doublebuffer.h"
#include "gimbals.h"
#include "newlibbt.h"
#include <curses.h>


enum { INTRO, SET_SAFETY_PRE, ZERO_PRE, GRAVCOMP_PRE, WALLDEFINE_PRE, WALLDEFINE_PT1, WALLDEFINE_PT2, WALLDEFINE_PT3, WALLDEFINE_POST, NORMAL };
static int walkthrough_mode = INTRO;
static int gravWalkStage = -1;

#define DEMO_MAX_OBJS 128
static OBJECT hapticObjects[DEMO_MAX_OBJS];  /* max 128 objects in space for the moment */
static int hapticObjectIndices[DEMO_MAX_OBJS]; /* for removal */
static PLANE hapticPlanes[DEMO_MAX_OBJS]; /* !!THIS DEMO IS STUCK WITH ONLY PLANE OBJECTS!! */ 

static double tempWallPt1[3];
static double tempWallPt2[3]; 
static double tempWallPt3[3];
static WAMLoc tempLoc;

static int walkthrough_done = 0;




#define disp_offset 15
#define disp_width 11


char divider[disp_offset + (disp_width * MAX_NODES)];
SC_move_list ml;
frame myFrame;

//wam_struct *WAM;

int done = 0;
int motor_position[7];
pthread_mutex_t disp_mutex;
int WACKYplay = 0;
int showWhere = 0;
int active = 0;

//control_thd_parms ctp;
//SimpleCtl *s;
//actuator_struct *act;
//int num_act;
//bus_struct *bus;
//int num_bus;
//uint64_t CPU_cycles_per_second;
//int showGimbals = 0;
//int showBusError = 0;
//int showLimits = 0;
//int processJacobian = 0;

DoubleBuffer_struct dbl;

//wam_vector wv;
//extern wam_struct WAM;
//=======================================================

int npucks, cpuck;
//int node;

//int use_safety_module;
int entryLine;
//int zeroing = 0, entry = 0;
int disp_puck = 0;
//int disp_help = 0;
int useTRC = 0;
long saved_bus_error = 0;

//int indexes[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
//int offsets[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int modes[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
long int temperatures[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
//double gimbals[3] = {0, 0, 0};
double commands[15];
double newpid[15];
//double A,B,C,D;
enum {HELP_SCREEN, MAIN_SCREEN, GCOMP_SCREEN, HOMING_SCREEN, TIMING_SCREEN, PLAYLIST_SCREEN};
int present_screen = MAIN_SCREEN;
int startup_stage = 0; //Set to 1 when main loop is started. Used to control sigint handler

int useGimbals = 0;

//==============================================================

/*==============================*
 * PRIVATE Function Prototypes  *
 *==============================*/
void sigint_handler();
void RenderScreen();
void RenderHelp();
void RenderScreenGcomp();
void ProcessInput(int c);
void Initialization();
void Shutdown();
void DisplayThread();
void StartDisplayThread();
void WACKYeval();
char *ifstr(int in, char *yes, char *no);
void ProcessJacobian();
void clearScreen(void);
void deleteClosePlane(WAMLoc *tipLoc);

/*==============================*
 * Functions                    *
 *==============================*/


/** Entry point for the application.
    Initializes the system and executes the main event loop.

*/

WAMdat wam;
extern wam_struct WAM;  // this is a hack that will be resolved - sc

void deleteClosePlane(WAMLoc *tipLoc)
{

  int i;
  double closestSoFar = MAXFLOAT;
  int closestIndex;
  double thisVal;
  
#define CLOSENESS_THRESHOLD 0.06 /* no idea - 6 cm?  */

  for(i=0;i<DEMO_MAX_OBJS;i++)
    {
      if(hapticObjectIndices[i] != -1)
	{
	  
	  if((thisVal = distanceFromPlane(tipLoc, &hapticPlanes[i])) < closestSoFar)
	    {
	      closestSoFar = thisVal;
	      closestIndex = i;
	    }
	  

	}

    }
  
  if(closestSoFar < CLOSENESS_THRESHOLD)
    {

      removeHapticObject(hapticObjectIndices[closestIndex]);
      hapticObjectIndices[closestIndex] = -1;
      

    }

}

int main(int argc, char **argv)
{

    char chr;
    char filename[50];
    int cnt, cnt2, num, i;
    int ret;
    int status[MAX_NODES];
    FILE *inFile;
    OBJECT blockObj1; /* trying to avoid malloc */
    BLOCK block1;
    RESISTANCELINE line1;
    WAMLoc blockC1, blockC2;
    

    
    Initialization();

    syslog(LOG_ERR, "starting initialize system", ret);
    ret = InitializeSystem("actuators.dat","buses.dat","motors.dat","pucks.dat");

    setSafetyLimits(2.0, 2.0, 2.0);  // ooh dangerous

    syslog(LOG_ERR, "done with initialize system: returned %d", ret);
    if (ret)
    {
        syslog(LOG_ERR, "Failed to initialize system: returned %d", ret);
        return (-1);
    }
    
    if(argc > 1)
    {
        if(!strcmp(argv[1],"-g")) // If gimbals are being used
        {
            initGimbals();
            useGimbals = 1;      
        }
    }
    

    //    ret = InitWAM(wam.wamDriverData, "wam.dat");


    /* This is a hack for now, so references to global WAM don't 
       have to be updated through all the driver code*/
    wam.wamDriverData = &WAM;

/*     blockC1.pos[0] = -0.56; */
/*     blockC1.pos[1] = -0.228; */
/*     blockC1.pos[2] = 0.1; */
    
/*     blockC2.pos[0] = -0.51; */
/*     blockC2.pos[1] = 0.23; */
/*     blockC2.pos[2] = 0.35; */
    
/*     /\*createBlock(&blockObj1, &block1, &blockC1, &blockC2, 1.0);*\/ */
    

/*     createResistanceLine(&blockObj1, &line1, &blockC1, &blockC2, 15); */

    
    InitHapticWAM(&wam);
 
   /* for demo purposes only here */
    for(i=0;i<DEMO_MAX_OBJS;i++)
      {
	hapticObjectIndices[i] = -1;
      }

    setSafetyLimits(2.0, 2.0, 2.0); // don't turn off right away




    DumpData2Syslog();


    npucks = wam.wamDriverData->num_actuators;
    

    //bus = GetBuses(&num_bus); //get pointer and number of buses.
    //s = GetWAMsc();

    MLconstruct(&ml, wam.wamDriverData->sc, 7, 50); //Initialize the playlist
    //cnt2 = 0;
    //ret = 1;

    if (InitDL(&dbl,1000,"datalog.dat"))//Initialize data logging, abort if there are memory problems
    {
        CloseSystem();
        closelog();
        endwin();
        exit(0);
    }

    start_control_threads(26, 2, WAMControlThread, (void *)&dbl);

    signal (SIGINT, sigint_handler); //register the interrupt handler
    startup_stage = 1;

    StartDisplayThread();

    while (!done)
    {
        //        if(processJacobian)
        //            ProcessJacobian();

              evalDL(&dbl); //check to see whether the data logger buffers are full and need writing
              WACKYeval();  //check to see whether we want to update to another random trajectory
              MLeval(&ml);  //Check to see whether the playlist needs attention


        if ((chr = getch()) != ERR) //Check buffer for keypress
            ProcessInput(chr);
        //        if(showGimbals)
        //            getGimbalsAngles(gimbals);


	//
	        //if(showWhere)
	            //whereAmI(&myFrame, WAM_4DOF_G);
            
	        //getProperty(wam.wamDriverData->act[0].bus, SAFETY_MODULE, SAVED_BUS_ERR, &saved_bus_error);
	        usleep(100000);

    }

    //    CloseDL(&dbl);
    Shutdown();
    CloseSystem();


}

#if 0
    void ProcessJacobian()
    {
        /* this will break with 7DOF, obviously */
        static double Jacobian[24];
        static int i,j;
        findJacobian(wam.wamDriverData->Jpos, Jacobian);
    
	        syslog(LOG_ERR, "jacobian:\n");
        for(i=0;i<6;i++)
        {
            syslog(LOG_ERR, "%f %f %f %f\n", Jacobian[i*4], Jacobian[i*4+1], Jacobian[i*4+2], Jacobian[i*4+3]);
        }
    
    }
#endif

/** Initialize curses, syslog, and the display mutex.

*/
void Initialization()
{
    int err;
    int i,cnt;

    /* Get the cycles-per-second of this machine */
    //CPU_cycles_per_second = SYSPAGE_ENTRY(qtime)->cycles_per_sec;

    /* Initialize the ncurses screen library */
    initscr();
    addstr("screen initalized\n");
    refresh();
    cbreak();
    noecho();
    timeout(0);
    clear();

    //* Initialize syslog */
    openlog("WAM", LOG_CONS | LOG_NDELAY, LOG_USER);
    addstr("syslog initalized\n");
    refresh();
    syslog(LOG_ERR, "syslog initalized by WAM application");

    /* Initialize the display mutex */
    err = pthread_mutex_init(&(disp_mutex), NULL);
    if (err)
    {
        syslog(LOG_ERR, "Could not initialize mutex for displays.");
    }

    /* Set the present active puck to 0 */
    cpuck = 0;

}

/** Traps the Ctrl-C signal.
    Quits the program gracefully when Ctrl-C is hit.

*/
void sigint_handler()
{
    if (!startup_stage)
    {
        endwin();
        exit(1);
    }
    else
        done = 1;
}

/** Shut the application down gracefully.
    Stops the control threads, ncurses, syslog, and frees the movelist.
*/
void Shutdown()
{
    // Put the gimbals into RESET
    //if(useGimbals) setProperty(0, 5, STATUS, FALSE, STATUS_RESET);
    
    stop_control_threads();
    MLdestroy(&ml);
    closelog();
    endwin();
}

/** Starts the display thread.
    Creates a new thread, sets its priority, and runs it as a display thread.
*/
void StartDisplayThread()
{
    pthread_t display_thd_id;
    pthread_attr_t display_attr;
    struct sched_param display_param;

    pthread_attr_init(&display_attr);
    pthread_attr_setinheritsched(&display_attr, PTHREAD_EXPLICIT_SCHED);
    display_param.sched_priority = 10;
    pthread_attr_setschedparam(&display_attr, &display_param);

    pthread_create(&display_thd_id, &display_attr, (void*)DisplayThread, 0);
    sched_setscheduler(display_thd_id, SCHED_FIFO, &display_param);
    if (display_thd_id == -1)
    {
        syslog(LOG_ERR, "start_display_thread:Couldn't start display thread!");
        exit( -1);
    }
}

/** Spins in a loop, updating the screen.
    Runs as its own thread, updates the screen.
*/
void DisplayThread()
{
    int cnt,err;

    /* Create the divider line */
    divider[0] = '\0';
    for(cnt = 0; cnt < disp_offset + wam.wamDriverData->num_actuators * disp_width; cnt++)
      {
        strcat(divider, "-");
      }


    clear();
    refresh();

    while (!done)
    {


      if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
        {
	  syslog(LOG_ERR, "display mutex failed: %d", err);
        }
      

	switch (present_screen)
        {
            case HELP_SCREEN:
                RenderHelp();
                break;
            case MAIN_SCREEN:
	      RenderScreen();
                break;
            case GCOMP_SCREEN:
            //RenderScreenGcomp();
                break;

        }
        pthread_mutex_unlock( &(disp_mutex) );
        usleep(100000);
	}

}

/** Locks the display mutex.
    Allows the user to enter on-screen data without fear of display corruption.
*/
void start_entry()
{
    int err;
    if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
    {
        syslog(LOG_ERR, "display mutex failed: %d", err);
    }
    move(entryLine, 1);
    echo();
    timeout( -1);
}

/** Unlocks the display mutex.
    Allows the computer to resume automatically updating the screen.
*/
void finish_entry()
{
    noecho();
    timeout(0);
    move(entryLine, 1);
    addstr("                                                                              ");
    refresh();
    pthread_mutex_unlock( &(disp_mutex) );
}

/** Returns a string based on the value of an integer. A code readibility construct.

*/
char *ifstr(int in, char *yes, char *no)
{
    if (in)
        return yes;
    else
        return no;
}

/** Show the help screen.
    Displays the on-screen help.
*/
void RenderHelp() //{{{
{
    int val;
    int cnt, idx;
    uint64_t result;
    int line;


    line = 0;
    mvprintw(line , 0, "'a' Power wam");
    ++line;
    mvprintw(line , 0, "'d' Idle wam");
    ++line;
    mvprintw(line , 0, " ");
    ++line;
    mvprintw(line , 0, "'[' Select previous joint");
    ++line;
    mvprintw(line , 0, "']' Select next joint");
    ++line;

    refresh();
}

/** Draw the main information screen.
    Dynamically draw the puck information on the screen.
*/
void RenderScreen() //{{{
{
    int val;
    int cnt, idx;
    uint64_t result;
    int line;
    wam_vector Mpos, Mtrq, Jpos, Jtrq, GCtrq,GCpos, t2, t3, t4;
    wam_vector pos,trq;
    double pi=3.14159,a,b,c,d;





    line = 0;
    mvprintw(line , 0, "Actuators are %s           Last Error: ", active ? "ACTIVATED  " : "DEACTIVATED"); ++line;
    //    mvprintw(line, 0, "%s", divider); ++line;         |


    if (disp_puck)
    {
        mvprintw(line , 0, "       PUCK: "); ++line;
        mvprintw(line , 0, "       Mode: "); ++line;
        mvprintw(line , 0, "        Pos: "); ++line;
        mvprintw(line , 0, "        Trq: "); ++line;
    }
    else
    {
        mvprintw(line , 0, "      JOINT: "); ++line;
        mvprintw(line , 0, "       Mode: "); ++line;
        mvprintw(line , 0, "      Angle: "); ++line;
        mvprintw(line , 0, "     Torque: "); ++line;
    }
    mvprintw(line , 0, "    PID Cmd: "); ++line;
    mvprintw(line , 0, " Trj Target: "); ++line;
    mvprintw(line , 0, "[G] Temp C : "); ++line;


    #if 0
        mvprintw(line , 0, "[p] Kp     : "); ++line;
        mvprintw(line , 0, "    Kd     : "); ++line;
        mvprintw(line , 0, "    Ki     : "); ++line;
        mvprintw(line , 0, "[v] Vel    : "); ++line;
        mvprintw(line , 0, "    Acc    : "); ++line;
    #endif

    //    mvprintw(line, 0, "%s", divider); ++line;
    mvprintw(line, 0, "[c] Command: "); ++line;
    mvprintw(line, 0, "%s", divider); ++line;

    mvprintw(line, 0, "  Play List: "); ++line;
    mvprintw(line, 0, "     Status: ");++line;
    mvprintw(line, 5, "[.]Stop  [,]One   [;]Play  [/]Loop  [>]Next"); ++line;
    mvprintw(line, 5, "[<]Prev  [+]Add   [-]Del   [v]Save  [l]Load"); ++line;

    mvprintw(line, 0, "%s", divider); ++line;
    if(showWhere)
    {
        mvprintw(line, 0, "   WhereAmI:", myFrame.x, myFrame.y, myFrame.z);    ++line;
    }
    entryLine = line;
    ++line;
    mvprintw(line, 0, "%s", divider); ++line;

    
    switch(walkthrough_mode)
      {

      case INTRO:
	mvprintw(line, 0, "Welcome to the Haptic WAM Interface Software Demo  "); 
	++line;
	mvprintw(line, 0, "Press any key to start a walkthrough of a few of  ");
	++line;
	mvprintw(line, 0, "the features of the WAM.   ");
	++line;
	mvprintw(line, 0, "Press 'x' at any time to exit.");
	++line;
	++line;
	break;
      case SET_SAFETY_PRE:
	mvprintw(line, 0, "First, you can set the safety limits on the WAM velocity.  ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Press 'L' to enter values for maximum joint velocity (rad/sec),  ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "elbow velocity (m/s), and tip velocity (m/s).  "  );
	clrtoeol();
	++line;
	mvprintw(line, 0, "Try values of 2, 2, and 2 to start.  ");
	clrtoeol();
	++line;
	clrtoeol();
	++line;
	break;


      case ZERO_PRE:
	mvprintw(line, 0, "Next, have the WAM automatically calibrate its joint angles.  ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "First, activate the WAM (on the pendant) and press 'a' to enable the WAM actuators.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Now, press 'Z' to initiate the zeroing process.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Or press 's' to skip this if it is already zeroed.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      case GRAVCOMP_PRE:
	mvprintw(line, 0, "Now that the WAM is zeroed, gravity compensation can be activated. ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Press 'd' to deactivate the WAM controllers.          ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Now, press 'T' to put the WAM in 'torque mode', which is required for gravcomp.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Next, press 'g' to activate gravity compensation.  ");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Now, press 'a' to activate the WAM activators and see gravcomp in action. ");
	clrtoeol();
	++line;
	break;

      case WALLDEFINE_PRE:
	mvprintw(line, 0, "After experimenting with the gravity compensation, you can create walls");
	++line;
	mvprintw(line, 0, "in space by defining 3 points that lie in a plane.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Try it now by pressing 'W'.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      case WALLDEFINE_PT1:
	mvprintw(line, 0, "Please place the WAM tip at the first point to define a wall.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Press 'b' to save the point.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      case WALLDEFINE_PT2:
	mvprintw(line, 0, "Please place the WAM tip at the second point to define a wall.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Press 'b' to save the point.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      case WALLDEFINE_PT3:
	mvprintw(line, 0, "Please place the WAM tip at the third point to define a wall.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Press 'b' to save the point.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      case WALLDEFINE_POST:
	
	mvprintw(line, 0, "A wall is now defined between the three points that you entered.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "You can delete a wall at any time by putting the WAM tip near the wall and pressing 'B'");
	clrtoeol();
	++line;
	mvprintw(line, 0, "You can now use all of these commands and more to control the WAM.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "Have fun!  Press any key to continue.");
	clrtoeol();
	++line;
	mvprintw(line, 0, "");
	clrtoeol();
	++line;
	break;

      default:
	mvprintw(line , 0, "[a] Activate       [d] Deactivate   [L] Limit WAM velocity  "); ++line; 
	mvprintw(line , 0, "[i,I] Idle Mode    [x] Exit         [g] (%c) Gravity toggle ", wam.wamDriverData->Gcomp ? '*' : ' '); ++line;
	mvprintw(line , 0, "[t,T] Torque Mode  [e] Error reset  [r] (%c) Ripple toggle  ", useTRC ? '*' : ' '); ++line;
	mvprintw(line , 0, "[p,P] PID Mode     [h] Home WAM     [z,Z] ZeroWAM (Man/Auto)"); ++line;
	mvprintw(line , 0, "[s,S] Start Trj    [ ] Joint/Puck                           "); ++line;
	
      }

    getWAMjoints(&Mpos, &Mtrq, &Jpos, &Jtrq);
    getWAMmotor_position(motor_position);
    

    // Convert the gimbals position data
    if(useGimbals)
        getGimbalsAngles(&Jpos.q[motor_position[5]]);
        
    //    calcGcomp2(&t2, &t3, &t4);

    line = 0;

    if (1)
    {
        if(saved_bus_error == 0)
            mvprintw(line, 48,"(No Error)                ");
        if(saved_bus_error & 0x00FF)
            mvprintw(line, 48,"Heartbeat missing (0x%02x) ", saved_bus_error);
        if(saved_bus_error & 0x0100)
            mvprintw(line, 48,"Over velocity on joint 1  ");
        if(saved_bus_error & 0x0200)
            mvprintw(line, 48,"Over velocity on joint 2  ");
        if(saved_bus_error & 0x0400)
            mvprintw(line, 48,"Over velocity on joint 3  ");
        if(saved_bus_error & 0x0800)
            mvprintw(line, 48,"Over velocity on joint 4  ");
        if(saved_bus_error & 0x1000)
            mvprintw(line, 48,"Over velocity at endtip   ");
        if(saved_bus_error & 0x2000)
            mvprintw(line, 48,"Over velocity at elbow    ");
        if(saved_bus_error & 0x4000)
            mvprintw(line, 48,"Enable torque exceeded    ");

        ++line;
    }


    fer (cnt, wam.wamDriverData->num_actuators)
    {
        line = 1;
        idx = wam.wamDriverData->act[cnt].puck.ID;
        
        if (disp_puck)
        {
            mvprintw(line , disp_offset + disp_width*cnt, "   %2d   ", idx); ++line;
            //mvprintw(line , disp_offset + disp_width*cnt, "%8d ", modes[idx]); ++line;
            if(modes[idx] == MODE_IDLE) 
                mvprintw(line , disp_offset + disp_width*cnt, "IDLE  ");
            else if (modes[idx] == MODE_TORQUE) 
                mvprintw(line , disp_offset + disp_width*cnt, "TORQUE");
            else if (modes[idx] == MODE_PID) 
                mvprintw(line , disp_offset + disp_width*cnt, "PID   ");
            else 
                mvprintw(line , disp_offset + disp_width*cnt, "UNK:%d ", modes[idx]);
            ++line;
            mvprintw(line , disp_offset + disp_width*cnt, "%8d ", wam.wamDriverData->act[cnt].puck.position); ++line;
            mvprintw(line , disp_offset + disp_width*cnt, "%8d ", wam.wamDriverData->act[cnt].puck.torque_cmd); ++line;
        }
        else
        {
            if (motor_position[cnt] == motor_position[cpuck])
                mvprintw(line , disp_offset + disp_width*cnt, " -[%2d]- ", motor_position[cnt]+1);
            else
                mvprintw(line , disp_offset + disp_width*cnt, "   %2d  ", motor_position[cnt]+1);
            ++line;

            if (wam.wamDriverData->sc[motor_position[cnt]].mode == MODE_IDLE)
                mvprintw(line , disp_offset + disp_width*cnt, "IDLE  ");
            else if (wam.wamDriverData->sc[motor_position[cnt]].mode == MODE_TORQUE)
                mvprintw(line , disp_offset + disp_width*cnt, "TORQUE");
            else if (wam.wamDriverData->sc[motor_position[cnt]].mode == MODE_PID)
                mvprintw(line , disp_offset + disp_width*cnt, "PID   ");
            else
                mvprintw(line , disp_offset + disp_width*cnt, "UNK:%d ", wam.wamDriverData->sc[motor_position[cnt]].mode);
            ++line;

            mvprintw(line, disp_offset + disp_width*cnt, "%+8.4f ", Jpos.q[motor_position[cnt]]); ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%+8.4f ", Jtrq.q[motor_position[cnt]]); ++line;
        }
        mvprintw(line, disp_offset + disp_width*cnt, "%+8.4f ", wam.wamDriverData->sc[motor_position[cnt]].pid.yref);    ++line;
        if (wam.wamDriverData->sc[motor_position[cnt]].trj.state == 0)
            mvprintw(line, disp_offset + disp_width*cnt, "STOPPED    ");
        else
            mvprintw(line, disp_offset + disp_width*cnt, "%+8.4f ", wam.wamDriverData->sc[motor_position[cnt]].trj.end);
        ++line;

        mvprintw(line, disp_offset + disp_width*cnt, "%8.1f ", temperatures[wam.wamDriverData->act[cnt].puck.ID] / 10.0);    ++line;

        //        ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%+8.4f ", commands[cnt]); ++line;
        ++line;

        if (ml.Nmoves)
            mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", ml.next[cnt].pos);
        else
            mvprintw(line, disp_offset + disp_width*cnt, " No Pts ");
        ++line;
        mvprintw(line, disp_offset, "%s, point %d of %d          ", ifstr(MLrunning(&ml), "RUNNING", "STOPPED"),ml.Cmove,ml.Nmoves); ++line;


        #if 0
            mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", wam.wamDriverData->sc[motor_position[cnt]].pid.Kp );    ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", wam.wamDriverData->sc[motor_position[cnt]].pid.Kd );    ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", wam.wamDriverData->sc[motor_position[cnt]].pid.Ki);    ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", wam.wamDriverData->sc[motor_position[cnt]].trj.vel);    ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", wam.wamDriverData->sc[motor_position[cnt]].trj.acc);    ++line;
    
    
            //mvprintw(line, disp_offset + disp_width*cnt, "%8ld ", vel[act[cnt].puck.ID]);    ++line;
            ++line;
            if (zeroing)
            {
                mvprintw(line, disp_offset + disp_width*cnt, "ZEROING");      ++line;
            }
            else if (wam.wamDriverData->sc[motor_position[cnt]].trj.state == 0)
            {
                mvprintw(line, disp_offset + disp_width*cnt, "STOPPED");      ++line;
            }
            else if (wam.wamDriverData->sc[motor_position[cnt]].trj.state == TRAJ_RUN)
            {
                mvprintw(line, disp_offset + disp_width*cnt, "RUNNING");      ++line;
            }
            else if (wam.wamDriverData->sc[motor_position[cnt]].trj.state == TRAJ_FILE)
            {
                mvprintw(line, disp_offset + disp_width*cnt, "TRAJ GO");      ++line;
            }
            else
            {
                mvprintw(line, disp_offset + disp_width*cnt, "UNK:%d  ", wam.wamDriverData->sc[motor_position[cnt]].trj.state);      ++line;
            }
    
            //++line; // Divider
            //if(showGimbals)
            //{
            //    mvprintw(line, disp_offset, "J5 = %8.4f, J6 = %8.4f, J7 = %8.4f\t\t\t", gimbals[0], gimbals[1], gimbals[2]);    ++line;
            //}
    
            ++line;
            ++line;
            if (ml.Nmoves)
                mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", ml.next[cnt].pos);
            else
                mvprintw(line, disp_offset + disp_width*cnt, " No Pts ");
            ++line;
        #endif
        // mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", act[motor_position[cnt]].angle);
        //  ++line;
        //  mvprintw(line, disp_offset + disp_width*cnt, "%8.2e ", Jpos.q[motor_position[cnt]]);
        //   ++line;
    }
    ++line;
    ++line;
    ++line;

    if(showWhere)
    {
        mvprintw(line, disp_offset, " X = %8.4f,  Y = %8.4f,  Z = %8.4f\t\t\t", myFrame.x, myFrame.y, myFrame.z);    ++line;
    }
    mvprintw(line, 0, "");

    refresh();
} //}}}

#if 0
    void RenderScreenGcomp() //{{{
    {
        int val;
        int cnt, idx;
        uint64_t result;
        int line;
        wam_vector Mpos, Mtrq, Jpos, Jtrq, GCtrq,GCpos, t2, t3, t4;
        wam_vector pos,trq;
        double pi=3.14159,a,b,c,d;
    
        line = 0;
        mvprintw(line, 0, "Communication Time %10.5f milliseconds  ", TurnaroundTime(0)); ++line;
        getGcompValues(&A,&B,&C,&D);
        mvprintw(line, 0, "Gcomp Calc-Coefficients: A: %8.4f B: %8.4f C: %8.4f D: %8.4f ", A,B,C,D); ++line;
        calcGcomp(&Jpos,&GCtrq,A,B,C,D);
        mvprintw(line, 0, "Gcomp Calc-Results     : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", GCtrq.q[0],GCtrq.q[1],GCtrq.q[2],GCtrq.q[3]); ++line;
        mvprintw(line, 0, "Gcomp Calc-Angles      : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", Jpos.q[0],Jpos.q[1],Jpos.q[2],Jpos.q[3]); ++line;
        set_wam_vector(&GCpos,0,-pi/2,-pi/2,0,0,0,0);
        a = -0.638225;
        b = -6.125;
        c = 0.700088;
        d = 9.8;
        calcGcomp(&GCpos,&GCtrq,a,b,c,d);
        mvprintw(line, 0, "Reference Gcomp Calc-Coefficients: A: %8.4f B: %8.4f C: %8.4f D: %8.4f ", a,b,c,d); ++line;
        mvprintw(line, 0, "Reference Gcomp Calc-Angles      : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", Jpos.q[0],Jpos.q[1],Jpos.q[2],Jpos.q[3]); ++line;
        mvprintw(line, 0, "Reference Gcomp Calc-Results     : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", GCtrq.q[0],GCtrq.q[1],GCtrq.q[2],GCtrq.q[3]); ++line;
    
        mvprintw(line, disp_offset, "               "); ++line;
        calcGcomp2(&t2, &t3, &t4);
        mvprintw(line, 0, "t2: %10.5f, %10.5f, %10.5f, %10.5f ", t2.q[0], t2.q[1], t2.q[2], t2.q[3]); ++line;
        mvprintw(line, 0, "t3: %10.5f, %10.5f, %10.5f, %10.5f ", t3.q[0], t3.q[1], t3.q[2], t3.q[3]); ++line;
        mvprintw(line, 0, "t4: %10.5f, %10.5f, %10.5f, %10.5f ", t4.q[0], t4.q[1], t4.q[2], t4.q[3]); ++line;
        ++line;
        mvprintw(line, 0, "Position offsets: %d, %d, %d, %d ", offsets[0], offsets[1], offsets[2], offsets[3]);
        entryLine = line;
        refresh();
    }
#endif

void clearScreen(void)
{
    int err;

    if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
    {
        syslog(LOG_ERR, "display mutex failed: %d", err);
    }
    clear();
    pthread_mutex_unlock( &(disp_mutex) );
}

/** Process user input.
    Handles the user's keypress, and performs the function desired by the user.
*/
void ProcessInput(int c) //{{{ Takes last keypress and performs appropriate action.
{
    //int val;
    int tmp;
    int cnt;
    //int ret;
    //int idx, err;
    //long ltmp;
    char chr;
    int i;
    double jVel, tVel, eVel;
    double ftmp;
    char fn[50];
    wam_vector wv;
    static int firstdone = 0;

 if(c=='x' || c=='X') /* eXit */
   {
     done = 1;
     return; 
   }

    if(walkthrough_mode == INTRO)
      {
	walkthrough_mode = SET_SAFETY_PRE;
      }
    else if(walkthrough_mode == SET_SAFETY_PRE)
      {

	if(c == 'L')
	  {
            start_entry();
            addstr("Enter JointVel (rad/s), TipVel (m/s), ElbowVel (m/s): ");
            refresh();
            scanw("%lf,%lf,%lf", &jVel, &tVel, &eVel);
            finish_entry();
            setSafetyLimits(jVel, tVel, eVel);
	    walkthrough_mode = ZERO_PRE;

	  }
      }
    else if(walkthrough_mode == ZERO_PRE)
      {
	if(c == 'a')
	  {
	    firstdone = 1;
	    fer(cnt, npucks) SetProp(cnt,MODE, MODE_TORQUE);
	    active = 1;
	    fer(cnt, npucks) // Get Mode
	      {
		modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
		tmp = GetProp(cnt, MODE);
		modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
	      }
	  }
	else if(firstdone && (c == 'Z'))
	  {
	   
	    useTRC = 1;
	    SetTRC(useTRC);
	    ZeroWAM();
	    walkthrough_mode = GRAVCOMP_PRE;
	  }
	else if(firstdone && (c == 's'))
	  {
	    useTRC=1;
	    SetTRC(useTRC);
	    walkthrough_mode = GRAVCOMP_PRE;
	  }
      }
    else if(walkthrough_mode == GRAVCOMP_PRE)
      {
	if(gravWalkStage == -1 && c == 'd')
	  {
            fer(cnt, npucks) SetProp(cnt,MODE, MODE_IDLE);
            active = 0;
            fer(cnt, npucks) // Get Mode
            {
	      modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
	      tmp = GetProp(cnt, MODE);
	      modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
            }
	    gravWalkStage = 0;
	  }
	
	    if(gravWalkStage == 0 && c == 'T')
	      {
		fer(cnt, npucks) SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_TORQUE);
		gravWalkStage = 1;
	      }
	    else if(gravWalkStage == 1 && c == 'g')
	      {
		toggleGcomp(wam.wamDriverData);
		refresh();
		gravWalkStage = 2;
	      }
	    else if(gravWalkStage == 2 && c == 'a')
	      {
		fer(cnt, npucks) SetProp(cnt,MODE, MODE_TORQUE);
		active = 1;
		fer(cnt, npucks) // Get Mode
		  {
		    modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
		    tmp = GetProp(cnt, MODE);
		    modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
		  }

		walkthrough_mode = WALLDEFINE_PRE;
	      }
      }
    else if(walkthrough_mode == WALLDEFINE_PRE)
      {
	if(c == 'W')
	  {
	    walkthrough_mode = WALLDEFINE_PT1;
	  }

      }
    else if(walkthrough_mode == WALLDEFINE_PT1)
      {
	if(c == 'b')
	  {
	    
	    WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &tempLoc);
	    tempWallPt1[0] = tempLoc.pos[0];
	    tempWallPt1[1] = tempLoc.pos[1];
	    tempWallPt1[2] = tempLoc.pos[2];
	    walkthrough_mode = WALLDEFINE_PT2;
	  }

      }
    else if(walkthrough_mode == WALLDEFINE_PT2)
      {
	if(c == 'b')
	  {
	    
	    WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &tempLoc);
	    tempWallPt2[0] = tempLoc.pos[0];
	    tempWallPt2[1] = tempLoc.pos[1];
	    tempWallPt2[2] = tempLoc.pos[2];
	    walkthrough_mode = WALLDEFINE_PT3;
	  }

      }
    else if(walkthrough_mode == WALLDEFINE_PT3)
      {
	if(c == 'b')
	  {
	    
	    WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &tempLoc);
	    tempWallPt3[0] = tempLoc.pos[0];
	    tempWallPt3[1] = tempLoc.pos[1];
	    tempWallPt3[2] = tempLoc.pos[2];
	    
	    i=0;
	    while(hapticObjectIndices[i] != -1)
	      {
		i++;
	      }
	    //check for > than num objs

	    createPlane(&hapticObjects[i], &hapticPlanes[i], tempWallPt1, tempWallPt2, tempWallPt3);
	    hapticObjectIndices[i] = addHapticObject(&hapticObjects[i]);

	    if(!walkthrough_done)
	      walkthrough_mode = WALLDEFINE_POST;
	    else
	      walkthrough_mode = NORMAL;
	  }

      }
    else if(walkthrough_mode == WALLDEFINE_POST)
      {

	walkthrough_mode = NORMAL;
	walkthrough_done = 1;

      }
    else // NORMAL - non-walkthrough
      {
	switch (c)
	  {
	  case 'B':
	    WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &tempLoc);
	    deleteClosePlane(&tempLoc);
	    break;
	  case 'W': /* this is kind of a hack but the right idea. */
	    walkthrough_mode = WALLDEFINE_PT1;
	    break;
	  case 'a': // Activate WAM, sets actuators to Torque mode (applies commanded torques)
            fer(cnt, npucks) SetProp(cnt,MODE, MODE_TORQUE);
            active = 1;
            fer(cnt, npucks) // Get Mode
	      {
                modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, MODE);
                modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
            }
            break;
        case 'd': // Deactivate WAM, sets actuators to Idle mode (resistive braking, ignores torque commands)
            fer(cnt, npucks) SetProp(cnt,MODE, MODE_IDLE);
            active = 0;
            fer(cnt, npucks) // Get Mode
            {
                modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, MODE);
                modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
            }
            break;
        case 'i': // Set present joint controller to Idle mode (sends zero torque)
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_IDLE);
            break;
        case 'I': // Set ALL joint controllers to Idle mode (sends zero torque)
            fer(cnt, npucks) SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_IDLE);
            break;
        case 't': // Set present joint controller to Torque mode
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_TORQUE);
            break;
        case 'T': // Set ALL joint controllers to Torque mode
            fer(cnt, npucks) SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_TORQUE);
            break;
        case 'p': // Set present joint controller to PID mode
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_PID);
            break;
        case 'P': // Set ALL joint controllers to PID mode
            fer(cnt, npucks) SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_PID);
            break;
        case 's': // Start Trajectory control on present puck
            SCstarttrj(&(wam.wamDriverData->sc[motor_position[cpuck]]), commands[cpuck]);
            break;
        case 'S': // Start Trajectory control on all pucks
            fer(cnt, npucks) SCstarttrj(&(wam.wamDriverData->sc[motor_position[cnt]]), commands[cnt]);
            break;
	    //        case 'x': case 'X': /* eXit */
	    //            done = 1;
	    //            break; 
        case 'e': // Error reset for safety system
            //setProperty(wam.wamDriverData->act[0].bus, SAFETY_MODULE, SAVED_BUS_ERR, TRUE, 0);
            break;
        case 'h': /* Home / ParkWAM */
            ParkWAM();
            break;  
        case ' ': /* Toggle puck/joint display */
            disp_puck = !disp_puck;
            clear();
            break;
        case 'g': /* Toggle gravity compensation */
	    toggleGcomp(wam.wamDriverData);
            refresh();
            break;
        case 'r': /* Toggle Torque Ripple Compensation */
            useTRC = !useTRC;
            SetTRC(useTRC);
            refresh();
            break;  
        case 'Z': /* Zero WAM */
            ZeroWAM();
            useTRC = 1;
            SetTRC(useTRC);
            break; 
        case 'z': /* Send zero-position to WAM */
            set_wam_vector(&wv, 0, 0, 0, 0, 0, 0, 0);
            SetWAMpos(&wv);
            break;
        case 'L': // Limit WAM velocity
            start_entry();
            addstr("Enter JointVel (rad/s), TipVel (m/s), ElbowVel (m/s): ");
            refresh();
            scanw("%lf,%lf,%lf", &jVel, &tVel, &eVel);
            finish_entry();
            setSafetyLimits(jVel, tVel, eVel);
            break; 
        case 'w':   // Get whereAmI()
            showWhere = !showWhere;
            clearScreen();
            break;                   
        case '~':
            WACKYplay = !WACKYplay;
            //logdata = !logdata;
            //if (logdata) DLon(&dbl);
            //else DLoff(&dbl);
            break;
        case '[': /* Select previous puck */
            cpuck--;
            if (cpuck < 0)
                cpuck = npucks - 1;
            break;
        case ']': /* Select next puck */
            cpuck++;
            if (cpuck >= npucks)
                cpuck = 0;
            break;
        case 'c': /* Enter trajectory end position for present puck */
            start_entry();
            addstr("Enter next command: ");
            refresh();
            scanw("%lf", &(commands[cpuck]));
            finish_entry();
            break;    
        case 'C': /* Enter trajectory end position for all pucks */
            start_entry();
            addstr("Enter next command for all pucks: ");
            refresh();
            scanw("%lf", &ftmp);
            fer (cnt, npucks)
            {
                commands[cnt] = ftmp;
            }
            finish_entry();
            break;
        case '_': /* Refresh display */
            clearScreen();
            break;
        case 'G':   /* Get temperatures */
            fer(cnt, npucks)
            {
                temperatures[wam.wamDriverData->act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, TEMP);
                temperatures[wam.wamDriverData->act[cnt].puck.ID] = tmp;
            }
            break;
        case ';':   /* Play through playlist once */
            MLplay(&ml);
            break;
        case ',': /* Execute present line in playlist */
            MLone(&ml);
            break;
        case '>': /* Show next line in playlist */
            MLnext(&ml);
            break;
        case '<': /* Show previous line in playlist */
            MLprev(&ml);
            break;
        case '/':   /* Loop through playlist */
            MLrepeatplay(&ml);
            break;
        case '.':   /* Stop playlist execution */
            MLstop(&ml);
            break;
        case 'l':   /* Load playlist from file */
            start_entry();
            addstr("Enter filename for playlist read: ");
            refresh();
            scanw("%s", fn);
            MLload(&ml, fn);
            finish_entry();
            break;
        case 'v':   /* Save current playlist to file */
            start_entry();
            addstr("Enter filename for playlist save: ");
            refresh();
            scanw("%s", fn);
            MLsave(&ml, fn);
            finish_entry();
            break;
        case '+':   /* Add to playlist */
            MLaddSC(&ml);
            break;
        case '-':   /* Delete from playlist */
            MLdel(&ml);
            break;

        case 27: //Handle and discard extended keyboard characters (like arrows)
            if ((chr = getch()) != ERR)
            {
                if (chr == 91)
                {
                    if ((chr = getch()) != ERR)
                    {
                        if (chr == 67) //Right arrow
                        {
                            cpuck++;
                            if (cpuck >= npucks)
                                cpuck = 0;
                        }
                        else if (chr == 68) //Left arrow
                        {
                            cpuck--;
                            if (cpuck < 0)
                                cpuck = npucks - 1;
                        }
                        else
                            while(getch()!=ERR){}
                    }
                }
                else
                    while(getch()!=ERR){}
            }
            break; 
            
        default:
	  while(getch()!=ERR){}
	  //syslog(LOG_ERR,"Caught unknown keyhit %d",c);
	  break;
	  }  
	
      }  // NORMAL walkthrough mode
}
         
#if 0        
        case 'i': /* Idle controllers */
            IdleWAM();
            break;
        case 'd': /* Deactivate actuators */
            IdleWAM();
            break;
        
        
        /*case 'A': // Get Index
                    fer(cnt, npucks)
                    {
                        indexes[act[cnt].puck.ID] = -1;
                        indexes[act[cnt].puck.ID] = GetProp(cnt, DRIVE_MODE);
                    }
                    break;*/
        case 'M': /* get Mode */
            fer(cnt, npucks)
            {
                modes[wam.wamDriverData->act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, MODE);
                //syslog(LOG_ERR,"%d Val1 %d",cnt,tmp);
                modes[wam.wamDriverData->act[cnt].puck.ID] = tmp;
                //syslog(LOG_ERR,"Val2 %d",modes[act[cnt].puck.ID]);
            }
            break;
        case '|': /*Get gravity compensation values */
            getGcompvalues();
            break;
        /*case '`':
                    GCompCollectData("gcompdata.dat",2,2,1);
                    break;*/



// changed to create wall b/c 'T' makes more sense for torque mode        
/*         case 'W': /\* Set mode to MODE_TORQUE for all pucks*\/ */
/*             fer(cnt, npucks) */
/*             SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_TORQUE); */
/*             break; */
        case 'y': /* Set mode to MODE_TORQUE for present puck */
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_PID);
            break;
        case 'w': /* Set mode to MODE_TORQUE for present puck */
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_TORQUE);
            break;
        

        case 'q': /* Set mode to MODE_IDLE for present puck */
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cpuck]]), MODE_IDLE);
            break;
        case 'Q': /* Set mode to MODE_IDLE for all pucks */
            fer(cnt, npucks)
            SCsetmode(&(wam.wamDriverData->sc[motor_position[cnt]]), MODE_IDLE);
            break;
        case 'p': /* Enter PID constants for present puck \bug Check to see if we are in control mode 0 and able to change gains*/
            start_entry();
            addstr("Enter pid constants \"Kp Kd Ki\": ");
            refresh();
            ret = scanw("%lf %lf %lf %lf", &(newpid[0]), &(newpid[1]), &(newpid[2]), &(newpid[3]));
            //if (ret == 3)
            SCsetpid(&(wam.wamDriverData->sc[motor_position[cpuck]]), newpid[0], newpid[1], newpid[2], newpid[3]);
            finish_entry();
            break;
        case 'P': /* Enter PID constants for all pucks \bug Check to see if we are in control mode 0 and able to change gains*/
            start_entry();
            addstr("Enter pid constants \"Kp Kd Ki\": ");
            refresh();
            ret = scanw("%lf %lf %lf %lf", &(newpid[0]), &(newpid[1]), &(newpid[2]), &(newpid[3]));
            //if (ret == 3)
            fer(cnt, npucks)
            SCsetpid(&(wam.wamDriverData->sc[motor_position[cnt]]), newpid[0], newpid[1], newpid[2], newpid[3]);
            finish_entry();
            break;
        case 'k': /* Enter Kill torque for present puck */
            start_entry();
            addstr("Enter Kill torque: ");
            refresh();
            ret = scanw("%lf", &(newpid[0]));
            wam.wamDriverData->sc[motor_position[cpuck]].kill_torque = newpid[0];
            finish_entry();
            break;
        case 'K': /* Enter Kill torque for all pucks */
            start_entry();
            addstr("Enter Kill torque: ");
            refresh();
            ret = scanw("%lf", &(newpid[0]));
            fer(cnt, npucks)
            wam.wamDriverData->sc[motor_position[cnt]].kill_torque = newpid[0];
            finish_entry();
            break;
        case 'v': /* Enter trajectory constants for present puck */
            start_entry();
            addstr("Enter trajectory constants \"Velocity Acceleration\": ");
            refresh();
            ret = scanw("%lf %lf", &(newpid[0]), &(newpid[1]));
            SCsettrjprof(&(wam.wamDriverData->sc[motor_position[cpuck]]), newpid[0], newpid[1]);
            finish_entry();
            break;
        case 'V': /* Enter trajectory constants for all pucks */
            start_entry();
            addstr("Enter trajectory constants \"Velocity Acceleration\": ");
            refresh();
            ret = scanw("%lf %lf", &(newpid[0]), &(newpid[1]));
            fer(cnt, npucks)
            SCsettrjprof(&(wam.wamDriverData->sc[motor_position[cnt]]), newpid[0], newpid[1]);
            finish_entry();
            break;

        
        
        //        case 'J':
        //            processJacobian = !processJacobian;
        //            break;
        case 'j':
            start_entry();
            addstr("Enter trajectory filename: ");
            refresh();
            scanw("%s", fn);
            finish_entry();
            playTrajectoryFile(fn, 1);
            break;
        case 'i':
            start_entry();
            addstr("Enter trajectory filename: ");
            refresh();
            scanw("%s", fn);
            finish_entry();
            teachMoveManual(fn);
            break;
        case 'I':
            start_entry();
            addstr("Enter trajectory filename: ");
            refresh();
            scanw("%s", fn);
            finish_entry();
            teachMove(fn);
            break;
        case 'l':
        //showLimits = !showLimits;
        //if(showLimits)
        //    getSafetyLimits(&jVel, &tVel, &eVel);
            break;
        
        /*        case 'N': // Enter gravity compensation constants by hand
                    start_entry();
                    addstr("Enter GComp constants \"A B C D\": ");
                    refresh();
                    ret = scanw("%lf %lf %lf %lf", &A,&B,&C,&D);

                    setGcompValues(A,B,C,D);
                    finish_entry();
                    break;
        */
        
        
        /*        case 'h':   // Toggle help display
                    clear();
                    disp_help = !disp_help;
                    if (disp_help) present_screen = HELP_SCREEN;
                    break;*/
        /*        case 'g':
                    clear();
                    if (present_screen == MAIN_SCREEN)
                      present_screen = GCOMP_SCREEN;
                    else if (present_screen == GCOMP_SCREEN)
                      present_screen = MAIN_SCREEN;

                  break;*/
        
        case 'g':   /* Get gimbals */
        //showGimbals = !showGimbals;

            break;

        case 'E':   /* Get Saved bus error */
        //saved_bus_error++;
            //getProperty(wam.wamDriverData->act[0].bus, SAFETY_MODULE, SAVED_BUS_ERR, &saved_bus_error);
            //setProperty(wam.wamDriverData->act[0].bus, SAFETY_MODULE, SAVED_BUS_ERR, TRUE, 0);

            //showBusError = 1;
            break;
        
        case 'o':
            setProperty(wam.wamDriverData->act[0].bus, SAFETY_MODULE, CONTROLLERS, TRUE, 0x1B);
            break;
        /*        case 'j':
                  fer(cnt, npucks)
                  {
                    offsets[cnt] = GetProp(cnt,POSITION_OFFSET);
                  }
                  break;*/

#endif       


/** Evaluate a playlist

  checks to see if we have reached the next point and loads a new trajectory if we have.
*/
void WACKYeval()
{
    int ret,cnt;
    double pos,vel,acc,min,max;

    if (WACKYplay)
    {
        ret = 0;
        for (cnt = 0; cnt < npucks; cnt++)
        {
            if (wam.wamDriverData->sc[cnt].trj.state == 0)
            {
                switch(cnt)
                {
                    case 0: // J1
                    min = -2.45;
                    max = +2.45;
                    break;
                    case 1: // J2
                    min = -1.95;
                    max = +1.95;
                    break;
                    case 2: // J3
                    min = -2.78;
                    max = +2.78;
                    break;
                    case 3: // J4
                    min = -0.90;
                    max = +2.00;
                    break;
                    default:
                    min = -1.40;
                    max = +1.40;
                    break;
                }
                pos = drand48() * (max - min) + min;
                //vel = (0.5-drand48()) * 1.4 * 2;
                SCstarttrj(&(wam.wamDriverData->sc[cnt]), pos);
            }
        }
    }
}
/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *             5A                                                         *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *                                                                      *
 *                        SUPPORTED COMPILERS                           *
 *                                                                      *
 *  This software has been compiled and tested with the following       *
 *  compilers:                                                          *
 *
 *  - QNX qcc
 *                                                                      *
 *  ******************************************************************  *
 *
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/
        
