///////////////////////////////////////////////////////////
//                                                       //
//           Automatically Generated C++ Code            //
//           BHand Control Center Version 1.0            //
//                                                       //
//                   Real Time Mode                      //
//                                                       //
//		BarrettHand Demo Program                         //
//      v1.0 990803 dr - Created Program                 //
//      v2.0 001120 ac - Modified Program, see note below   
//      v4.5 030919 bz - Made acceleration smooth
//						Tries to re-init 5 times after error
//						Other minor enhancements
//
///////////////////////////////////////////////////////////

// Modified on 001120 by ac to allow user selection of serial port.


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <conio.h>
#include <ncurses.h>


#include "BHand.h"

BHand  bh;           // Handles all hand communication
int    value;        // Hand parameter obtained with Get
int    result;       // Return value (error) of all BHand calls
int    port;		 // User's chosen serial port

int    szbuffer;     // Size of data buffers
int*   pdata[4][8];  // Pointers to data buffers
char   chr;

char _kbhit(void)
{
    //char chr;
    return((chr = getch()) != ERR);
}

char _getch(void)
{
    return(chr);
}

///////////////////////////////////////////////////////////
// Error Handler - called whenever result!=0
// Feel free to customize this function

void Error(void)
{
	int err = 0;
    char errStr[200];
    
    sprintf(errStr,"ERROR: %d\n%s\n", result, bh.ErrorMessage(result) );
	addstr( errStr );refresh();
	

	//Re-initialize the hand - BZ 19Sep2003
	addstr( "\r\t\tInitializing Hand...             " );refresh();
	while(err < 5)
	{
		if( result=bh.InitHand("") )
			++err;
		else
			break;
	}
	if(err == 5)
	{
		addstr("\n\nCould not initialize hand!  Check cable & power.");
		addstr("\n\nPress Ctrl-C to exit..."); refresh();
		while(1);
	}

}


///////////////////////////////////////////////////////////
//  Initialize hand, set timeouts and baud rate

void Initialize(void)
{

	addstr( "\n\n\r\t\tInitializing Software..." );
	addstr( "\n\n\r\t\tPlease enter desired serial port> " ); refresh();
	//scanf("%d", &port);
    while(getch()!=ERR){}
    echo();
    timeout( -1);
	scanw("%d", &port);
    noecho();
    timeout(0);
    addstr( "\n" );

	if( result=bh.InitSoftware(port,THREAD_PRIORITY_TIME_CRITICAL) )   //allow use of any serial port
		Error();

	if( result=bh.ComSetTimeouts(0,100,25000,100,5000) )
		Error();

	//if( result=bh.Baud(38400) )
	//	Error();

//AC 010711 Removed FDZ and FIP setting for compliance w/ firmware v4.21 (version 4.21)

//	if( result=bh.Command("SFSET FDZ 251") )
//		Error();

//	if( result=bh.Command("SFSET FIP 0") )
//		Error();

	if( result=bh.Command ( "SFSET MPE 1000" ) )  //AC 010711 spread MPE was 100 changed so
		Error();								 //program would crash less	(version 4.21)

	addstr( "\r\t\tInitializing Hand...             " );

	if( result=bh.InitHand("") )
		Error();


}




///////////////////////////////////////////////////////////
//  Set parameters, allocate data buffers, load files

void PrepareRealTime(void)
{
	szbuffer = 2000;

	for( int m=0; m<4; m++ )
		for( int v=0; v<8; v++ )
		{
			pdata[m][v] = new int[szbuffer];
			memset((void*)pdata[m][v], 0, szbuffer*sizeof(int));
		}

	if( result=bh.RTSetFlags( "123", 1, 1, 0, 0, 1, 0, 1, 0, 1 ))
		Error();

}


///////////////////////////////////////////////////////////
//  Run RealRime loop, return 1 if interrupted with a key

int RunRealTime1(void)
{
	double var[4][7];
	int N=0;
	DWORD  time, tmstart;
	bool terminate=false;
	int pos, m;
    char a[30];
// AC 010711 Set control parameters so RT velocity control will work w/ firmware v4.21 (revision 4.21)
/*	if( result=bh.Command ( "123FSET SAMPLE 16" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

 Don't mess with the spread - BZ 19Sep2003
	if( result=bh.Command ( "SFSET FDZ 248" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }	
	
	if( result=bh.Command ( "SFSET FIP 0" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SFSET FPG 100" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FDZ 221" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }	
	
	if( result=bh.Command ( "GFSET FIP 66" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FPG 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
*/

	if( result=bh.Command ( "123FSET SAMPLE 63" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FPG 42" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

    if( result=bh.RTStart( "123" ))
        Error();
        
	tmstart = GetTickCount();
    
	if( result=bh.RTUpdate())
        Error();

	while( !terminate && !_kbhit() )
	{
		time = GetTickCount() - tmstart;
        //sprintf(a,"%ld.",GetTickCount());addstr(a);refresh();

		pos = N%szbuffer;
		for( m=0; m<4; m++ )
		{
			pdata[m][0][pos] = bh.RTGetPosition( m+'1' );
			pdata[m][1][pos] = bh.RTGetVelocity( m+'1' );
			pdata[m][2][pos] = bh.RTGetStrain( m+'1' );
			pdata[m][3][pos] = bh.RTGetDeltaPos( m+'1' );
			pdata[m][5][pos] = (int)time;
		}
		pos = (N+1)%szbuffer;

		for( int m=0; m<4; m++ )
		{
			var[m][0] = bh.RTGetPosition( m+'1' );
			var[m][1] = bh.RTGetVelocity( m+'1' );
			var[m][2] = bh.RTGetStrain( m+'1' );
			var[m][3] = bh.RTGetDeltaPos( m+'1' );
			var[m][4] = pdata[m][4][N%szbuffer];
			var[m][5] = (double)time;
			var[m][6] = (double)N;
		}

		value = (int)((200.00) * (sin( (2.00) * ((3.14) * ((var[0][5]) / (2000.00))) )));
		bh.RTSetVelocity( '1', value );
		pdata[0][6][pos] = value;

		value = (int)((200.00) * (sin( ((2.00) * ((3.14) * ((var[1][5]) / (2000.00)))) - ((2.00) * ((3.14) / (3.00))) )));
		bh.RTSetVelocity( '2', value );
		pdata[1][6][pos] = value;

		value = (int)((200.00) * (sin( ((2.00) * ((3.14) * ((var[2][5]) / (2000.00)))) - ((3.14) / (3.00)) )));
		bh.RTSetVelocity( '3', value );
		pdata[2][6][pos] = value;

		terminate = (0<(int)((var[0][5]) > (8000.00)));

		N++;
		if( result=bh.RTUpdate())
            Error();
	}

	if( result=bh.RTAbort())
        Error();
        
	
	if( result=bh.Command ( "FDEF" ) ) //AC 010711 Restore control parameters to defaults (version 4.21)
		Error();						
	
	if( _kbhit() )
		{ _getch(); return 1; }	
	else
		return 0;
}

int RunRealTime2(void)
{
	double var[4][7];
	int N=0;
	DWORD  time, tmstart;
	bool terminate=false;
	int pos, m;

// AC 010711 Set control parameters so RT velocity control will work w/ firmware v4.21 (revision 4.21)
/*	if( result=bh.Command ( "123FSET SAMPLE 16" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

 Don't mess with the spread - BZ 19Sep2003
	if( result=bh.Command ( "SFSET FDZ 248" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }	
	
	if( result=bh.Command ( "SFSET FIP 0" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SFSET FPG 100" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FDZ 221" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }	
	
	if( result=bh.Command ( "GFSET FIP 66" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FPG 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
*/

	if( result=bh.Command ( "GFSET SAMPLE 63" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GFSET FPG 42" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	bh.RTStart( "123" );
	tmstart = GetTickCount();
	bh.RTUpdate();

	while( !terminate && !_kbhit() )
	{
		time = GetTickCount() - tmstart;

		pos = N%szbuffer;
		for( m=0; m<4; m++ )
		{
			pdata[m][0][pos] = bh.RTGetPosition( m+'1' );
			pdata[m][1][pos] = bh.RTGetVelocity( m+'1' );
			pdata[m][2][pos] = bh.RTGetStrain( m+'1' );
			pdata[m][3][pos] = bh.RTGetDeltaPos( m+'1' );
			pdata[m][5][pos] = (int)time;
		}
		pos = (N+1)%szbuffer;

		for( int m=0; m<4; m++ )
		{
			var[m][0] = bh.RTGetPosition( m+'1' );
			var[m][1] = bh.RTGetVelocity( m+'1' );
			var[m][2] = bh.RTGetStrain( m+'1' );
			var[m][3] = bh.RTGetDeltaPos( m+'1' );
			var[m][4] = pdata[m][4][N%szbuffer];
			var[m][5] = (double)time;
			var[m][6] = (double)N;
		}

		value = (var[0][0] > 16000) ? 0 : (int)(215.00);
		bh.RTSetVelocity( '1', value );
		pdata[0][6][pos] = value;
		
		value = (int)(((var[2][0]) > (5000.00)) ? (250.00) : (0.00));
		bh.RTSetVelocity( '2', value );
		pdata[1][6][pos] = value;
		
		value = (int)(((var[0][0]) > (5000.00)) ? (235.00) : (0.00));
		bh.RTSetVelocity( '3', value );
		pdata[2][6][pos] = value;
		
		terminate = (0<(int)((var[0][5]) > (2000.00)));

		N++;
		bh.RTUpdate();
	}

	bh.RTAbort();
	
	if( result=bh.Command ( "FDEF" ) ) //AC 010711 Restore control parameters to defaults (version 4.21)
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	else
		return 0;
}

int Execute1(void)
{
	int i;

	/* Set up FPG, SAMPLE, ACCEL, and Max Velocities for fingers 1-3 - BZ 19Sep2003 */
	if( result=bh.Command ( "123FSET FPG 10" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	
	if( result=bh.Command ( "123FSET MCV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123FSET MOV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123FSET SAMPLE 63" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	
	if( result=bh.Command ( "123FSET ACCEL 4" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* Set up FPG, SAMPLE, ACCEL, and Max Velocities for spread - BZ 19Sep2003 */
	if( result=bh.Command ( "4FSET FPG 8" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	
	if( result=bh.Command ( "4FSET MCV 100" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "4FSET MOV 100" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "4FSET SAMPLE 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	
	if( result=bh.Command ( "4FSET ACCEL 8" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* Perform supervisory demonstration */
	if( result=bh.Command ( "GO" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SO" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SC" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1M 5000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3M 10000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2M 15000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123SO" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

//AC 010711 Removed following command because it was casing fingers to hit eachother (version 4.21)

	//if( result=bh.Command ( "123SC" ) )
	//	Error();
	//if( _kbhit() )
	//	{ _getch(); return 1; }

	if( result=bh.Command ( "SC" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }
	
	if( result=bh.Command ( "123C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SM 1570" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "12M 11500" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	
	/* Do the "Back and Forth" */
	if( result=bh.Command ( "3FSET DP 11500" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	for(i = 0; i < 3; i++)
	{
		if( result=bh.Command ( "1FSET DP 11500" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }

		if( result=bh.Command ( "2FSET DP 6000" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }

		if( result=bh.Command ( "GM" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }
		
		if( result=bh.Command ( "1FSET DP 6000" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }

		if( result=bh.Command ( "2FSET DP 11500" ) )
			Error();

		if( result=bh.Command ( "GM" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }
	}

	if( result=bh.Command ( "123M 11500" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* End of Back and Forth */

	if( result=bh.Command ( "1FSET MCV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2FSET MCV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3FSET MCV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123M 5000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SC" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123M 2000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* Do finger jumble */

	if( result=bh.Command ( "SM 1000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2M 14000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3M 11500" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* End finger Jumble */

	if( result=bh.Command ( "1FSET MCV 65" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2FSET MCV 65" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3FSET MCV 85" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "SFSET MCV 50" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123SC" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1FSET MCV 75" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2FSET MCV 75" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3FSET MCV 75" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1M 13000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3M 13000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2M 13000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1M 17000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3M 17000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2M 17000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1M 9000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "3M 9000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "2M 9000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "1234O" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	/* Do Spread Inc Close with finger movement */

	if( result=bh.Command ( "SFSET MCV 100" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123FSET MCV 200" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "123FSET DP 4000" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "4FSET DS 630" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	for(i = 0; i < 5; i++)
	{
		if( result=bh.Command ( "123M 11000" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }

		if( result=bh.Command ( "123M" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }

		if( result=bh.Command ( "SIC" ) )
			Error();
		if( _kbhit() )
			{ _getch(); return 1; }
	}


	if( result=bh.Command ( "C" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	if( result=bh.Command ( "GO" ) )
		Error();
	if( _kbhit() )
		{ _getch(); return 1; }

	return 0;
}

/*
///////////////////////////////////////////////////////////
//  Save all buffers into a text file

void SaveData(char* name)
{
	FILE* fp = fopen(name, "wt");
	if( !fp )
		return;

	for( int r=0; r<szbuffer; r++ )
	{
		for( int v=0; v<8; v++ )
		{
			for( int m=0; m<4; m++ )
				faddstr(fp, "%d ", pdata[m][v][r]);
			faddstr(fp, "  ");
		}
		faddstr(fp, "\n");
	}

	fclose(fp);
}
*/

void PrintBanner()
{
	//system("cls");
	addstr("\n\t\t---------------------------------------");
	addstr("\n\t\t|                                     |");
	addstr("\n\t\t|     Barrett Technology, Inc.        |");
	addstr("\n\t\t|     BarrettHand Software Example    |");
	addstr("\n\t\t|     BarrettHand Demo Program v4.5   |");
	addstr("\n\t\t|                                     |");
	addstr("\n\t\t---------------------------------------");
    refresh();


}

void exitProgram(void)
{
    bh.Command("T");
    bh.Reset();
    endwin(); 
    exit(0); 
}

///////////////////////////////////////////////////////////
//  Main function - initialize, execute

void main(void)
{
	int i=0;
	
    /* Initialize the ncurses screen library */
    initscr();
    refresh();
    cbreak();
    noecho();
    timeout(0);
    clear();
	
	PrintBanner();
	Initialize();
	
	addstr("\n\n\r\t\tThis program will run a demonstration until a key is pressed.");
    refresh();
	PrepareRealTime();

	while(1)
	{
        if(Execute1())
		{
			exitProgram();
		}
		if(RunRealTime1())
		{
			exitProgram();
		}

		if(Execute1())
		{
			exitProgram();
		}
		
		for(i=0;i<3;i++)
		{
			bh.Command("GO");
			if(RunRealTime2())
			{
				exitProgram();
			}
		}
	
	}

	//SaveData( "ConferenceWiggle" );
}
