/* Decode: 
    Traveler Hauptman 06/20/2001

  Decodes binary data log files into text comma delimited files.
  
Binary format:
data type        function
int               # of fields
for each field:
  int             field data type: 0=int 1=long 2=double
  char[8]         field name: 8 character string not 0 terminated (not a std C string)
rest of file:
int               record index number
long              length of data in bytes
void[data length] data
  
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int fieldlist[50];
int fieldsize;
int fieldcnt;
void *data,*dataidx;
char names[400];

int main(int argc, char **argv)
{
  FILE *inf,*outf;
  char infilename[100],outfilename[100];
  int target_record = -1;
  int numfields=0;
  char *nameidx;
  long cnt,idx;
  int current_index;
  long length;
  int *intdata;
  double *doubledata;
  long *longdata;
  
  if (argc < 3 || argc > 4) 
  {
    printf("\nUsage: >convert input_file output_file [record#]\n");
    exit(1);
  }
  printf("\nstarting program\n");
  strcpy(infilename,argv[1]);
  strcpy(outfilename,argv[2]);
  printf("filenames copied\n");
  if (argc == 4) 
  {
    target_record = atoi(argv[3]);
  }
  
  //open input file  
  if ((inf = fopen(infilename,"rb"))==NULL)
  {
    printf("\nUnable to open input file: %s\n",infilename);
    exit(1);
  }
  printf("Input file opened\n");
  //open output file
  if ((outf = fopen(outfilename,"w"))==NULL)
  {
    printf("\nUnable to open input file: %s\n",outfilename);
    exit(1);
  }
  printf("Output file opened\n");
  //figure out fields
  //print fields 
  fread(&fieldcnt,sizeof(int),1,inf);
  nameidx = names;
  printf("field cnt:%d\n",fieldcnt);
  for (cnt = 0;cnt < fieldcnt;cnt++)
  {
    fread(&(fieldlist[cnt]),sizeof(int),1,inf);
    printf("\n-%d- ",fieldlist[cnt]);
    fread(&(names[8*cnt]),sizeof(char),8,inf);
    for (idx=0;idx<8;idx++)
    {     
      putchar(names[idx+8*cnt]);
    } 
  }
  printf("\nHeader read & processed\n");

  while(!feof(inf))
  {
    if(fread(&current_index,sizeof(int),1,inf)==0) break;
    fread(&length,sizeof(long),1,inf);
    printf("index: %d length: %ld\n",current_index,length);
   
    if ((data = malloc(length))==NULL) 
    {
      printf("\nAborting because of memory allocation error.\n");
      exit(1);
    }
    printf("memory allocated\n");
    fread(data,sizeof(char),length,inf);
    printf("data read\n");

    if (target_record==-1 || target_record==current_index)
    {
      dataidx = data;
      cnt = 0;
      while(cnt<length)
      {
        fprintf(outf,"%d, ",current_index);
        
        for (idx = 0;idx < fieldcnt;idx++)
        {
          switch (fieldlist[idx])
          {
          case 0://integer
            intdata = (int *)dataidx;
            fprintf(outf,"%d",*intdata);
            //printf("did an integer %d\n",*intdata);
            intdata++;
            dataidx = (void *)intdata;
            cnt+=sizeof(int);
            
            break;
          case 1://long
            longdata = (long *)dataidx;
            fprintf(outf,"%ld",*longdata);
            longdata++;
            dataidx = (void *)longdata;
            //printf("did an long\n");
            cnt+=sizeof(long);
            break;
          case 2://double
            doubledata = (double *)dataidx;
            fprintf(outf,"%g",*doubledata);
            doubledata++;
            dataidx = (void *)doubledata;
            //printf("did an double\n");
            cnt+=sizeof(double);
            break;       
          }
          if (idx < fieldcnt-1) fprintf(outf,", ");
        
        }
        fprintf(outf,"\n");
      }
    } 
    free(data);   
  }
  fclose(inf);
  fclose(outf);
}
