#ifndef _BTWAM_H
#define _BTWAM_H

#include "btsystem.h"
#include "SimpleControl.h"

#define WAM2004

//joint ratios
#ifndef WAM2004
#define mN1  35.87
#define mN2  28.21
#define mN3  28.21
#define mN4  17.77
#define mn3  1.68
#endif

#ifdef WAM2004
#define mN1  42.0
#define mN2  28.25
#define mN3  28.25
#define mN4  18.0
#define mn3  1.68
#endif

#define mN5  10.27
#define mN6  10.27
#define mN7  14.93
#define mn6  1


typedef struct {
  double q[10];
} wam_vector;

typedef struct {
  int id;
  double A,B,C,D,E,F;
  int Gcomp,Ocomp; //0 = no gravity comp, 1 = gravity comp
  actuator_struct *act;
  int num_actuators;
  int wrist_attached; //0 no wrist, 1 yes wrist is attached.
  int motor_position[7];
  int zero_order[7];
  wam_vector zero_offsets;
  wam_vector stop_torque;
  wam_vector park_location;
  SimpleCtl sc[7];
  wam_vector Mpos,Mtrq,Jpos,Jtrq;
  wam_vector Kp,Kd,Ki,saturation;
  wam_vector vel,acc;
  double sample_rate;
}wam_struct;



/* Define a structure to hold the output of the whereAmI() routine */
typedef struct 
{ 
    double x;                // Translation in x direction
    double y;                // Translation in y direction
    double z;                // Translation in z direction
    double rotMatrix[3][3];  // The rotation matrix
}frame; 

/* Define WAM types for the whereAmI() routine.
    4DOF    = 4-DOF WAM with standard outer link
    4DOF_G  = 4-DOF WAM with gimbals on outer link
    7DOF    = 7-DOF WAM
*/
enum {WAM_4DOF, WAM_4DOF_G, WAM_7DOF} wamType;


int ActAngle2Mpos(wam_vector *Mpos); //Extracts actuators 0 thru 6 into wam_vector positions
int Mtrq2ActTrq(wam_vector *Mtrq);  //Packs wam_vector of torques into actuator array
void Mpos2Jpos(wam_vector * Mpos, wam_vector * Jpos); //convert motor angle to joint angle
void Jpos2Mpos(wam_vector * Jpos, wam_vector * Mpos);
void Jtrq2Mtrq(wam_vector * Jtrq, wam_vector * Mtrq); //conbert joint torque to motor torque
void calcGcomp(wam_vector * Jpos, wam_vector * Jtrq,double A,double B,double C,double D); //Calculate gravity compensation torques
//void calcGcomp2(wam_vector * t2, wam_vector * t3,wam_vector * t4);
//void setGcomp(int onoff);
//void setGcompValues(double A,double B,double C,double D);
//void getGcompValues(double *A,double *B,double *C,double *D);

void getOcompValues(double *E,double *F);
void toggleOcomp();
//void getGcompvalues();
void Jpos2Cpos(wam_vector * Jpos, wam_vector * Cpos);
//void GCompCollectData(const char *filename,int j2,int j3, int j4);
//void GCompFileDump(FILE *out);
//void GCompSample(wam_vector *res, double p1, double p2, double p3, double p4);


void FindWAMstop(int idx);
void FindWAMcenter(int idx);
void GentleFindWAMcenter(int idx);
void GentleFindWAMstop(int idx);


void zero_wam_vector(wam_vector *wv);
void set_wam_vector(wam_vector *dest, double s0, double s1, double s2, double s3, double s4, double s5, double s6);
void scale_wam_vector(wam_vector *wv, double a);
void copy_wam_vector(wam_vector *dest, wam_vector *scr);	
double dist_wam_vector(wam_vector *dest, wam_vector *src);
int read_wam_vector(FILE *in,wam_vector *wv);
int write_wam_vector(FILE *out,wam_vector *wv);
int dump_wam_vector(wam_vector *wv);

void getWAMjoints(wam_vector *Mpos,wam_vector *Mtrq,wam_vector *Jpos,wam_vector *Jtrq);
void getWAMmotor_position(int *mp);

int whereAmI(frame *theFrame, int wamType);
wam_struct * GetWAM(void);
int playViaTrajectoryFile(char *fileName, double timeScale);


#endif _BTWAM_H


/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/
