/*
   Module .............libbt                                
   File ...............doubblebuffer.c                                     
   Author .............Traveler Hauptman                                 
   Creation Date ......April 3, 2002                                 
                                                                       
              Copyright (c) 2002 Barrett Technology, Inc.              
                         139 Main Street                               
                        Kendall/MIT Square                             
                   Cambridge, MA  02142-1528  USA                      
                                                                      
                         All rights reserved.                          
                                                                       
 ***********************************************************************
 *                            DISCLAIMER                               *
 *                                                                     *
 *  This software and related documentation are provided to you on     *
 *  an as is basis and without warranty of any kind.  No warranties,   *
 *  express or implied, including, without limitation, any warranties  *
 *  of merchantability or fitness for a particular purpose are being   *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett    *
 *  Technology, Inc. be liable for any lost development expenses, lost *
 *  lost profits, or any incidental, special, or consequential damage. *
 *                                                                     *
 ***********************************************************************  
 *                                                                      
 *  NOTES:                                                              
 *   Uses some code & concepts developed at Nortwestern University 
 *   by Traveler Hauptman                                                                      
 *  REVISION HISTORY:                                                   
 *                                           
 *                                                                      
 *======================================================================*/


#include <syslog.h>
#include "doublebuffer.h"
#include <stdio.h>
#include <stdlib.h>
/*****************************************************************************/

int InitDL(DoubleBuffer_struct *db,int size, char *filename)
{
  db->Log_Data = 0;
  db->DLwritten = 0;
  db->Log_File_Open = 0;
  db->DLctr = 0;
  db->DLidx = 0;
  if ((db->DLfile = fopen(filename,"w"))==NULL)
  {
    syslog(LOG_ERR,"InitDL:Could not open datalogger file %s",filename);
    return -1;
  }
  db->Log_File_Open = 1;
  db->buffersize = size;
  if ((db->DLbuffer1 = malloc(size * sizeof(data_to_log)))==NULL)
  {
    syslog(LOG_ERR,"InitDL:Could not allocate buffer1",filename);
    fclose(db->DLfile);
    return -1;
  }
  if ((db->DLbuffer2 = malloc(size * sizeof(data_to_log)))==NULL)
  {
    syslog(LOG_ERR,"InitDL:Could not allocate buffer2",filename);
    free(db->DLbuffer1);
    fclose(db->DLfile);
    return -1;
  }
  InitDataFile(db);
  db->DL = db->DLbuffer1;
  return 0;
}

void DLon(DoubleBuffer_struct *db)
{
  db->Log_Data = 1;
}

void DLoff(DoubleBuffer_struct *db)
{
  db->Log_Data = 0;
}

void CloseDL(DoubleBuffer_struct *db)
{
  free(db->DLbuffer1);
  free(db->DLbuffer2);
  fclose(db->DLfile);
}

data_to_log * GetDLptr(DoubleBuffer_struct *db)
{

  return &(db->DL[db->DLidx]);
}

//Writes the header of a binary data file.
void InitDataFile(DoubleBuffer_struct *db)
{ 
  FILE *datfile;
  int tmp;
  long len;
  
  datfile = db->DLfile;
  tmp = 7;                               
  fwrite(&tmp,sizeof(int),1,datfile);   //6 fields
  
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J1 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J2 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J3 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J4 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J5 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J6 pos  ");
  tmp = 2;
  fwrite(&tmp,sizeof(int),1,datfile);   //field = double
  fprintf(datfile,"J7 pos  ");
  
}

/* UpdateDL increments the DL index and switches the pointer to the other buffer if necessary*/
void UpdateDL(DoubleBuffer_struct *db)
{
  db->DLidx++;                      //Increment current array index
  
  if (db->DLidx >= db->buffersize)            //If our index exceeds our array size
  {
    db->DLidx = 0;                  //reset our index to zero
    
    if (db->DL == db->DLbuffer1)        //If we are currently pointed to buffer 1
    {
      db->DL = db->DLbuffer2;           //Point to buffer 2
      db->DLwritten = 1;            //Indicate that buffer 1 is full
    }
    else                        //else
    { 
      db->DL = db->DLbuffer1;           //Point to buffer 1
      db->DLwritten = 2;            //Indicate that buffer 2 is full
    }
  }
}

/* DLwrite checks to see if one of the buffers got filled and if so, writes it to a file */
void DLwrite(DoubleBuffer_struct *db)
{
  data_to_log *DLout;
  int Ridx;
  long Rlength;
  
  if (db->Log_Data) //If data logging is turned on
  {
    if (db->DLwritten){               //If any buffer is full
      if (db->Log_File_Open){ //If our data logging files are open and everything is peachy
      
        if (db->DLwritten == 1)       //If buffer1 is full
          DLout = db->DLbuffer1;      //Set DLout to point the same place as DLbuffer1
        else if (db->DLwritten == 2)  //Else If buffer2 is full
          DLout = db->DLbuffer2;      //Set DLout to point the same place as DLbuffer1
          
        Ridx = db->DLctr;                           //Record index = full buffer counter
        fwrite(&Ridx,sizeof(int),1,db->DLfile);     //Write Record index as a binary integer
        Rlength = db->buffersize*sizeof(data_to_log);         //Calculate the Record length
        fwrite(&Rlength,sizeof(long),1,db->DLfile);     //Write the Record length in bytes
        
        fwrite(DLout,sizeof(data_to_log),db->buffersize,db->DLfile);  //Write all the data in binary form
        db->DLctr++;                                        //increment the record index                        
      }
      db->DLwritten=0;                                      //Reset full buffer indicator to zero
    }
  }
}
void evalDL(DoubleBuffer_struct *db)
{
  if (db->DLwritten) DLwrite(db);  
}

