/* demo.trj
Time    J1  J2  J3  J4  J5  J6  J7
0.0     0.0 1.5 1.5 0.0 1.1 1.1 0.0
0.1     0.2 1.7 1.7 0.0 1.1 1.1 0.2
0.2     0.4 1.9 1.8 0.0 1.1 1.1 0.5
*/
#include <syslog.h>
#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/syspage.h>
#include <sys/neutrino.h>
#include <math.h>

#include "SimpleControl.h"
#include "btwam.h"
#include "traj.h"

#define MAX_LINE_LENGTH     255

enum{ TEACH_DONE, TEACH_ADD_ONE, TEACH_ADD_CONTINUOUS };

extern int shutdown_threads;
int teachMode = TEACH_DONE;
FILE *teachFile;
extern uint64_t ms_timer;
uint64_t startTime;
double teachTime;
pthread_t record_thd_id;

void RecordingThread(void);

#if 0
void evaluateTrajectory(pointList_t *pts, int idx, double dt, double *retval)
{
       
}
#endif

void evaluate_traj_file(trajectory *traj, int idx, double dt, double *retVal)
{
    trajectory *dummy;
    //double retVal;
    
    dummy = traj;
    if (traj->state == TRAJ_FILE)
    {
        //If this is the first time through, initialize the trajectory
        if(traj->t == 0)
        {
            syslog(LOG_ERR, "traj->t == 0");
            traj->row = 0;
            traj->t2 = 0;
            traj->x2 = (traj->tf)[traj->row][idx+1];
        }
        
        traj->t += dt;
        
        if(traj->t >= traj->t2)
        {
            traj->t1 = traj->t2;
            traj->x1 = traj->x2;
            ++traj->row;
            if(traj->row >= traj->rows)
            {
                traj->state = TRAJ_DONE;
                traj->cmd = 0;
            }
            else
            {
                traj->t2 = (traj->tf)[traj->row][0];
                traj->x2 = (traj->tf)[traj->row][idx+1];
                traj->end = (traj->tf)[traj->row][idx+1];
                traj->cmd = (traj->x2 - traj->x1) * dt / (traj->t2 - traj->t1);
                syslog(LOG_ERR, "IDX %d: cmd = %lf, x2 = %lf, x1 = %lf, t2 = %lf, t1 = %lf",
                                idx, traj->cmd, traj->x2, traj->x1, traj->t2, traj->t1);
            }
        }
        if (isnan(traj->cmd)){
          syslog(LOG_ERR,"evaluate_traj_file: returning NAN");
          syslog(LOG_ERR,"evaluate_traj_file: Old *traj %lx New *traj %lx",traj,dummy);
        }
        
        *retVal = traj->cmd;
        
        //return(retVal);
    }
    //else
    //    return(0.0);
}

int init_traj_file(char *fileName, double ***trajList, int *rows, int *columns)
{
    char        line[MAX_LINE_LENGTH];
    int         charCt, i, row, columnCt;
    double      a[8];
    FILE        *inFile;
    
    //Open the file
    if((inFile = fopen(fileName, "r")) == NULL)
    {
        syslog(LOG_ERR, "Trajectory file '%s' not found!", fileName);
        return(1);
    }
    syslog(LOG_ERR, "Opened the file");
    
    //Scan the file
    *rows = 0;
    while(1)
    {
        if(fgets(line, MAX_LINE_LENGTH, inFile) == NULL)        
            break; //Reached EOF
        if(line[0] == '#')
        {
            syslog(LOG_ERR, "Found a comment: [%s]", line);
            continue; //Skip comments
        }
        *columns = sscanf(line, "%lf %lf %lf %lf %lf %lf %lf %lf",
                            &a[0],&a[1],&a[2],&a[3],&a[4],&a[5],&a[6],&a[7]);
        if(*columns > 1)
        {
            syslog(LOG_ERR, "Read row %d with %d columns", *rows, *columns);
            columnCt = *columns;
            ++*rows;
        }
    }
    
    //Allocate memory for the file
    if( (*trajList = (double **)malloc(*rows * sizeof(double *))) == NULL)
    {
        syslog(LOG_ERR, "Insufficient memory to load trajectory file.  malloc() failed for %d rows", *rows);
        return(2);
    }
    syslog(LOG_ERR, "Initialized the main array: %d rows", *rows);
    for(i = 0; i < *rows; i++)
    {
        if(((*trajList)[i] = (double *)malloc(columnCt * sizeof(double))) == NULL)
        {
            syslog(LOG_ERR, "Insufficient memory to load trajectory file.  malloc() failed for %d columns", columnCt);
            return(2);
        }
    }
    syslog(LOG_ERR, "Initialized the sub-arrays");

    //Load the file into memory
    rewind(inFile);
    row = 0;
    while(1)
    {
        if(fgets(line, MAX_LINE_LENGTH, inFile) == NULL)
            break; //Reached EOF
        if(line[0] == '#')
        {
            syslog(LOG_ERR, "Found a comment: [%s]", line);
            continue; //Skip comments
        }
        *columns = sscanf(line, "%lf %lf %lf %lf %lf %lf %lf %lf",
                            &a[0],&a[1],&a[2],&a[3],&a[4],&a[5],&a[6],&a[7]);
        if(*columns > 1)
        {    
            for(i = 0; i < *columns; i++)
                (*trajList)[row][i] = a[i];
            ++row;
        }
    }
    syslog(LOG_ERR, "Loaded the data into the arrays");
    //Close the file
    fclose(inFile);
    
    *columns = columnCt;
    
    return(0);
}

int teachMove(char *fileName)
{
    wam_vector jpos, garbage;
    char buffer[2];
    int readCt;
    int fd;
    
    fd = fileno(stdin);
    
    //Create the teachFile
    if((teachFile = fopen(fileName, "w")) == NULL)
    {
        syslog(LOG_ERR, "teachMove: could not create teachFile %s", fileName);
        return(1);
    }
    fprintf(teachFile, "# %s\n# WAM trajectory file\n#\n# Format:\n"
                       "# Time(s)  J1  J2  J3  J4  J5  J6  J7\n", fileName);
                           
    teachTime = 0;
    //Add the present position to the output file
    getWAMjoints(&garbage, &garbage, &jpos, &garbage);
    fprintf(teachFile, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
        teachTime, jpos.q[0], jpos.q[1], jpos.q[2], 
        jpos.q[3], jpos.q[4], jpos.q[5], jpos.q[6]);
    
    //Add points to teachFile
    startTime = ClockCycles();
    teachMode = TEACH_ADD_CONTINUOUS;
    
    do
    {
        usleep(100000);
        readCt = read(fd, buffer, 2);
    }while(buffer[0] != 27); //Hit ESC to stop
    
    //Stop the recording
    teachMode = TEACH_DONE;
    usleep(200000); //Wait for recording to complete
    
    //Close the teachFile
    fclose(teachFile);
    
    //Clear the file pointer
    teachFile = NULL;
    
    return(0);
}

int teachMoveManual(char *fileName)
{
    char c;
    wam_vector jpos,jlast, garbage;
    char buffer[2];
    int readCt;
    int fd;
    
    fd = fileno(stdin);
    
    //Record joint angles upon user input
    //Hit any key to record present position
    //Hit ESC to stop
    
    //Create the teachFile
    if((teachFile = fopen(fileName, "w")) == NULL)
    {
        syslog(LOG_ERR, "teachMove: could not create teachFile %s", fileName);
        return(1);
    }
    fprintf(teachFile, "# %s\n# WAM trajectory file\n#\n# Format:\n"
                       "# Time(s)  J1  J2  J3  J4  J5  J6  J7\n", fileName);
    //Initialize the time
    teachTime = 0;
    
    //Add the present position to the output file
    getWAMjoints(&garbage, &garbage, &jpos, &garbage);
    fprintf(teachFile, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
        teachTime, jpos.q[0], jpos.q[1], jpos.q[2], 
        jpos.q[3], jpos.q[4], jpos.q[5], jpos.q[6]);
    while(1)
    {
        readCt = read(fd, buffer, 2);
        if(readCt != -1)
        {
            if(buffer[0] == 27)
                break;
            else
            {
                //Increase the time by 5 seconds
               
                
                //Add point to teachFile
                teachMode = TEACH_ADD_ONE;
            }
        }
        usleep(100000);
    }
    
    //Close the teachFile
    fclose(teachFile);
    
    //Clear the file pointer
    teachFile = NULL;
    
    return(0);
}

/** Starts the display thread.
    Creates a new thread, sets its priority, and runs it as a display thread.
*/
void StartRecordingThread()
{
    pthread_attr_t recording_attr;
    struct sched_param recording_param;

    pthread_attr_init(&recording_attr);
    pthread_attr_setinheritsched(&recording_attr, PTHREAD_EXPLICIT_SCHED);
    recording_param.sched_priority = 10;
    pthread_attr_setschedparam(&recording_attr, &recording_param);

    pthread_create(&record_thd_id, &recording_attr, (void*)RecordingThread, 0);
    sched_setscheduler(record_thd_id, SCHED_FIFO, &recording_param);
    if (record_thd_id == -1)
    {
        syslog(LOG_ERR, "StartRecordingThread:Couldn't start recording thread!");
        exit( -1);
    }
}

/** Spins in a loop, updating the screen.
    Runs as its own thread, updates the screen.
*/
void RecordingThread()
{
    wam_vector jpos,jlast, garbage;
    uint64_t now;
    double CPU_cycles_per_second,timescale;
    
    /* Get the cycles-per-second of this machine */
    CPU_cycles_per_second = (double)(SYSPAGE_ENTRY(qtime)->cycles_per_sec);
    
    while (!shutdown_threads)
    {
        
       //Record joint angles at 10Hz
        switch(teachMode)
        {
            case TEACH_ADD_CONTINUOUS:
                now = ClockCycles();
                if(now > startTime)
                    teachTime = (now - startTime) / CPU_cycles_per_second;
                else
                    teachTime = (UINT64_MAX - startTime + now) / CPU_cycles_per_second;
                    
                //Add the present position to the output file
                getWAMjoints(&garbage, &garbage, &jpos, &garbage);
                if(teachFile)
                    fprintf(teachFile, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
                        teachTime, jpos.q[0], jpos.q[1], jpos.q[2], 
                        jpos.q[3], jpos.q[4], jpos.q[5], jpos.q[6]);
            break;
            case TEACH_ADD_ONE:
                //Add the present position to the output file
                copy_wam_vector(&jlast,&jpos);
                getWAMjoints(&garbage, &garbage, &jpos, &garbage);
                timescale = dist_wam_vector(&jlast,&jpos);
                if(teachFile){
                 teachTime=teachTime+2*timescale;
                  fprintf(teachFile, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
                    teachTime, jpos.q[0], jpos.q[1], jpos.q[2], 
                        jpos.q[3], jpos.q[4], jpos.q[5], jpos.q[6]);
                }
                //Reset the teachMode
                teachMode = TEACH_DONE;
            break;
            default:
                //No nothing
            break;
        }
    
        
        usleep(100000); // Run at 10Hz
    }

    pthread_exit(NULL);
}