#include <math.h>
#include <syslog.h>
#include "Dynamics.h"
#include "Kinematics.h"
#include "MatrixMath.h"

static int debug_kin = 0;

static double gravity[3] = { 0.0, 0.0, -9.81 };

//BLOCK MEGAblock;


void InitWAMDynamics(WAMDynamics *wamDyn, int dof)
{

    wamDyn->DOF = dof;
    
    /*  LinkMasses in kilograms
        LinkCOGLoc 0, 1, 2 are x, y, z in meters 
     */
    
    wamDyn->LinkMasses[0]     = 12.044;
    wamDyn->LinkCOGLoc[0][0]  =  0.0;
    wamDyn->LinkCOGLoc[0][1]  =  0.1405;
    wamDyn->LinkCOGLoc[0][2]  = -0.0061;
    
    wamDyn->LinkMasses[1]     =  5.903;
    wamDyn->LinkCOGLoc[1][0]  = -0.0001;
    wamDyn->LinkCOGLoc[1][1]  = -0.0166;
    wamDyn->LinkCOGLoc[1][2]  =  0.0096;

    wamDyn->LinkMasses[2]     =  2.08;
    wamDyn->LinkCOGLoc[2][0]  = -0.0443;
    wamDyn->LinkCOGLoc[2][1]  =  0.2549;
    wamDyn->LinkCOGLoc[2][2]  =  0.0003;
#if 0
    //Blank Link
    wamDyn->LinkMasses[3]     =  1.1348;  //1.1348;0.465;3.2828
    wamDyn->LinkCOGLoc[3][0]  =  0.0146;
    wamDyn->LinkCOGLoc[3][1]  =  0.0001;
    wamDyn->LinkCOGLoc[3][2]  =  0.1308;  //0.1308

    wamDyn->LinkMasses[4]     =  2.1480;//2.148;   //1.1348; gimbals = 1.2; blank and weight = 2.148
    wamDyn->LinkCOGLoc[4][0]  =  0.0000;  //0.0;0.1308; gimbals = 0;
    wamDyn->LinkCOGLoc[4][1]  =  0.0000;  //0.0;0.0001; gimbals = 0;
    wamDyn->LinkCOGLoc[4][2]  =  0.0235;  //0.0235;0.1308; gimbals = 0.07; blank and weight = 0.0235
#endif
#if 1
    //Wrist + Hand    
    wamDyn->LinkMasses[3]     =  2.765;//0.465;
    wamDyn->LinkCOGLoc[3][0]  =  0.0;//0.356;
    wamDyn->LinkCOGLoc[3][1]  =  0.0;
    wamDyn->LinkCOGLoc[3][2]  =  0.17;
    
    wamDyn->LinkMasses[4]     =  0.001;
    wamDyn->LinkCOGLoc[4][0]  =  0.0;
    wamDyn->LinkCOGLoc[4][1]  =  0.0;
    wamDyn->LinkCOGLoc[4][2]  =  0.0;
    
    wamDyn->LinkMasses[5]     =  0.001;
    wamDyn->LinkCOGLoc[5][0]  =  0.0;
    wamDyn->LinkCOGLoc[5][1]  =  0.0;
    wamDyn->LinkCOGLoc[5][2]  =  0.0;
    
    wamDyn->LinkMasses[6]     =  0.001;
    wamDyn->LinkCOGLoc[6][0]  =  0.0;
    wamDyn->LinkCOGLoc[6][1]  =  0.0;
    wamDyn->LinkCOGLoc[6][2]  =  0.0;
#endif
#if 1    
    wamDyn->LinkMasses[7]     =  1.2;
    wamDyn->LinkCOGLoc[7][0]  =  0.0;
    wamDyn->LinkCOGLoc[7][1]  =  0.0;
    wamDyn->LinkCOGLoc[7][2]  =  0.065;
#endif   
}



void GravityAndTipForceTorques(WAMDynamics *wamDyn, WAMKinematics *wamKin, wam_vector *joint_angles, wam_vector *joint_vel, wam_vector *joint_acc, WAMLoc *tipforce, wam_vector *torques, int gravityon)
{

  static double pstar[MAXDOF][3];
  static double first = 1;
  static int i,j,k;
  static double wd[3];
  static double w[3];
  static double v[3];
  static double vd[3];
  static double rotMatrices[MAXDOF][9];
  static double transRotMatrices[MAXDOF][9];
  static double hold1[3], hold2[3], hold3[3];
  static double z0[3] = {0.0, 0.0, 1.0 };
  static double vhat[3];
  static double Fs[MAXDOF][3];
  static double f[3];
  static double nn[3];
  static double I[] = { 1,0,0,0,1,0,0,0,1 };
  static double *R;
  static double *tR;
  static double countdown = 100;
  static WAMLoc outputforce;
 

  if(first == 1)
    {
      
      for(i=0;i<(wamDyn->DOF);i++)

	{
	  pstar[i][0] = wamKin->As[i];
	  pstar[i][1] = wamKin->Ds[i] * sin(wamKin->Alphas[i]);
	  pstar[i][2] = wamKin->Ds[i] * cos(wamKin->Alphas[i]);
	}

      first=0;

    }

  
  findDHtransform_deltaTheta(wamKin, joint_angles);

  wd[0] = wd[1] = wd[2] = 0.0;
  w[0] = w[1] = w[2] = 0.0;
  v[0] = v[1] = v[2] = 0.0;
  if(gravityon)
    {
      vd[0] = gravity[0];
      vd[1] = gravity[1];
      vd[2] = gravity[2];
    }
  else
    {
      vd[0] = 0.0;
      vd[1] = 0.0;
      vd[2] = 0.0;
    }



  // forward recursion
  for(i=0;i<(wamDyn->DOF);i++)
    {

      // transpose is same as inverse in this case, see Robot Toolbox Dynamics notes
      transRotMatrices[i][0] = rotMatrices[i][0] = wamKin->transforms[i][0];
      transRotMatrices[i][3] = rotMatrices[i][1] = wamKin->transforms[i][1];
      transRotMatrices[i][6] = rotMatrices[i][2] = wamKin->transforms[i][2];
      transRotMatrices[i][1] = rotMatrices[i][3] = wamKin->transforms[i][4];
      transRotMatrices[i][4] = rotMatrices[i][4] = wamKin->transforms[i][5];
      transRotMatrices[i][7] = rotMatrices[i][5] = wamKin->transforms[i][6];
      transRotMatrices[i][2] = rotMatrices[i][6] = wamKin->transforms[i][8];
      transRotMatrices[i][5] = rotMatrices[i][7] = wamKin->transforms[i][9];
      transRotMatrices[i][8] = rotMatrices[i][8] = wamKin->transforms[i][10];


      // wd = Rot * (wd + z0 * qdd(i) + cross(w,z0 * qd(i)))
      
      // z0 * qd
      hold2[0] = hold2[1] = 0.0;
      hold2[2] = z0[2] * joint_vel->q[i];

      // cross with w then + z0 * qdd + wd
      cross(hold1, w, hold2);

      hold1[0] = hold1[0] + wd[0];
      hold1[1] = hold1[1] + wd[1];
      hold1[2] = hold1[2] + z0[2] * joint_acc->q[i] + wd[2];
      
      mtrx_mult(wd, transRotMatrices[i], 3, 3, hold1, 3, 1);



      // w = Rot * (w + z0*Qd)

      hold1[0] = hold1[1] = 0;
      hold1[2] = z0[2] * joint_vel->q[i];

      hold1[0] = hold1[0] + w[0];
      hold1[1] = hold1[1] + w[1];
      hold1[2] = hold1[2] + w[2];

      mtrx_mult(w, transRotMatrices[i], 3, 3, hold1, 3, 1);
      


      //vd = cross(wd, pstar) + cross(w, cross(w, pstar)) + tR*vd
      
      cross(hold1, w, pstar[i]);
      cross(hold2, w, hold1);


      mtrx_mult(hold1, transRotMatrices[i], 3, 3, vd, 3, 1);

      cross(hold3, wd, pstar[i]);

      vd[0] = hold3[0] + hold2[0] + hold1[0];
      vd[1] = hold3[1] + hold2[1] + hold1[1];
      vd[2] = hold3[2] + hold2[2] + hold1[2];
      

      // Vhat = cross(wd,r) + cross(w, cross(w,r)) + vd 

      cross(hold2, w, wamDyn->LinkCOGLoc[i]);
      cross(hold1, w, hold2);

      cross(hold2, wd, wamDyn->LinkCOGLoc[i]);
      
      vhat[0] = hold1[0] + hold2[0] + vd[0];
      vhat[1] = hold1[1] + hold2[1] + vd[1];
      vhat[2] = hold1[2] + hold2[2] + vd[2];


      // total force on this link
      Fs[i][0] = vhat[0] * wamDyn->LinkMasses[i];
      Fs[i][1] = vhat[1] * wamDyn->LinkMasses[i];
      Fs[i][2] = vhat[2] * wamDyn->LinkMasses[i];

      // moment on this link.  We are using point masses so not implemented.
      // should be
      // N = link.I * wd + cross(w, link.I*w);
      // where link.I is 6x6 symmetric inertial matrix
      
    }
  


  //backward recursion

  FindTransformedPoint(wamKin,  joint_angles, tipforce, &outputforce);

  //f[0] = tipforce->pos[0];
  //f[1] = tipforce->pos[1];
  //f[2] = tipforce->pos[2];
  f[0] = outputforce.pos[0];
  f[1] = outputforce.pos[1];
  f[2] = outputforce.pos[2];
  
  //  syslog(LOG_ERR, "forces tip %f %f %f\noutput %f %f %f", tipforce->pos[0], tipforce->pos[1] ,tipforce->pos[2], outputforce.pos[0], outputforce.pos[1], outputforce.pos[2]);
  nn[0] = outputforce.angle[0];
  nn[1] = outputforce.angle[1];
  nn[2] = outputforce.angle[2];


  for(i=((wamDyn->DOF)-1); i>=0; i--)
    {
      
      if(i==((wamDyn->DOF)-1))
	{
	  R = I;
	  tR = I;
	}
      else
	{
	  R = rotMatrices[i+1];
	  tR = transRotMatrices[i+1];
	}
      
      // nn = R * (nn + cross(R' * pstar, f)) + 
      // cross(pstar + r, Fs(i)) +  --- N(i) {no N(i) for us}

      // cross pstar + r, Fm(i)

      hold2[0] = pstar[i][0] +  wamDyn->LinkCOGLoc[i][0];
      hold2[1] = pstar[i][1] +  wamDyn->LinkCOGLoc[i][1];
      hold2[2] = pstar[i][2] +  wamDyn->LinkCOGLoc[i][2];

      cross(hold1, hold2, Fs[i]);
      
      

      // R * (nn + cross(R' * pstar, f))

      mtrx_mult(hold2, tR, 3, 3, pstar[i], 3, 1);
 
      cross(hold3, hold2, f);

      hold3[0] = hold3[0] + nn[0];
      hold3[1] = hold3[1] + nn[1];
      hold3[2] = hold3[2] + nn[2];

      mtrx_mult(hold2, R, 3, 3, hold3, 3, 1);
      
      nn[0] = hold2[0] + hold1[0];
      nn[1] = hold2[1] + hold1[1];
      nn[2] = hold2[2] + hold1[2];

      // f = R * f + Fs(i)
      
      mtrx_mult(hold1, R, 3, 3, f, 3, 1);
      f[0] = hold1[0] + Fs[i][0];
      f[1] = hold1[1] + Fs[i][1];
      f[2] = hold1[2] + Fs[i][2];
      

      // torque = nn' * (R' * z0) + frictional stuff {ignored}


      mtrx_mult(hold1, transRotMatrices[i], 3, 3, z0, 3,1);

      // note we transpose nn just to be redefining dims - since a vector

      mtrx_mult(&torques->q[i], nn, 1, 3, hold1, 3, 1);
      


    }



}


