#include <syslog.h>
#include <math.h>
#include "btdriver.h"
#include "Haptics.h"

enum { NONE, TOP, BOTTOM, LEFT, RIGHT, FRONT, BACK };

extern WAMdat wam; /* need to use ID to do this right or use objects*/




static int teIndex;
static int feIndex;
static int obIndex;


static int numForceEffectors;
static int numHapticObjects;

static ForceEffector forceEffectors[MAX_FORCE_EFFECTORS];
static OBJECT *hapticObjects[MAX_HAPTIC_OBJECTS];


static double lastPos[3]; /* for velocity calc */



void computeWAMLoc() /* Based on global wam */
{
  WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &wam.wamTipLoc);

}

int InitHaptics(WAMdat *wam)
{
  numForceEffectors = 0;
  
  teIndex = registerTorqueEffector(effectGandFComp);

  feIndex = registerForceEffector(hapticObjectInteractor);

}

/* REDO THIS TO WORK WITH OBJECTS */
/*void deleteClosePlane(WAMLoc *tipLoc)
  {

  int i;
  double closestSoFar = MAXFLOAT;
  int closestIndex;
  double thisVal;
  
  #define CLOSENESS_THRESHOLD 0.06 

  for(i=0;i<MAX_HAPTIC_OBJECTS;i++)
  {
  if(hapticObjectIndices[i] != -1)
  {
	  
  if((thisVal = distanceFromPlane(tipLoc, &hapticPlanes[i])) < closestSoFar)
  {
  closestSoFar = thisVal;
  closestIndex = i;
  }
	  

  }

  }
  
  if(closestSoFar < CLOSENESS_THRESHOLD)
  {

  removeHapticObject(hapticObjectIndices[closestIndex]);
  hapticObjectIndices[closestIndex] = -1;
      

  }

  }
*/

int hapticObjectInteractor(WAMdat *wd, WAMLoc *tipforce, void *data)
{

  static int i;

  for(i=0;i<numHapticObjects;i++)
    {
      if(hapticObjects[i] != NULL)
	{

	  (*(hapticObjects[i]->interact))(&wd->wamTipLoc, tipforce, i);  

	}

    }
}
int addHapticObject(OBJECT *object)
{
  int i;

  for(i=0;i<numHapticObjects;i++)
    {
      if((i < numHapticObjects) && (hapticObjects[i] == NULL))
	{
	  hapticObjects[i] = object;

	  return(i);


	}

    }
  hapticObjects[numHapticObjects] = object;
  
  return(numHapticObjects++);
  
}


void removeHapticObject(int index)
{
  hapticObjects[index] = NULL;
}




int registerForceEffector(ForceEffector thisForceEffector)
{

  int i;

  for(i=0;i<numForceEffectors;i++)
    {
      if((i < numForceEffectors) && (forceEffectors[i] == NULL))
	{
	  forceEffectors[i] = thisForceEffector;
	  return(numForceEffectors);


	}

    }
  forceEffectors[numForceEffectors] = thisForceEffector;
  
  numForceEffectors++;
  return(numForceEffectors);
  
}


void deregisterForceEffector(int index)
{
  forceEffectors[index] = NULL;
}


/* performs gravity compensation calculations, with all haptic forces */
/* also updates tip location */
int effectGandFComp(wam_struct * thisWAM, void *data)
{

  static int cnt;
  static int idx;
  static int i;
  static wam_vector          Gtrq;
  static wam_vector          Gvel;
  static wam_vector          Gacc;
  static WAMLoc          tipforce; // force applied to tip
  static WAMLoc          wamloc; // actual position of WAM
  static int countdown = 50;

  
  
  zero_wam_vector(&Gtrq);
  zero_wam_vector(&Gvel);
  /* THESE NEED TO BE SET */

  zero_wam_vector(&Gacc);

  tipforce.pos[0] = 0.0;
  tipforce.pos[1] = 0.0;
  tipforce.pos[2] = 0.0;
  tipforce.angle[0] = 0.0;
  tipforce.angle[1] = 0.0;
  tipforce.angle[2] = 0.0;

 
  /* tool frame - immobile */
  thisWAM->Jpos.q[thisWAM->num_actuators+1] = 0.0; /* +1 is masstest hack */
    
  /* find where we are in space for use in force computation */
  /* this wam needs to be connected to the DriverData struct */
  /* by id or something - not a global wam. later */

  WAMFWKin(&wam.wamKin, &wam.wamDriverData->Jpos, &wam.wamTipLoc); //moved to WAMControlThread

  /*  if(countdown == 0 )
      {
      countdown = 5;
      }
      countdown--;*/
      

  for(i=0;i<numForceEffectors;i++)
    {
      if(forceEffectors[i] != NULL)
	{
	  (*forceEffectors[i])(&wam, &tipforce,i);
	}
    }


  GravityAndTipForceTorques(&wam.wamDyn, &wam.wamKin, &(thisWAM->Jpos), &(Gvel), &(Gacc), &(tipforce), &Gtrq, thisWAM->Gcomp);
  
  Gtrq.q[0] = -1.0 * Gtrq.q[0];
  Gtrq.q[1] = -1.0 * Gtrq.q[1];
  Gtrq.q[2] = -1.0 * Gtrq.q[2];
  Gtrq.q[3] = -1.0 * Gtrq.q[3];
  
  Gtrq.q[4] = -1.0 * Gtrq.q[4];
  Gtrq.q[5] = -1.0 * Gtrq.q[5];
  Gtrq.q[6] = -1.0 * Gtrq.q[6];
  
  for (cnt = 0; cnt < thisWAM->num_actuators; cnt++)
    {
      idx = thisWAM->motor_position[cnt];
      thisWAM->Jtrq.q[idx] += Gtrq.q[idx];
    }

}



int getGcomp(wam_struct *WAM)
{
  return WAM->Gcomp;
}
/** Enable or disable the gravity compensation calculation
 */
void setGcomp(wam_struct *WAM, int onoff)
{
  WAM->Gcomp = onoff;

}
/** Toggle the gravity compensation calculation
 */
void toggleGcomp(wam_struct *WAM)
{
  WAM->Gcomp = !WAM->Gcomp;

}





void createBlock(OBJECT *inputObj, BLOCK *inputBlock, double *c1, double *c2, double visc)
{
 
  inputObj->viscosity = visc;
  inputObj->interact = blockInteract;
  inputObj->data = (void *) inputBlock;

  if(c1[0] > c2[0]){
    
    inputBlock->xPlaneNeg = c2[0];
    inputBlock->xPlanePos = c1[0];
  } else {
    inputBlock->xPlanePos = c2[0];
    inputBlock->xPlaneNeg = c1[0];
  }


  if(c1[1] > c2[1]){
    
    inputBlock->yPlaneNeg = c2[1];
    inputBlock->yPlanePos = c1[1];

  } else {
    inputBlock->yPlanePos = c2[1];
    inputBlock->yPlaneNeg = c1[1];
  }

  if(c1[2] > c2[2]){
    
    inputBlock->zPlaneNeg = c2[2];
    inputBlock->zPlanePos = c1[2];
  } else {
    inputBlock->zPlanePos = c2[2];
    inputBlock->zPlaneNeg = c1[2];
  }

}


void createContainer(OBJECT *inputObj, CONTAINER *inputBlock, double *c1, double *c2)
{
 
  inputObj->interact = containerInteract;
  inputObj->data = (void *) inputBlock;

  if(c1[0] > c2[0]){
    
    inputBlock->xPlaneNeg = c2[0];
    inputBlock->xPlanePos = c1[0];
  } else {
    inputBlock->xPlanePos = c2[0];
    inputBlock->xPlaneNeg = c1[0];
  }


  if(c1[1] > c2[1]){
    
    inputBlock->yPlaneNeg = c2[1];
    inputBlock->yPlanePos = c1[1];

  } else {
    inputBlock->yPlanePos = c2[1];
    inputBlock->yPlaneNeg = c1[1];
  }

  if(c1[2] > c2[2]){
    
    inputBlock->zPlaneNeg = c2[2];
    inputBlock->zPlanePos = c1[2];
  } else {
    inputBlock->zPlanePos = c2[2];
    inputBlock->zPlaneNeg = c1[2];
  }

}

void createField(OBJECT *inputObj, FIELD *inputField, double *c1, double *c2)
{

  double a,b,c, dist;
  inputObj->interact = (void *) fieldInteract;
  inputObj->data = (void *) inputField;
  a = c2[0] - c1[0];
  b = c2[1] - c1[1];
  c = c2[2] - c1[2];

  dist = sqrt(a*a + b*b + c*c);
  
  inputField->F[0] = -30.0 * a/dist;
  inputField->F[1] = -30.0 * b/dist;
  inputField->F[2] =  -30.0 * c/dist;

 
}



void createSphere(OBJECT *inputObj, SPHERE *inputSphere, double *c1, double *c2)
{

  double a,b,c;
  inputObj->interact = (void *) sphereInteract;
  inputObj->data = (void *) inputSphere;
  
  a = c1[0] - c2[0];
  b = c1[1] - c2[1];
  c = c1[2] - c2[2];
  inputSphere->center.pos[0] = c1[0];
  inputSphere->center.pos[1] = c1[1];
  inputSphere->center.pos[2] = c1[2];

  inputSphere->radius = sqrt(a * a + b * b + c * c);


}

void fieldInteract(WAMLoc *cursorpos, WAMLoc *tipforce, int id)
{

  static int i;
  static FIELD *field;
  field = (FIELD *) (hapticObjects[id]->data);

  for(i=0;i<3;i++)
    {
      tipforce->pos[i] = field->F[i];
    }


}
void sphereInteract(WAMLoc *cursorpos, WAMLoc *tipforce, int id)
{

  static double dist, dX, dY, dZ;
  static SPHERE *sphere;

  sphere = (SPHERE *) (hapticObjects[id]->data);

  dX = cursorpos->pos[0] - sphere->center.pos[0];
  dY = cursorpos->pos[1] - sphere->center.pos[1];
  dZ = cursorpos->pos[2] - sphere->center.pos[2];

  dist = sqrt(dX * dX + dY * dY + dZ * dZ);

  if(dist < sphere->radius)
    {
      /* normalize then scale */
      tipforce->pos[0] = dX / dist * -30.0;
      tipforce->pos[1] = dY / dist * -30.0;
      tipforce->pos[2]  = dZ / dist * -30.0;

    }

}



/* this keeps you within a block */
int  containerInteract(WAMLoc *curpos, WAMLoc *tipforce, int id)
{

  static int whichside = NONE;

  static CONTAINER *thisblock;

  static double dist;
  static double temp;

  thisblock = (CONTAINER *) (hapticObjects[id]->data);

  tipforce->pos[0] = tipforce->pos[1] = tipforce->pos[2] = 0.0 ;
  tipforce->angle[0] = tipforce->angle[1] = tipforce->angle[3] = 0.0;
  //    if(whichside == NONE)
  //      {
  dist = 0.0;
  if(curpos->pos[0] < thisblock->xPlaneNeg)
    {
      whichside = LEFT;
      dist = thisblock->xPlaneNeg - curpos->pos[0];
    }
  if(curpos->pos[0] > thisblock->xPlanePos)
    {
      
      temp = curpos->pos[0] - thisblock->xPlanePos;
      
      if(temp > dist)
	{
	  whichside = RIGHT;
	  dist = temp;
	}
    }      
  if(curpos->pos[1] < thisblock->yPlaneNeg)
    {
      
      temp = thisblock->yPlaneNeg - curpos->pos[1];
      
      if(temp > dist)
	{
	  whichside = BACK;
	  dist = temp;
	}
    }
  if(curpos->pos[1] > thisblock->yPlanePos)
    {
      
      temp = curpos->pos[1] - thisblock->yPlanePos;
      
      if(temp > dist)
	{

	  whichside = FRONT;
	  dist = temp;
	}
    }
  
  if(curpos->pos[2] < thisblock->zPlaneNeg)
    {
      temp = thisblock->zPlaneNeg - curpos->pos[2];

      if(temp > dist)
	{
	  whichside = BOTTOM;
	  dist = temp;
	}
    }
  if(curpos->pos[2] > thisblock->zPlanePos)
    {
      temp = curpos->pos[2] - thisblock->zPlanePos;
      if(temp > dist)
	{
	  whichside = TOP;
	  dist = temp;
	}
    }
  /*    } */
  if(curpos->pos[0] > thisblock->xPlaneNeg && curpos->pos[0] < thisblock->xPlanePos && curpos->pos[1] > thisblock->yPlaneNeg && curpos->pos[1] < thisblock->yPlanePos && curpos->pos[2] > thisblock->zPlaneNeg && curpos->pos[2] < thisblock->zPlanePos)
    {
      whichside = NONE;
    }
      /* }*/

  if(whichside == LEFT)
    {
      tipforce->pos[0] = -30.0;
    }
  else if(whichside == RIGHT)
    {
      tipforce->pos[0] = 30.0;
    }
  else if(whichside == BACK)
    {
      tipforce->pos[1] = -30.0;
    }
  else if(whichside == FRONT)
    {
      tipforce->pos[1] = 30.0;
    }
  else if(whichside == BOTTOM)
    {
      tipforce->pos[2] = -30.0;
    }
  else if(whichside == TOP)
    {
      tipforce->pos[2] = 30.0;
    }
       

}


/* this keeps you out of a block */
int  blockInteract(WAMLoc *curpos, WAMLoc *tipforce, int id)
{

  static BLOCK *thisblock;
  static double dist;
  static double temp;
  static int which;
  thisblock = (BLOCK *) (hapticObjects[id]->data);
  

  //  tipforce->pos[0] = tipforce->pos[1] = tipforce->pos[2] = 0.0 ;
  tipforce->angle[0] = tipforce->angle[1] = tipforce->angle[3] = 0.0;

  if(curpos->pos[0] > thisblock->xPlaneNeg && curpos->pos[0] < thisblock->xPlanePos && curpos->pos[1] > thisblock->yPlaneNeg && curpos->pos[1] < thisblock->yPlanePos && curpos->pos[2] > thisblock->zPlaneNeg && curpos->pos[2] < thisblock->zPlanePos)
    {
      dist = curpos->pos[0] - thisblock->xPlaneNeg;
      tipforce->pos[0] = 30.0;
      
      temp = thisblock->xPlanePos - curpos->pos[0];
      if(temp < dist)
	{
	  tipforce->pos[0] = -30.0;
	  dist = temp;
	}

      temp = curpos->pos[1] - thisblock->yPlaneNeg;
      if(temp < dist)
	{
	  tipforce->pos[1] = 30.0;
	  tipforce->pos[0] = 0.0;
	  dist = temp;
	}
	  
      temp = thisblock->yPlanePos - curpos->pos[1];
      if(temp < dist)
	{
	  tipforce->pos[1] = -30.0;
	  tipforce->pos[0] = 0.0;
	  dist = temp;
	}
      temp = curpos->pos[2] - thisblock->zPlaneNeg;
      if(temp < dist)
	{
	  tipforce->pos[2] = 30.0;
	  tipforce->pos[0] = 0.0;
	  tipforce->pos[1] = 0.0;
	  dist = temp;
	}

      temp = thisblock->zPlanePos - curpos->pos[2];
      if(temp < dist)
	{
	  tipforce->pos[2] = -30.0;
	  tipforce->pos[1] = 0.0;
	  tipforce->pos[0] = 0.0;
	  dist = temp;
	}


    }
}

int resistanceLineInteract(WAMLoc *wloc, WAMLoc *tipforce, int id)
{

  static double lD[3];
  static double dist[3];
  static double mag;
  static double sum1, sum2, sum3;
  static double normal[3];
  static double curvel[3];
  static double distance;
  static RESISTANCELINE *thisLine;
  static int countdown = 0;

  thisLine = (RESISTANCELINE *) (hapticObjects[id]->data);

  dist[0] = thisLine->e1.pos[0] - wloc->pos[0];
  dist[1] = thisLine->e1.pos[1] - wloc->pos[1];
  dist[2] = thisLine->e1.pos[2] - wloc->pos[2];

  lD[0] = thisLine->e2.pos[0] - thisLine->e1.pos[0];
  lD[1] = thisLine->e2.pos[1] - thisLine->e1.pos[1];
  lD[2] = thisLine->e2.pos[2] - thisLine->e1.pos[2];

  mag = sqrt(lD[0] * lD[0] + lD[1] * lD[1] + lD[2] * lD[2]);

  lD[0] = lD[0] / mag;
  lD[1] = lD[1] / mag;
  lD[2] = lD[2] / mag;
  
  sum1 = dist[1] * lD[2] - dist[2] * lD[1];
  sum2 = dist[2] * lD[0] - dist[0] * lD[2];
  sum3 = dist[0] * lD[1] - dist[1] * lD[0];

  
  distance = sqrt(sum1 * sum1 + sum2 * sum2 + sum3 * sum3);
  /* denominator magnitude is 1 b/c lD is normalized *

  /* find perpendicular to line through point */
  /* sum123 hold first cross product */
  
  normal[0] = sum2 * lD[2] - sum3 * lD[1];
  normal[1] = sum3 * lD[0] - sum1 * lD[2];
  normal[2] = sum1 * lD[1] - sum2 * lD[0];


  mag = sqrt(normal[0] * normal[0] + normal[1] * normal[1] + normal[2] * normal[2]);

  /* bring towards path */
  if(mag > 0.05)
    {
      normal[0] = normal[0] / mag;
      normal[1] = normal[1] / mag;
      normal[2] = normal[2] / mag;
      
      
      tipforce->pos[0] = normal[0] * 25 ;//* thisLine->force;// * distance;
	tipforce->pos[1] = normal[1] * 25; //* thisLine->force;// * distance;
	tipforce->pos[2] = normal[2] * 25 ;//* thisLine->force;// * distance;
      
      
      /*      if(countdown == 0)
	{

//	  syslog(LOG_ERR, "tf %f %f %f %f", tipforce->pos[0], tipforce->pos[1], tipforce->pos[2], mag);
	  
	  countdown=100;

	}
      countdown--;
      */


	
    }
  else /* resist motion */
    {


      curvel[0] = wloc->pos[0] - lastPos[0];
      curvel[1] = wloc->pos[1] - lastPos[1];
      curvel[2] = wloc->pos[2] - lastPos[2];
      
      mag = sqrt(curvel[0] * curvel[0] + curvel[1] * curvel[1] + curvel[2] * curvel[2]);
      //      syslog(LOG_ERR,"in path %f", mag);
      
      if(mag > 0.0001) /* if is really moving */
	{
	  curvel[0] = curvel[0] / mag; /* normalize */
	  curvel[1] = curvel[1] / mag;
	  curvel[2] = curvel[2] / mag;
	  


	  tipforce->pos[0] += curvel[0] * 20.0;
	  tipforce->pos[1] += curvel[1] * 20.0;
	  tipforce->pos[2] += curvel[2] * 20.0;

	  //	  syslog(LOG_ERR, "resisting %f", tipforce->pos[2]);
	}
    }

  lastPos[0] = wloc->pos[0];
  lastPos[1] = wloc->pos[1];
  lastPos[2] = wloc->pos[2];



}

void createResistanceLine(OBJECT *inputObj, RESISTANCELINE *inputLine, WAMLoc *e1, WAMLoc *e2, double force)
{
  inputObj->interact = resistanceLineInteract;
  inputObj->data = (void *) inputLine;

  inputLine->e1.pos[0] = e1->pos[0];
  inputLine->e1.pos[1] = e1->pos[1];
  inputLine->e1.pos[2] = e1->pos[2];


  inputLine->e2.pos[0] = e2->pos[0];
  inputLine->e2.pos[1] = e2->pos[1];
  inputLine->e2.pos[2] = e2->pos[2];

  inputLine->force = force;



}


int createPlane(OBJECT *inputObj, PLANE *inputPlane, double *pt1, double *pt2, double *pt3)
{
  double rel2[3];
  double rel3[3];
  double WAMBase[] = {0.0, 0.0, 0.0}; /*(right?)*/
  double dotRes;
  double divisor;

  inputObj->interact = planeInteract;
  inputObj->data = (void *) inputPlane;

  rel2[0] = pt2[0] - pt1[0];
  rel2[1] = pt2[1] - pt1[1];
  rel2[2] = pt2[2] - pt1[2];

  rel3[0] = pt3[0] - pt1[0];
  rel3[1] = pt3[1] - pt1[1];
  rel3[2] = pt3[2] - pt1[2];
   
  cross(inputPlane->normal, rel2, rel3);
  divisor = sqrt(inputPlane->normal[0] * inputPlane->normal[0] + inputPlane->normal[1] * inputPlane->normal[1] + inputPlane->normal[2] * inputPlane->normal[2]);


  inputPlane->normal[0] /= divisor;
  inputPlane->normal[1] /= divisor;
  inputPlane->normal[2] /= divisor;

  dot(&inputPlane->distance, inputPlane->normal, pt1);

  inputPlane->distance = -inputPlane->distance;
  
  dot(&dotRes, inputPlane->normal, pt3);
  dotRes += inputPlane->distance;
  

  if(dotRes > 0.0)
    inputPlane->WAMSideSign = 1; /* WAM base is on same side as normal */
  else /* assume plane does not go through WAM */
    inputPlane->WAMSideSign = -1; /* WAM base opposite to normal */

}

double distanceFromPlane(WAMLoc *loc, PLANE *plane)
{
  double dist;

  //  dot(&dist, plane->normal, loc->pos);
  //dist += plane->distance;
  // equivalent


  dist = plane->normal[0] * loc->pos[0] + plane->normal[1] * loc->pos[1] + plane->normal[2] * loc->pos[2] + plane->distance;

  dist = fabs(dist);
  


  return(dist);
}


int planeInteract(WAMLoc *wd, WAMLoc *tipforce, int id)
{

  static double tipSideDist;
  static PLANE *thisPlane;
  static int tipSide;

  thisPlane = (PLANE *) (hapticObjects[id]->data);

  //  dot(&tipSideDist, thisPlane->normal, wd->pos);
  //tipSideDist += thisPlane->distance;
  // normal, dist is params A B C D
  

  tipSideDist = thisPlane->normal[0] * wd->pos[0] + thisPlane->normal[1] * wd->pos[1] + thisPlane->normal[2] * wd->pos[2] + thisPlane->distance;
  

  //  syslog(LOG_ERR,"pos %f %f %f", wd->pos[0], wd->pos[1], wd->pos[2]);

  if(tipSideDist > 0.0)
    tipSide = 1;
  else 
    tipSide = -1;


  // take out += business, later

  if((tipSide == 1)) //&& (thisPlane->WAMSideSign == -1))
    {
      /* push towards base, opposite normal (is this right?) */
     
      tipforce->pos[0] += thisPlane->normal[0] * 30.0; /* adjust this 10.0 */
      tipforce->pos[1] += thisPlane->normal[1] * 30.0;
      tipforce->pos[2] += thisPlane->normal[2] * 30.0;


    }
  /*  else if((tipSide == -1) && (thisPlane->WAMSideSign == 1))
    {
      tipforce->pos[0] += thisPlane->normal[0] * -30.0;
      tipforce->pos[1] += thisPlane->normal[1] * -30.0;
      tipforce->pos[2] += thisPlane->normal[2] * -30.0;

    }
  else
    {
      //      tipforce->pos[0] += 0.0;
      //      tipforce->pos[1] += 0.0;
      //      tipforce->pos[2] += 0.0;

    }*/


}
