#ifndef __HAPTICS_H__
#define __HAPTICS_H__


#include "Kinematics.h"
#include "Dynamics.h"
#include "btdriver.h"


#define MAX_FORCE_EFFECTORS 16
#define MAX_HAPTIC_OBJECTS 256

typedef struct {
  
  wam_struct *wamDriverData; //temporarily a pointer
  WAMLoc wamTipLoc;
  WAMKinematics wamKin;
  WAMDynamics wamDyn;


} WAMdat;


typedef int (*ForceEffector)(WAMdat *wd, WAMLoc *resultForce, int id);

typedef int (*ObjectInteract)(WAMLoc *wamTipLoc, WAMLoc *resultForce, int id);


typedef struct {

  int type;
  ObjectInteract interact;
  double viscosity;
  void *data;

} OBJECT;




typedef struct resistanceline {
  
  WAMLoc e1;
  WAMLoc e2;
  double force;

} RESISTANCELINE;


typedef struct {
  double xPlaneNeg;
  double xPlanePos;
  double yPlaneNeg;
  double yPlanePos;
  double zPlaneNeg;
  double zPlanePos;
} BLOCK;

typedef struct {
  double xPlaneNeg;
  double xPlanePos;
  double yPlaneNeg;
  double yPlanePos;
  double zPlaneNeg;
  double zPlanePos;
} CONTAINER;




typedef struct {
  double normal[3];
  double distance;
  int WAMSideSign;
} PLANE;

typedef struct {

  WAMLoc center;
  double radius;

} SPHERE;
 
typedef struct {
  
  double F[3];
} FIELD;


int effectGandFComp(wam_struct *thisWAM, void *data);
int hapticObjectInteractor(WAMdat *wd, WAMLoc *tipforce, void *data);
int blockInteract(WAMLoc *curpos, WAMLoc *tipforce, int id);
int containerInteract(WAMLoc *curpos, WAMLoc *tipforce, int id);
void sphereInteract(WAMLoc *cursorpos, WAMLoc *tipforce, int id);
int resistanceLineInteract(WAMLoc *wd, WAMLoc *tipforce, int id);
void fieldInteract(WAMLoc *curpos, WAMLoc *tipforce, int id);

int planeInteract(WAMLoc *wd, WAMLoc *tipforce, int id);
void deleteClosePlane(WAMLoc *tipLoc);



#endif
