#include <math.h>
#include <syslog.h>

#include "Kinematics.h"
#include "MatrixMath.h"
#include "btwam.h"
#include "newlibbt.h"

void InitWAMKinematics(WAMKinematics *wamKin, int dof)
{

    int i;

    InitKinematics(wamKin, dof);

    /*  Alpha = Rotation about x from Frame-1 to Frame
        D = Distance along z-1 axis from x-1 to x
        A = Distance along x axis from z-1 to z
     */
     
    wamKin->Alphas[0]   = -PI/2.0;
    wamKin->Ds[0]       = 0.0;
    wamKin->As[0]       = 0.0;
    
    wamKin->Alphas[1]   = PI/2.0;
    wamKin->Ds[1]       = 0.0;
    wamKin->As[1]       = 0.0;
    
    wamKin->Alphas[2]   = -PI/2.0;
    wamKin->Ds[2]       = 0.558;
    wamKin->As[2]       = 0.0476;
#if 0    
    // Blank Link
    wamKin->Alphas[3]   = PI/2.0;
    wamKin->Ds[3]       = 0.0;
    wamKin->As[3]       = -0.0476;
    
    wamKin->Alphas[4]   = 0.0;
    wamKin->Ds[4]       = 0.3574;
    wamKin->As[4]       = 0.0;
#endif
#if 1
    // Wrist
    wamKin->Alphas[3]   = PI/2.0;
    wamKin->Ds[3]       = 0.0;
    wamKin->As[3]       = -0.0476;
    
    wamKin->Alphas[4]   = -PI/2.0;
    wamKin->Ds[4]       = 0.31;
    wamKin->As[4]       = 0.0;
    
    wamKin->Alphas[5]   = PI/2.0;
    wamKin->Ds[5]       = 0.0;
    wamKin->As[5]       = 0.0;
    
    wamKin->Alphas[6]   = 0.0;
    wamKin->Ds[6]       = 0.06;
    wamKin->As[6]       = 0.0;
#endif
    for(i=0;i<(wamKin->dof);i++)
    {
        initLinkTransform(wamKin->Alphas[i], wamKin->As[i], 0.0, wamKin->Ds[i], wamKin->transforms[i]);

    }

}

#if 0
void InitWAMKinematics(WAMKinematics *wamKin, int dof){

  int i;

  InitKinematics(wamKin, dof);

  wamKin->Alphas[0] = -PI/2.0;
  wamKin->Alphas[1] = PI/2.0;
  wamKin->Alphas[2] = -PI/2.0;
  wamKin->Alphas[3] = PI/2.0;
  wamKin->Ds[2] = 0.558;
  wamKin->As[2] = 0.0476;
  wamKin->As[3] = -0.0476;

#ifdef MASSTEST
  wamKin->Ds[4] = 0.0;//0.3574;
#endif


  for(i=0;i<(wamKin->dof);i++)
    {
      initLinkTransform(wamKin->Alphas[i], wamKin->As[i], 0.0, wamKin->Ds[i], wamKin->transforms[i]);

    }

}
#endif

void InitKinematics(WAMKinematics *wamKin, int dof)
{
  int i;

  wamKin->dof = dof;

  memset(wamKin->As, 0x0, dof * sizeof(double));
  memset(wamKin->Ds, 0x0, dof * sizeof(double));
  memset(wamKin->Alphas,0x0, dof * sizeof(double));
  memset(wamKin->transforms, 0x0, dof * 16 * sizeof(double));



  for(i=0;i<dof;i++)
    {
      wamKin->transforms[i][15] = 1.0;

    }

}

void initLinkTransform(double alpha, double a, double theta, double d, double *transform)
{
  double cosTheta, sinTheta, cosAlpha, sinAlpha;
  
  cosTheta = cos(theta);
  sinTheta = sin(theta);
  cosAlpha = cos(alpha);
  sinAlpha = sin(alpha);

  transform[0] = cosTheta;
  transform[1] = -1 * sinTheta * cosAlpha;
  transform[2] = sinTheta * sinAlpha;
  transform[3] = a * cosTheta;
  transform[4] = sinTheta;
  transform[5] = cosTheta * cosAlpha;
  transform[6] = -1 * cosTheta * sinAlpha;
  transform[7] = a * sinTheta;
  /*  transform[8] keep at 0*/
  transform[9] = sinAlpha;
  transform[10] = cosAlpha;
  transform[11] = d;

}



void findDHtransform_deltaTheta(WAMKinematics *wamKin, wam_vector *inputAngles)
{
  static int i;
  static double cosAlpha, sinAlpha, a, sinTheta, cosTheta;


  for(i=0;i<wamKin->dof;i++)
    {

      cosAlpha = cos(wamKin->Alphas[i]);
      sinAlpha = sin(wamKin->Alphas[i]);
      a = wamKin->As[i];
      sinTheta = sin(inputAngles->q[i]);
      cosTheta = cos(inputAngles->q[i]);

      wamKin->transforms[i][0] = cosTheta;
      wamKin->transforms[i][1] = -1 * sinTheta * cosAlpha;
      wamKin->transforms[i][2] = sinTheta * sinAlpha;
      wamKin->transforms[i][3] = a * cosTheta;
      wamKin->transforms[i][4] = sinTheta;
      wamKin->transforms[i][5] = cosTheta * cosAlpha;
      wamKin->transforms[i][6] = -1 * cosTheta * sinAlpha;
      wamKin->transforms[i][7] = a * sinTheta;     
      


    }

}

void WAMFWKin(WAMKinematics *wamKin,  wam_vector *inputAngles, WAMLoc *outputLoc)
{

  int i;
  double dest[16];
  double holder;
  double sBeta;
  double U[16] = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };

  findDHtransform_deltaTheta(wamKin, inputAngles);

  for(i=0;i<(wamKin->dof);i++)
    {
       mtrx_mult(dest, U,4,4,wamKin->transforms[i],4,4);
      mtrx_cpy(dest, U, 4,4);
    }


  outputLoc->pos[0] = dest[3];
  outputLoc->pos[1] = dest[7];
  outputLoc->pos[2] = dest[11];



  
  /* ZYZ euler angles - algorithm from Craig */
  holder = U[8] * U[8] + U[9] * U[9];
  holder = sqrt(holder);
  outputLoc->angle[0] = atan2(holder, U[10]);
  sBeta = sin(outputLoc->angle[0]);

  if(sBeta == 0.0)
    {
      /* degenerate solution. */
      
    }
  else
    {

      outputLoc->angle[1] = atan2(U[6] / sBeta, U[2] / sBeta);
      outputLoc->angle[2] = atan2(U[9] / sBeta, -1 * U[10] / sBeta);
				 
    }

}


void FindTransformedPoint(WAMKinematics *wamKin,  wam_vector *inputAngles, WAMLoc *inputLoc, WAMLoc *outputLoc)
{

  int i;
  double dest[16];
  double rot_mtrx[9];
  double i_rot_mtrx[9];
  double holder;
  double sBeta;
  double U[16] = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };
  static int count=100;

  //U[3] = inputLoc->pos[0];
      //U[7] = inputLoc->pos[1];
     //    U[11] = inputLoc->pos[2];
  
  findDHtransform_deltaTheta(wamKin, inputAngles);

  for(i=0;i<(wamKin->dof);i++)
    {
      mtrx_mult(dest, U,4,4,wamKin->transforms[i],4,4);
      // mtrx_mult(dest, wamKin->transforms[i],4,4,U,4,4);
      mtrx_cpy(dest, U, 4,4);
    }
  count--;
  if(count==0)
    {
      count=100;
      //      mtrx_log(U, 4, 4, "dest");
    }
  //    outputLoc->pos[0] = dest[3];
  //   outputLoc->pos[1] = dest[7];
  //   outputLoc->pos[2] = dest[11];
   
  rot_mtrx[0] = dest[0];
  rot_mtrx[1] = dest[1];
  rot_mtrx[2] = dest[2];
  rot_mtrx[3] = dest[4];
  rot_mtrx[4] = dest[5];
  rot_mtrx[5] = dest[6];
  rot_mtrx[6] = dest[8];
  rot_mtrx[7] = dest[9];
  rot_mtrx[8] = dest[10];

  mtrx_transpose(rot_mtrx, i_rot_mtrx, 3,3);

  //    outputLoc->pos[0] = dest[0] * inputLoc->pos[0] + dest[1] * inputLoc->pos[1] + dest[2] * inputLoc->pos[2];
  //    outputLoc->pos[1] = dest[4] * inputLoc->pos[0] + dest[5] * inputLoc->pos[1] + dest[6] * inputLoc->pos[2];
  //    outputLoc->pos[2] = dest[8] * inputLoc->pos[0] + dest[9] * inputLoc->pos[1] + dest[10] * inputLoc->pos[2];
  // could just multiply here, whoops
    outputLoc->pos[0] = i_rot_mtrx[0] * inputLoc->pos[0] + i_rot_mtrx[1] * inputLoc->pos[1] + i_rot_mtrx[2] * inputLoc->pos[2];
       outputLoc->pos[1] = i_rot_mtrx[3] * inputLoc->pos[0] + i_rot_mtrx[4] * inputLoc->pos[1] + i_rot_mtrx[5] * inputLoc->pos[2];
       outputLoc->pos[2] = i_rot_mtrx[6] * inputLoc->pos[0] + i_rot_mtrx[7] * inputLoc->pos[1] + i_rot_mtrx[8] * inputLoc->pos[2];


  // !!!!!!!!! REMOVED ANGLE INFO
  
  /* ZYZ euler angles - algorithgm from Craig */
  //  holder = U[8] * U[8] + U[9] * U[9];
  //  holder = sqrt(holder);
  //  outputLoc->angle[0] = atan2(holder, U[10]);
  //  sBeta = sin(outputLoc->angle[0]);

  if(sBeta == 0.0)
    {
      /* degenerate solution. */
      
    }
  else
    {

      //      outputLoc->angle[1] = atan2(U[6] / sBeta, U[2] / sBeta);
      //      outputLoc->angle[2] = atan2(U[9] / sBeta, -1 * U[10] / sBeta);
				 
    }

}




