#include <syslog.h>
#include "newlibbt.h"
#include "Kinematics.h"


void findJacobian(WAMKinematics *wamKin, wam_vector *input_angles, double *Jacobian)
{
  /* Jacobian is 6 * DOF matrix */
  /* Transforms is MAXDOF * 16 matrix where DOF in the first dimension are used */
  /* inputAngles is a vector of number of joints */

  int i;
  double *thisTransform;


  double I[16] = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };
  double U[16] = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 };
  double dest[16];
  double On[3]; /* vec */
  double Ja[3]; /* vec of angles for Jacobian */
  double d[3]; /* vec for jacobian xyz */
  double Od[3];
  WAMLoc outputLoc;

  
  //findDHtransform_deltaTheta(inputAngles);

  WAMDHFWKin(input_angles,  &outputLoc); 
  On[0] = outputLoc.pos[0];
  On[1] = outputLoc.pos[1];
  On[2] = outputLoc.pos[2];

  syslog(LOG_ERR, "loc: %f %f %f\n", On[0], On[1], On[2]);
  //  memset(I, 0x0, 16);
  for(i=-1;i<(wamKin->dof-1);i++)
    {
      if(i==-1)
	thisTransform = I;
      else
      thisTransform = wamKin->transforms[i];
      
      mtrx_mult(dest, U, 4,4, thisTransform,4,4);
      mtrx_cpy(dest, U, 4, 4);
      
      *(Jacobian+i+3*wamKin->dof+1) = Ja[0] = U[2]; /* 4th row of jacobian, i'th col */
      *(Jacobian+i+4*wamKin->dof+1) = Ja[1] = U[6];
      *(Jacobian+i+5*wamKin->dof+1) = Ja[2] = U[10];
      
      Od[0] = On[0] - U[3];
      Od[1] = On[1] - U[7];
      Od[2] = On[2] - U[11];

      /* cross product */
      *(Jacobian+i+1) = Ja[1] * Od[2] - Ja[2] * Od[1];
      *(Jacobian+i+wamKin->dof+1) = Ja[2] * Od[0] - Ja[0] * Od[2];
      *(Jacobian+i+wamKin->dof*2+1) = Ja[0] * Od[1] - Ja[1] * Od[0];

    }

}

void StaticF2torque(WAMKinematics *wamKin, wam_vector *input_angles,  double *fVec, wam_vector *torques)
{
  /* fVec is 6x1 Force-moment composition of vectors */

  static double Jacobian[24], Jt[24];
  
  findJacobian(wamKin, input_angles, Jacobian);

  mtrx_transpose(Jacobian, Jt, 6, 4);

  //  mtrx_print(Jacobian, 6, 4, "jacobian");
  //  mtrx_print(Jt, 4, 6, "jt");

  mtrx_mult(torques->q, Jt, 4, 6, fVec, 6, 1);


}
