/* main.c */

#define toupper(c)      ( ((c >= 'a') && (c <= 'z')) ? c - ('a' - 'A') : c )
#define HOLE            (0x0000)
#if 0
#define SMOOTH_CT       (5)


#define REVOLUTIONS     (2)
#define STEP            (1)
#define TR_SAMPLE_TIME  (0.005)
#define WAIT_TIME       (0.0025)
#endif

#define TR_GAIN         (0.8)
#define GAIN_WINDOW     (1)
#define pi              (3.14159)

#define TEST_TR         (0)
#define COLLECT_TR      (1)

#include <stdio.h>
#include <syslog.h>
#include <sys/neutrino.h>
#include <sys/syspage.h>
#include <inttypes.h>
#include <malloc.h>
#include <math.h>
#include <semaphore.h>

#include "btcan.h"
#include "SimpleControl.h"

//extern  local_info_t    priv_data;
int     REVOLUTIONS_F, REVOLUTIONS_R, SMOOTH_CT;
int     STEP;
double  TR_SAMPLE_TIME;
double  WAIT_TIME;
double  kp, kd, ki;
int     compTorque;
extern sem_t   timer_semaphore;
extern int     shutdown_threads;

typedef struct
{
    int puckID;
    int motorSN;
    int encCtsPerRev;
    int mechOffset;
    int iOffsetB;
    int iOffsetC;
}motorInfo_t;

typedef struct{char s[20];}string20;

typedef struct
{
    motorInfo_t *m;
    CANdev_t    *c;
    PIDregulator *p;
}CThreadParams_t;

const    string20	statusTxt[]={
        {"STATUS_OFFLINE"},
    	{"STATUS_RESET"},
        {"STATUS_SHUTDOWN"},
        {"STATUS_INITIALIZE"},
        {"STATUS_ERR"},
        {"STATUS_READY"},
        {"STATUS_COMMUTATING"},
    };

void PuckControlThread(void *data);

void showMenu(void){
    printf("\nMENU");
    printf("\n--------");
    printf("\nE)numerate bus status");
    printf("\nM)easure CAN speed");
    printf("\nD)ownload firmware");
    printf("\nO)ld firmware download");
    printf("\nC)ollect TR data");
    printf("\nT)est TR data");
    
    printf("\n\nYour Choice: ");
}

void handleMenu(char c, CANdev_t *CANdev){
    char        fname[32];
    int         node;
    int         err;
    uint64_t    time1, time2;
    float       delta;
    long        reply;
    int         status[MAX_NODES];
    int         i;
    double      cps;
    int         newID;
    
    
    switch(c){
    case 'E':
        printf("\n\nCAN bus enumeration (status)\n");
        getBusStatus(CANdev, status);
        for(i = 0; i < MAX_NODES; i++)
        {
            if(i % 2)
                printf("\t\tNode %2d: %s", i, statusTxt[status[i]+1].s);
            else
                printf("\nNode %2d: %s", i, statusTxt[status[i]+1].s);
        }
        printf("\n");
        break;
        
    case 'M':
        printf("\n\nCAN Speed Test...\n");
        cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
        while(1)
        {
            time1 = ClockCycles();
            getProperty(CANdev, 25, STATUS, &reply);
            time2 = ClockCycles();
            delta = (double)(time2 - time1) / cps;
            printf("\r%f", delta); fflush(stdout);
        }
        
    case 'D':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Get new Node ID */
        if(node > 0)
        {
            printf("\nChange node ID to (enter 0 for no change): ");
            err = scanf("%d", &newID);
            if(err == EOF) /* No input data */
                newID = 0; /* Zero signifies "no change" */
        }
        
        /* Download the firmware */
        err = downloadFirmware(CANdev, node, newID, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'O':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Download the firmware */
        err = downloadFirmware_Old(CANdev, node, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'C':
        /* Collect TR Data */
        doTR(CANdev, COLLECT_TR);
        break;
        
    case 'T':
        doTR(CANdev, TEST_TR);
        break;
        
    default:
    
        break;
    }
}

int doTR(CANdev_t *CANdev, int doCollect)
{
    motorInfo_t motorInfo;
    int         err;
    FILE        *in;
    CThreadParams_t CThreadParams;
    PIDregulator pid;
    long        actPos;
    
    if((in = fopen("tr.info","r")) == NULL)
    {
        printf("\nFailed to open tr.info file!");
        return(1);
    }
    fscanf(in, "%*s : %d", &motorInfo.puckID);
    fscanf(in, "%*s : %d", &motorInfo.motorSN);
    fscanf(in, "%*s : %d", &motorInfo.encCtsPerRev);
    fscanf(in, "%*s : %d %*s : %d %*s : %d", &motorInfo.mechOffset, &motorInfo.iOffsetB, &motorInfo.iOffsetC);
    fscanf(in, "%*s : %d %*s : %d %*s : %d %*s : %d %*s : %lf %*s : %lf", &REVOLUTIONS_F, &REVOLUTIONS_R, &STEP, &SMOOTH_CT, &TR_SAMPLE_TIME, &WAIT_TIME);
    fscanf(in, "%*s : %lf %*s : %lf %*s : %lf", &kp, &kd, &ki);
    
    fclose(in);
    
    printf("CANID      : %d\n"
            "MOTOR_SN   : %d\n"
            "CTS_PER_REV: %d\n"
            "MECH_OFFSET: %d \n"
            "IOFFSET_B  : %d \n"
            "IOFFSET_C  : %d\n"
            "FWD_REVS   : %d \n"
            "REV_REVS   : %d \n"
            "STEP_SIZE  : %d \n"
            "SMOOTH_CT  : %d \n"
            "SAMP_TIME  : %0.3f \n"
            "WAIT_TIME  : %0.3f\n"
            "KP         : %0.3f \n"
            "KD         : %0.3f \n"
            "KI         : %0.3f\n", motorInfo.puckID, motorInfo.motorSN, motorInfo.encCtsPerRev,
            motorInfo.mechOffset, motorInfo.iOffsetB, motorInfo.iOffsetC,
            REVOLUTIONS_F, REVOLUTIONS_R, STEP, SMOOTH_CT, TR_SAMPLE_TIME, WAIT_TIME,
            kp, kd, ki);
   
    getchar();
    
#if 0
    /* Prompt for necessary info */
    printf("\nPlease enter puck ID: ");
    scanf("%d", &motorInfo.puckID);
    printf("\nPlease enter motor SN: ");
    scanf("%d", &motorInfo.motorSN);
    printf("\nPlease enter the number of encoder cts per revolution: ");
    scanf("%d", &motorInfo.encCtsPerRev);
    printf("\nPlease enter motor %d's offsets [mechOffset iOffsetB iOffsetC]: ", motorInfo.motorSN);
    scanf("%d %d %d", &motorInfo.mechOffset, &motorInfo.iOffsetB, &motorInfo.iOffsetC);
#endif    

    PIDinit(&pid, kp, kd, ki, 1); /* Kp, Kd, Ki, dt */
    
    CThreadParams.m = &motorInfo;
    CThreadParams.c = CANdev;
    CThreadParams.p = &pid;
    
    initTR(CANdev, &motorInfo);
    
    compTorque = 0;
    getProperty(CANdev, motorInfo.puckID, ACTUAL_POSITION, &actPos);
    pid.yref = actPos;
    
    syslog(LOG_ERR, "Starting control threads");
    start_control_threads( 40, 1, PuckControlThread, (void*)&CThreadParams);
    
    usleep(1000000);
    
    /* Collect the data */
    if(doCollect)
        err = collectTR(&CThreadParams);
    else
        err = testTR(&CThreadParams);
    
    return(err);
}

int collectTR(CThreadParams_t *CThreadParams)
{
    int         *Tc, *Tf, *Tr;
    long        reply;
    long        actPos, modPos;
    uint64_t    startTime;
    float       delta;
    int         torque;
    long        i;
    int         c;
    int         rev;
    float       torqueSum;
    long        samples;
    int         count;
    char        fileName[32];
    FILE        *outFile, *trFile;
    int         rev2;
    int         length;
    int         end, start;
    float       inc, sum;
    double      cps;
    
    motorInfo_t *motorInfo;
    CANdev_t    *CANdev;
    PIDregulator *pid;
    
    motorInfo = CThreadParams->m;
    CANdev = CThreadParams->c;
    pid = CThreadParams->p;
    
    syslog(LOG_ERR, "Allocating TR arrays");
    
    /* Allocate TorqueRipple Forward Compensation array space */
    Tf = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tf == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Reverse Compensation array space */
    Tr = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tr == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
       
    /* Allocate TorqueRipple Compensation array space */
    Tc = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tc == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Fill the initial arrays with holes */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        Tf[i] = Tr[i] = HOLE;
    }
    

    /* Write the data to a file */
    sprintf(fileName, "motor%d.raw", motorInfo->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "Dir\tRev#\tSamples\tCmdPos\tActPos\tTCmdTrq\n");

    sprintf(fileName, "motor%d.tr", motorInfo->motorSN);
    if((trFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    
    if(getFilesize(trFile) > 50)
    {
        fscanf(trFile, "%d", &motorInfo->encCtsPerRev);
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            fscanf(trFile, "%d %d %d", &Tf[i], &Tr[i], &Tc[i]);
        }
    }
    
    syslog(LOG_ERR, "Looking for DRIVE_MODE == 2");

    /* Spin the motor until DRIVE_MODE == 2 */
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_TORQUE);
    getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
    while(reply != 2)
    {
        pid->yref += 25;
        usleep(1000);
        getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
        printf("\nerr = %f", pid->e);
    }
    
    /* Short delay */
    usleep(50000); 
    
    /* Get the motor near its zero (index pulse) position */
    getProperty(CANdev, motorInfo->puckID, MECH_ANGLE, &reply);
    printf("\n1 = %ld", reply);
    //getch();
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    printf("\n2 = %ld", reply);
    pid->y = reply;
    printf("\n2.5 = %f", pid->yref);
    PIDreset(pid);
    printf("\n3 = %ld", reply);
    if(reply > motorInfo->mechOffset) /* If the present mech_angle > offset */
    {
        /* Go backwards */
        for(pid->yref = reply; pid->yref > motorInfo->mechOffset; pid->yref-=10)
        {
            /* Short delay */
            printf("\n3.5 = %f", pid->e);
            usleep(1000);        
        }
    }else
    {
        /* Go forwards */
        for(pid->yref = reply; pid->yref < motorInfo->mechOffset; pid->yref+=10)
        {
            /* Short delay */
           printf("\n3.5 = %f", pid->e);
            usleep(1000);        
        }
    }
    printf("\n4 = %f", pid->e);
    pid->yref = motorInfo->mechOffset;
    
    /* Short delay */
    usleep(50000); 
    
    /* Set the new "zero" position here near the index pulse */
    getProperty(CANdev, motorInfo->puckID, MECH_ANGLE, &reply);
    reply -= motorInfo->mechOffset;
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    pid->y = reply;
    PIDreset(pid);
    
    /* Check the POSITION_OFFSET */
    getProperty(CANdev, motorInfo->puckID, POSITION_OFFSET, &reply);
    printf("\nFinal POSITION_OFFSET = %ld\n", reply);
#if 0    
    if(reply > 50)
    {
        printf("\nFind Index FAILED!\n");
        exit(1);
    }
    if(reply < -50)
    {
        printf("\nFind Index FAILED!\n");
        exit(1);
    }
#endif    
    //while(1)
    //   usleep(1000);
    pid->Ki = 0;    
    pid->yref = 0;
    
    /* Short delay */
    usleep(500000); 
    
    cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
    
    syslog(LOG_ERR, "Collecting data");
    
    /* Collect the raw forward data */
    for(rev = 0; rev < REVOLUTIONS_F; rev++)
    {
        for(pid->yref = rev * motorInfo->encCtsPerRev; 
            pid->yref < (rev + 1) * motorInfo->encCtsPerRev; 
            pid->yref += STEP)
        {
            samples = 0;
            torqueSum = 0;
            
            /* Calc autoranged position */
            modPos = (long)pid->yref % motorInfo->encCtsPerRev;
            compTorque = Tf[modPos];
                
            startTime = ClockCycles();
            do
            {
                /* Find the new torque value */
                torque = (int)pid->lastresult + compTorque;
    
                /* Calculate elapsed time at this position */
                delta = (float)(ClockCycles() - startTime) / cps;
                if(delta > WAIT_TIME)
                {
                    /* Record the torque sum */
                    torqueSum += torque;
                    ++samples;
                }
                //syslog(LOG_ERR, "Delta: %lf, SampleT: %lf", delta, TR_SAMPLE_TIME);
                usleep(50);
            }while(delta < TR_SAMPLE_TIME); /* Run for TR_SAMPLE_TIME seconds */
            
            /* Save the average torque */
            Tf[modPos] = (int)(torqueSum / samples);
            fprintf(outFile, "F\t%d\t%d\t%ld\t%ld\t%d\n", rev, samples, modPos, (long)pid->y % motorInfo->encCtsPerRev, Tf[modPos]);
            printf("F\t%d\t%d\t%ld\t%ld\t%d\n", rev, samples, modPos, (long)pid->y % motorInfo->encCtsPerRev, Tf[modPos]);
        }
        
        /* Average every SMOOTH_CT data points of Tf into Tc */
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            torqueSum = 0;
            for(count = 0; count < SMOOTH_CT; count++)
            {
                modPos = i + count - (SMOOTH_CT-1) / 2;
                modPos %= motorInfo->encCtsPerRev;
                if(modPos < 0) modPos += motorInfo->encCtsPerRev;
                torqueSum += Tf[modPos];
                
            }
            Tc[i] = torqueSum / SMOOTH_CT;
        }
        
        /* Save the smoothed values back into Tf */
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            Tf[i] = Tc[i];
        }
        
        /* Save this cycle */
        rewind(trFile);
        fprintf(trFile, "%d\n", motorInfo->encCtsPerRev);
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            fprintf(trFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
        }
        
    }
    
    rev2 = -1;
    //setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, (REVOLUTIONS_R * motorInfo->encCtsPerRev)-1 );
    for(rev = REVOLUTIONS_R; rev > 0; rev--)
    {  
        rev2++;
        /* Collect the raw reverse data */
        for(pid->yref = (rev * motorInfo->encCtsPerRev)-1; 
            pid->yref >= (rev - 1) * motorInfo->encCtsPerRev; 
            pid->yref -= STEP)
        {
            samples = 0;
            torqueSum = 0;
            
            /* Calc autoranged position */
            modPos = (long)pid->yref % motorInfo->encCtsPerRev;
            compTorque = Tr[modPos];
            
            startTime = ClockCycles();
            do
            {
                /* Find the new torque value */
                torque = (int)pid->lastresult + compTorque;
  
                /* Calculate elapsed time at this position */
                delta = (float)(ClockCycles() - startTime) / cps;
                if(delta > WAIT_TIME)
                {
                    /* Record the torque sum */
                    torqueSum += torque;
                    ++samples;
                }
                usleep(50);
            }while(delta < TR_SAMPLE_TIME); /* Run for TR_SAMPLE_TIME seconds */
             
            /* Save the average torque */
            Tr[modPos] = (int)(torqueSum / samples);
            fprintf(outFile, "R\t%d\t%d\t%ld\t%ld\t%d\n", rev2, samples, modPos, (long)pid->y % motorInfo->encCtsPerRev, Tr[modPos]);
            printf("R\t%d\t%d\t%ld\t%ld\t%d\n", rev2, samples, modPos, (long)pid->y % motorInfo->encCtsPerRev, Tr[modPos]);
        }
        
       /* Average every SMOOTH_CT data points of Tf into Tc */
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            torqueSum = 0;
            for(count = 0; count < SMOOTH_CT; count++)
            {
                modPos = i + count - (SMOOTH_CT-1) / 2;
                modPos %= motorInfo->encCtsPerRev;
                if(modPos < 0) modPos += motorInfo->encCtsPerRev;
                torqueSum += Tr[modPos];
                
            }
            Tc[i] = torqueSum / SMOOTH_CT;
        }
        
        /* Save the smoothed values back into Tr */
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            Tr[i] = Tc[i];
        }
        
        /* Save this cycle */
        rewind(trFile);
        fprintf(trFile, "%d\n", motorInfo->encCtsPerRev);
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
        {
            fprintf(trFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
        }
  
    }
    
    /* Close the RAW data file */
    fclose(outFile);

    /* Close *.tr data file */
    fclose(trFile);
    
    /* Smooth it out */
    /* Average every SMOOTH_CT data points of Tf and Tr, then average Tf and Tr into Tc */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        torqueSum = 0;
        for(count = 0; count < SMOOTH_CT; count++)
        {
            modPos = i + count - (SMOOTH_CT-1) / 2;
            modPos %= motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            torqueSum += Tf[modPos];
            torqueSum += Tr[modPos];
        }
        Tc[i] = torqueSum / (SMOOTH_CT * 2);
    }
    
    /* Write the data to a file */
    sprintf(fileName, "motor%d.tr", motorInfo->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "%d\n", motorInfo->encCtsPerRev);
    
    /* Output the data */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        /* Save the forward data, reverse data, and smoothed data */
        fprintf(outFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
    }
    
    fclose(outFile);
    
    printf("\nDone collecting TR data!\n");
    
    /* Clean up */
    free(Tf);
    free(Tr);
    free(Tc);
    return(0);
}

int testTR(CThreadParams_t *CThreadParams)
{
    int         i;
    char        fileName[32];
    FILE        *in;
    int         *Tf, *Tr, *Tc;
    int        modPos;
    long        reply = 0;
    int         torque;
    int         junk;
    double      gain;
    long        posDiff;
    long        actPos;
    motorInfo_t *motorInfo;
    CANdev_t    *CANdev;
    PIDregulator *pid;
    
    motorInfo = CThreadParams->m;
    CANdev = CThreadParams->c;
    pid = CThreadParams->p;
    
    printf("\nPlease enter *.tr file name: ");
    scanf("%s", fileName);
    
    if((in = fopen(fileName, "r")) == NULL)
    {
        printf("\nFailed to open torque ripple file!");
        return(1);
    }

    /* Allocate TorqueRipple Compensation array space */
    Tf = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tf == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Compensation array space */
    Tr = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tr == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Compensation array space */
    Tc = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(Tc == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    fscanf(in, "%d", &motorInfo->encCtsPerRev); 
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        fscanf(in, "%d %d %d", &Tf[i], &Tr[i], &Tc[i]);
    }
    fclose(in);

#if 0    
    /* Spin the motor until DRIVE_MODE == 2 */
    
    compTorque = 0;
    getProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, &actPos);
    pid->yref = actPos;
    PIDreset(pid);
    
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_TORQUE);
    do
    {
        getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
        pid->yref += 25;
        usleep(1000);
    }while(reply != 2);
    
    /* Zero the position here at this arbitrary point */
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, 0);
    pid->yref = 0;
    PIDreset(pid);
    
    /* Check the POSITION_OFFSET */
    getProperty(CANdev, motorInfo->puckID, POSITION_OFFSET, &reply);
    printf("\nFinal POSITION_OFFSET = %ld", reply);
#endif

    syslog(LOG_ERR, "Looking for DRIVE_MODE == 2");

    /* Spin the motor until DRIVE_MODE == 2 */
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_TORQUE);
    getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
    while(reply != 2)
    {
        pid->yref += 25;
        usleep(1000);
        getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
        printf("\nerr = %f", pid->e);
    }
   
    /* Short delay */
    usleep(50000); 
    
    /* Get the motor near its zero (index pulse) position */
    getProperty(CANdev, motorInfo->puckID, MECH_ANGLE, &reply);
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    pid->y = reply;
    PIDreset(pid);
#if 0     
    if(reply > motorInfo->mechOffset) /* If the present mech_angle > offset */
    {
        /* Go backwards */
        for(pid->yref = reply; pid->yref > motorInfo->mechOffset; pid->yref-=10)
        {
            /* Short delay */
            usleep(1000);        
        }
    }else
    {
        /* Go forwards */
        for(pid->yref = reply; pid->yref < motorInfo->mechOffset; pid->yref+=10)
        {
            /* Short delay */
            usleep(1000);        
        }
    }
    
    pid->yref = motorInfo->mechOffset;
    
    /* Short delay */
    usleep(50000); 
    
    /* Set the new "zero" position here near the index pulse */
    getProperty(CANdev, motorInfo->puckID, MECH_ANGLE, &reply);
    reply -= motorInfo->mechOffset;
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    pid->y = reply;
    PIDreset(pid);
#endif    
    /* Check the POSITION_OFFSET */
    getProperty(CANdev, motorInfo->puckID, POSITION_OFFSET, &reply);
    printf("\nFinal POSITION_OFFSET = %ld\n", reply);
        
    /* TR Test */
    pid->Ki = 0;
    pid->Kd = 0;
    printf("\nTR Test, press ctrl-c to exit..."); fflush(stdout);
    do
    {
        pid->yref = pid->y;
        modPos = ((long)pid->y + reply) % motorInfo->encCtsPerRev;
        if(modPos < 0) modPos += motorInfo->encCtsPerRev;
        compTorque = Tc[modPos];
        //printf("\ncompTorque = %d\t", compTorque);
        usleep(1000);
        
    }while(1);
    
}

int initTR(CANdev_t *CANdev, motorInfo_t *motorInfo)
{
    long reply;
    //int torque;
    char c;
    int i;

    syslog(LOG_ERR, "Check the status of puck %d", motorInfo->puckID);

    /* Check the status of the puck */
    getProperty(CANdev, motorInfo->puckID, STATUS, &reply);
    syslog(LOG_ERR, "Puck status is %s", statusTxt[reply+1].s);
    switch(reply)
    {
        case STATUS_RESET:
            /* Get the puck out of RESET */
            setProperty(CANdev, motorInfo->puckID, STATUS, STATUS_INITIALIZE);
        
            /* Wait a moment */
            usleep(1000);
            
            /* Check the status of the puck */
            getProperty(CANdev, motorInfo->puckID, STATUS, &reply);
            
            if(reply != STATUS_READY)
            {
                printf("\nSomething is wrong with puck %d!  Try cycling the power.\n");
                return(1);
            }
            break;
        case STATUS_READY:
            /* Ready to go */
            
            break;
        default:
            printf("\nSomething is wrong with puck %d!  Try cycling the power.\n");
            return(1);
    }
    
    /* Send the offsets */
    setProperty(CANdev, motorInfo->puckID, MOTOR_OFFSET, motorInfo->mechOffset);
    setProperty(CANdev, motorInfo->puckID, I_OFFSET_B, motorInfo->iOffsetB);
    setProperty(CANdev, motorInfo->puckID, I_OFFSET_C, motorInfo->iOffsetC);
    setProperty(CANdev, motorInfo->puckID, CTS_PER_REV, motorInfo->encCtsPerRev);

    /* Go live with zero torque */
    setProperty(CANdev, motorInfo->puckID, COMMANDED_TORQUE, 0);
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_IDLE);
    
    
}

int main( int argc, char **argv )
{
    char            c;
    int             err;
    
    CANdev_t        CANdev;
    int             CANfd;
 
    /* Initialize syslogd */
    openlog("LOG_ERR",LOG_CONS | LOG_NDELAY, LOG_USER);
    syslog(LOG_ERR, "syslog initalized");
    
    /* Initialize CAN */
    if(err = initCAN("/dev/dpm_00", &CANdev.CANfd))
    {
        syslog(LOG_ERR, "initCAN returned err=%d", err);
    }
    if(err = pthread_mutex_init(&CANdev.mutex, NULL))
    {
        syslog(LOG_ERR, "pthread_mutex_init returned err=%d", err);
    }
    CANdev.channel = 0;

    /* Show Menu */
    showMenu();
    
    /* Get Choice */
    c = getchar();
    //printf("\nYou pressed ASCII: %d", c);
    c = toupper(c);
    
    /* Handle Menu */
    handleMenu(c, &CANdev);
    
    return(0);
}

void PuckControlThread(void *data)
{
    long actPos;
    CThreadParams_t *CThreadParams;
    
    CThreadParams = (CThreadParams_t*)data;
    
    syslog(LOG_ERR, "Started PuckControlThread");
    while (!shutdown_threads)
    {
        sem_wait(&timer_semaphore);
        //syslog(LOG_ERR, "Timer posted");
        getProperty(CThreadParams->c, CThreadParams->m->puckID, ACTUAL_POSITION, &actPos);
        //syslog(LOG_ERR, "Ctrl Thread Position received");
        /* Update PID */
        CThreadParams->p->y = actPos; 
        PIDcalc(CThreadParams->p);

        /* Set the new torque */
        setProperty_noVerify(CThreadParams->c, CThreadParams->m->puckID, COMMANDED_TORQUE, (int)CThreadParams->p->lastresult + compTorque);

    }

    pthread_exit(NULL);
}

#if 0    
    /* Fill in Forward holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {
        /* Find the last non-HOLE */
        start = i;
        while(Tf[start] == HOLE)
        {
            --start;
            if(start < 0)
                start = motorInfo->encCtsPerRev;
        }
        
        /* Find the next non-HOLE */
        end = i;
        while(Tf[end] == HOLE)
        {
            ++end;
            if(end > motorInfo->encCtsPerRev-1)
                end = 0;
        }
        
        /* Find the length of the hole */
        if(end > start)
            length = end - start;
        else
            length = end + motorInfo->encCtsPerRev - start;
        
        /* Find the increment */
        inc = (float)(Tf[end] - Tf[start]) / (float)length;
        sum = 0;
        
        /* Fill in data */
        for(end = 1; end < length; end++)
        {
            ++start;
            sum += inc;
            if(start == motorInfo->encCtsPerRev)
                start = 0;
            Tf[start] = (int)sum;
        }
    }

    /* Fill in Reverse holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {
        /* Find the last non-HOLE */
        start = i;
        while(Tr[start] == HOLE)
        {
            --start;
            if(start < 0)
                start = motorInfo->encCtsPerRev;
        }
        
        /* Find the next non-HOLE */
        end = i;
        while(Tr[end] == HOLE)
        {
            ++end;
            if(end > motorInfo->encCtsPerRev-1)
                end = 0;
        }
        
        /* Find the length of the hole */
        if(end > start)
            length = end - start;
        else
            length = end + motorInfo->encCtsPerRev - start;
        
        /* Find the increment */
        inc = (float)(Tr[end] - Tr[start]) / (float)length;
        sum = 0;
        
        /* Fill in data */
        for(end = 1; end < length; end++)
        {
            ++start;
            sum += inc;
            if(start == motorInfo->encCtsPerRev)
                start = 0;
            Tr[start] = (int)sum;
        }
    }

#endif
           
#if 0     
    /* Fill in holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {

        /* Fill in forward data */
        count = 0;
        while(Tf[i+count] == HOLE)
        {
            ++count;
            if(i + count >= motorInfo->encCtsPerRev /* * REVOLUTIONS */)
                count -= motorInfo->encCtsPerRev /* * REVOLUTIONS */;
        }
        Tf[i] = Tf[i+count];
                
        /* Fill in reverse data */
        count = 0;
        while(Tr[i+count] == HOLE)
        {
            ++count;
            if(i + count >= motorInfo->encCtsPerRev /* * REVOLUTIONS */)
                count -= motorInfo->encCtsPerRev /* * REVOLUTIONS */;
        }
        Tr[i] = Tr[i+count];
        
    }

   
    /* Average each REVOLUTIONS data points */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        for(rev = 1; rev < REVOLUTIONS; rev++)
        {
            Tf[i] = Tf[i] + Tf[(rev * motorInfo->encCtsPerRev) + i];
            Tr[i] = Tr[i] + Tr[(rev * motorInfo->encCtsPerRev) + i];
        }
        Tf[i] = Tf[i] / REVOLUTIONS;
        Tr[i] = Tr[i] / REVOLUTIONS;
    }

    /* Average forward and reverse torques */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        Tc[i] = (Tf[i] + Tr[i]) / 2;
    }
    
    /* Now run the filter BACKWARDS to eliminate the spatial phase introduced *
     * above.  Likewise, average every SMOOTH_CT data points. */                       
    for(i = motorInfo->encCtsPerRev-1; i >= 0; i--)
    {
        torqueSum = 0;
        for(count = 0; count < SMOOTH_CT; count++)
        {
            modPos = i + count - (SMOOTH_CT-1) / 2;
            modPos %= motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            torqueSum += Tc[modPos];
        }
        Tc[i] = torqueSum / SMOOTH_CT;
    }

    /* Apply a gain */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        Tc[i] = (int)((float)Tc[i] * TR_GAIN);
    }
#endif

#if(0) /* Sample tr.info file */

PuckID : 5
MotorSN : 10
CtsPerRev : 6600
mechOffset : 233
iOffsetB : 575
iOffsetC : 575
Rev_F : 5
Rev_R : 5
StepSize : 1
SmoothCt : 12
SampleTime : 0.010
WaitTime : 0.0025
Kp : 2
Kd : 10
Ki : 0

#endif