/* main.c */

#define toupper(c)      ( ((c >= 'a') && (c <= 'z')) ? c - ('a' - 'A') : c )
#define HOLE            (-1)

#define TEST_TR         (0)
#define COLLECT_TR      (1)

#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif

#include <stdio.h>
#include <syslog.h>
#include <sys/neutrino.h>
#include <sys/syspage.h>
#include <inttypes.h>
#include <malloc.h>
#include <math.h>
#include <semaphore.h>

#include "btcan.h"
#include "SimpleControl.h"

//extern  local_info_t    priv_data;
int     compTorque;
extern sem_t   timer_semaphore;
extern int     shutdown_threads;

typedef struct
{
    int     puckID;
    int     motorSN;
    long    encCtsPerRev;
    long    mechOffset;
    int     iOffsetB;
    int     iOffsetC;
}motorInfo_t;

typedef struct
{
    int     fwd;
    int     rev;
    long    samplesPerRev;
    int     smoothCt;
    int     epsilon;
    int     waitSamples;
    double  kp;
    double  kd;
    double  ki;
}trSetup_t;

typedef struct{char s[20];}string20;

typedef struct
{
    motorInfo_t     *m;
    CANdev_t        *c;
    PIDregulator    *p;
    trSetup_t       *s;
}CThreadParams_t;

const    string20	statusTxt[]={
        {"STATUS_OFFLINE"},
    	{"STATUS_RESET"},
        {"STATUS_SHUTDOWN"},
        {"STATUS_INITIALIZE"},
        {"STATUS_ERR"},
        {"STATUS_READY"},
        {"STATUS_COMMUTATING"},
    };

void PuckControlThread(void *data);

void showMenu(void){
    printf("\nMENU");
    printf("\n--------");
    printf("\nE)numerate bus status");
    printf("\nM)easure CAN speed");
    printf("\nD)ownload firmware");
    printf("\nO)ld firmware download");
    printf("\nC)ollect TR data");
    printf("\nT)est TR data");
    
    printf("\n\nYour Choice: ");
}

void handleMenu(char c, CANdev_t *CANdev){
    char        fname[32];
    int         node;
    int         err;
    uint64_t    time1, time2;
    float       delta;
    long        reply;
    int         status[MAX_NODES];
    int         i;
    double      cps;
    int         newID;
    
    
    switch(c){
    case 'E':
        printf("\n\nCAN bus enumeration (status)\n");
        getBusStatus(CANdev, status);
        for(i = 0; i < MAX_NODES; i++)
        {
            if(i % 2)
                printf("\t\tNode %2d: %s", i, statusTxt[status[i]+1].s);
            else
                printf("\nNode %2d: %s", i, statusTxt[status[i]+1].s);
        }
        printf("\n");
        break;
        
    case 'M':
        printf("\n\nCAN Speed Test...\n");
        cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
        while(1)
        {
            time1 = ClockCycles();
            getProperty(CANdev, 25, STATUS, &reply);
            time2 = ClockCycles();
            delta = (double)(time2 - time1) / cps;
            printf("\r%f", delta); fflush(stdout);
        }
        
    case 'D':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Get new Node ID */
        if(node > 0)
        {
            printf("\nChange node ID to (enter 0 for no change): ");
            err = scanf("%d", &newID);
            if(err == EOF) /* No input data */
                newID = 0; /* Zero signifies "no change" */
        }
        
        /* Download the firmware */
        err = downloadFirmware(CANdev, node, newID, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'O':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Download the firmware */
        err = downloadFirmware_Old(CANdev, node, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'C':
        /* Collect TR Data */
        doTR(CANdev, COLLECT_TR);
        break;
        
    case 'T':
        doTR(CANdev, TEST_TR);
        break;
        
    default:
    
        break;
    }
}

int doTR(CANdev_t *CANdev, int doCollect)
{
    motorInfo_t motorInfo;
    int         err;
    FILE        *in;
    CThreadParams_t CThreadParams;
    PIDregulator pid;
    long        actPos;
    trSetup_t   trSetup;
    
    if((in = fopen("tr.info","r")) == NULL)
    {
        printf("\nFailed to open tr.info file!");
        return(1);
    }
    fscanf(in, "%*s : %d", &motorInfo.puckID);
    fscanf(in, "%*s : %d", &motorInfo.motorSN);
    fscanf(in, "%*s : %ld", &motorInfo.encCtsPerRev);
    fscanf(in, "%*s : %ld %*s : %d %*s : %d", &motorInfo.mechOffset, &motorInfo.iOffsetB, &motorInfo.iOffsetC);
    fscanf(in, "%*s : %d %*s : %d %*s : %ld %*s : %d %*s : %d %*s : %d", &trSetup.fwd, &trSetup.rev, &trSetup.samplesPerRev, &trSetup.smoothCt, &trSetup.epsilon, &trSetup.waitSamples);
    fscanf(in, "%*s : %lf %*s : %lf %*s : %lf", &trSetup.kp, &trSetup.kd, &trSetup.ki);
    
    fclose(in);
    
    printf( "CANID      : %d\n"
            "MOTOR_SN   : %d\n"
            "CTS_PER_REV: %d\n"
            "MECH_OFFSET: %d \n"
            "IOFFSET_B  : %d \n"
            "IOFFSET_C  : %d\n"
            "FWD_REVS   : %d \n"
            "REV_REVS   : %d \n"
            "SAMP/REV   : %ld \n"
            "SMOOTH_CT  : %d \n"
            "STOPPED e  : %d \n"
            "STOPPED CT : %d \n"
            "KP         : %0.3f \n"
            "KD         : %0.3f \n"
            "KI         : %0.3f\n", motorInfo.puckID, motorInfo.motorSN, motorInfo.encCtsPerRev,
            motorInfo.mechOffset, motorInfo.iOffsetB, motorInfo.iOffsetC,
            trSetup.fwd, trSetup.rev, trSetup.samplesPerRev, trSetup.smoothCt, trSetup.epsilon, trSetup.waitSamples,
            trSetup.kp, trSetup.kd, trSetup.ki);
   
    getchar();

    PIDinit(&pid, trSetup.kp, trSetup.kd, trSetup.ki, 1); /* Kp, Kd, Ki, dt */
    
    CThreadParams.m = &motorInfo;
    CThreadParams.c = CANdev;
    CThreadParams.p = &pid;
    CThreadParams.s = &trSetup;
    
    initTR(CANdev, &motorInfo);
    
    compTorque = 0;
    getProperty(CANdev, motorInfo.puckID, ACTUAL_POSITION, &actPos);
    pid.yref = actPos;
    
    syslog(LOG_ERR, "Starting control threads");
    start_control_threads( 40, 1, PuckControlThread, (void*)&CThreadParams);
    
    usleep(1000000);
    
    /* Collect the data */
    if(doCollect)
        err = collectTR(&CThreadParams);
    else
        err = testTR(&CThreadParams);
    
    return(err);
}

void motorSettle(CANdev_t *CANdev, int puckID, int epsilon, int waitSamples)
{
    long reply, lastReply;
    int ctsRemaining;
    
    getProperty(CANdev, puckID, ACTUAL_POSITION, &lastReply);
    usleep(1000);
    
    ctsRemaining = waitSamples;
    while(ctsRemaining)
    {
        getProperty(CANdev, puckID, ACTUAL_POSITION, &reply);
        if(abs(reply - lastReply) <= epsilon) // If we are within episilon
        {
            --ctsRemaining;
            
        }else{
            ctsRemaining = waitSamples;
            usleep(1000);
        }
        
        lastReply = reply;    
    }
}

void fillHoles(int *aCts, int *aTrq, long ctsPerRev)
{
    int i;
    long    sampleLocA, sampleLocB;
    int     sampleValA, sampleValB;
    double  inc, sum;
    int     done = FALSE;
    
    /* Go from aCts[0] to aCts[1], [1] to [2], etc */
    sampleLocA = aCts[0];
    i = 1;
    while(!done)
    {
        if(aCts[i] != HOLE)
        {
            sampleLocB = aCts[i];
        }else{
            sampleLocB = aCts[0];
            done = TRUE;
        }
        
        /* Calculate increment */
        inc = (double)(aTrq[sampleLocB] - aTrq[sampleLocA]) / (sampleLocB - sampleLocA);
        
        ++sampleLocA;
        if(sampleLocA == ctsPerRev) sampleLocA = 0;
        
        sum = 0;
        while(sampleLocA != sampleLocB)
        {
            sum += inc;
            aTrq[sampleLocA] = (int)sum;
            ++sampleLocA;
            if(sampleLocA == ctsPerRev) sampleLocA = 0; 
        }
        ++i;   
    }
}

void smoothData(int *Tc, int *Tf, long ctsPerRev, int smoothCt)
{
    int     i;
    int     torqueSum;
    long    modPos;
    int     count;
    
    /* Average every SMOOTH_CT data points of Tf into Tc */
    for(i = 0; i < ctsPerRev; i++)
    {
        torqueSum = 0;
        for(count = 0; count < smoothCt; count++)
        {
            modPos = i + count - (smoothCt-1) / 2;
            modPos %= ctsPerRev;
            if(modPos < 0) modPos += ctsPerRev;
            torqueSum += Tf[modPos];
            
        }
        Tc[i] = torqueSum / smoothCt;
    }
    
    /* Save the smoothed values back into Tf */
    for(i = 0; i < ctsPerRev; i++)
    {
        Tf[i] = Tc[i];
    }
}

void outputData(FILE *outFile, int *aTrq, long ctsPerRev, char dir, int revolution)
{
    int i;
    
    for(i = 0; i < ctsPerRev; i++)
    {
        fprintf(outFile, "%c\t%d\t%d\n", dir, revolution, aTrq[i]);
    }
}

int collectTR(CThreadParams_t *CThreadParams)
{
    long    pOffset;
    int     *aTf;
    int     *aTr;
    int     *aCts;
    FILE    *outFile;
    long    reply;
    int     revolution;
    long    sampleDistance;
    long    modPos;
    int     i;
    char    fileName[20];
    long    torqueSum;
    int     count;
    
    motorInfo_t     *motorInfo;
    CANdev_t        *CANdev;
    PIDregulator    *pid;
    trSetup_t       *trSetup;
    
    motorInfo = CThreadParams->m;
    CANdev = CThreadParams->c;
    pid = CThreadParams->p;
    trSetup = CThreadParams->s;
    
    sampleDistance = motorInfo->encCtsPerRev / trSetup->samplesPerRev;
    
    syslog(LOG_ERR, "Allocating TR arrays");
    
    if((aTf = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int))) == NULL);
    {
        syslog(LOG_ERR, "Could not allocate aTrq array");
        exit(1);
    }
    
    if((aTr = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int))) == NULL);
    {
        syslog(LOG_ERR, "Could not allocate aTrq array");
        exit(1);
    }
    
    if((aCts = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int))) == NULL);
    {
        syslog(LOG_ERR, "Could not allocate aCts array");
        exit(1);
    }
        
    /* Write the *.raw file */
    sprintf(fileName, "motor%d.raw", motorInfo->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        syslog(LOG_ERR, "Unable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "Dir\tRev#\tSamples\tCmdPos\tActPos\tCmdTrq\n");

    syslog(LOG_ERR, "Looking for DRIVE_MODE == 2");

    /* Spin the motor until DRIVE_MODE == 2 */
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_TORQUE);
    getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
    while(reply != 2)
    {
        pid->yref += 25;
        usleep(1000);
        getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
        //printf("\nerr = %f", pid->e);
    }
    
    /* Set the present position to itself (force pOffset calculation) */
    getProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, &reply);
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    
    /* Wait for motor position to settle */ 
    motorSettle(CANdev, motorInfo->puckID, trSetup->epsilon, trSetup->waitSamples);
    
    /* Record the position offset */
    getProperty(CANdev, motorInfo->puckID, POSITION_OFFSET, &pOffset);
    
    //cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
    
    syslog(LOG_ERR, "Collecting data");
    
    // Clear aTrq to zeros
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
        aTf[i] = 0;
        
    for(revolution = 1; revolution <= trSetup->fwd; revolution++)
    {
        // Clear aCts to zeros
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
            aCts[i] = HOLE;
            
        for(i = 0; i < trSetup->samplesPerRev; i++)
        {
            /* Go to the new position */
            pid->yref += sampleDistance;
            modPos = ((long)pid->yref + pOffset) % motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            compTorque = aTf[modPos];
            
            /* Wait for motor position to settle */ 
            motorSettle(CANdev, motorInfo->puckID, trSetup->epsilon, trSetup->waitSamples);
            
            /* Record the torque */
            aCts[i] = modPos;
            aTf[modPos] = (int)pid->lastresult + compTorque;
        }
        
        // Fill the holes
        fillHoles(aCts, aTf, motorInfo->encCtsPerRev);
        
        // Smooth the data
        smoothData(aCts, aTf, motorInfo->encCtsPerRev, trSetup->smoothCt);
        
        // Output the data
        outputData(outFile, aTf, motorInfo->encCtsPerRev, 'F', revolution);
    }
    
    // Clear aTrq to zeros
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
        aTr[i] = 0;
        
    for(revolution = 1; revolution <= trSetup->rev; revolution++)
    {
        // Clear aCts to zeros
        for(i = 0; i < motorInfo->encCtsPerRev; i++)
            aCts[i] = HOLE;
            
        for(i = 0; i < trSetup->samplesPerRev; i++)
        {
            /* Go to the new position */
            pid->yref -= sampleDistance;
            modPos = ((long)pid->yref + pOffset) % motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            compTorque = aTr[modPos];
            
            /* Wait for motor position to settle */ 
            motorSettle(CANdev, motorInfo->puckID, trSetup->epsilon, trSetup->waitSamples);
            
            /* Record the torque */
            aCts[i] = modPos;
            aTr[modPos] = (int)pid->lastresult + compTorque;
        }
        
        // Fill the holes
        fillHoles(aCts, aTr, motorInfo->encCtsPerRev);
        
        // Smooth the data
        smoothData(aCts, aTr, motorInfo->encCtsPerRev, trSetup->smoothCt);
        
        // Output the data
        outputData(outFile, aTr, motorInfo->encCtsPerRev, 'R', revolution);
    }
    
    /* Close the RAW data file */
    fclose(outFile);
    
    /* Smooth it out */
    /* Average every SMOOTH_CT data points of Tf and Tr, then average Tf and Tr into Tc */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        torqueSum = 0;
        for(count = 0; count < trSetup->smoothCt; count++)
        {
            modPos = i + count - (trSetup->smoothCt-1) / 2;
            modPos %= motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            torqueSum += aTf[modPos];
            torqueSum += aTr[modPos];
        }
        aCts[i] = torqueSum / (trSetup->smoothCt * 2);
    }
    
    /* Write the data to a file */
    sprintf(fileName, "motor%d.tr", motorInfo->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "%d\n", motorInfo->encCtsPerRev);
    
    /* Output the data */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        /* Save the combined data */
        fprintf(outFile, "%d\n", aCts[i]);
    }
    
    /* Close *.tr data file */
    fclose(outFile);
    
    printf("\nDone collecting TR data!\n");
    
    /* Clean up */
    free(aTf);
    free(aTr);
    free(aCts);
    
    return(0);
}


    
int testTR(CThreadParams_t *CThreadParams)
{
    long    pOffset;
    int     *aTrq;
    long    reply;
    long    modPos;
    int     i;
    FILE    *in;
    char    fileName[20];
    
    motorInfo_t     *motorInfo;
    CANdev_t        *CANdev;
    PIDregulator    *pid;
    trSetup_t       *trSetup;
    
    motorInfo = CThreadParams->m;
    CANdev = CThreadParams->c;
    pid = CThreadParams->p;
    trSetup = CThreadParams->s;
    
    //printf("\nPlease enter *.tr file name: ");
    //scanf("%s", fileName);
    sprintf(fileName, "motor%d.tr", motorInfo->puckID);
    
    if((in = fopen(fileName, "r")) == NULL)
    {
        printf("\nFailed to open torque ripple file!");
        return(1);
    }

    /* Allocate TorqueRipple Compensation array space */
    aTrq = (int *)malloc(motorInfo->encCtsPerRev * sizeof(int));
    if(aTrq == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate aTrq buffer!");
        return(1);
    }
    
    fscanf(in, "%d", &motorInfo->encCtsPerRev); 
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        fscanf(in, "%d", &aTrq[i]);
    }
    fclose(in);

    syslog(LOG_ERR, "Looking for DRIVE_MODE == 2");

    /* Spin the motor until DRIVE_MODE == 2 */
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_TORQUE);
    getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
    while(reply != 2)
    {
        pid->yref += 25;
        usleep(1000);
        getProperty(CANdev, motorInfo->puckID, DRIVE_MODE, &reply);
        //printf("\nerr = %f", pid->e);
    }
    
    /* Set the present position to itself (force pOffset calculation) */
    getProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, &reply);
    setProperty(CANdev, motorInfo->puckID, ACTUAL_POSITION, reply);
    
    /* Wait for motor position to settle */ 
    motorSettle(CANdev, motorInfo->puckID, trSetup->epsilon, trSetup->waitSamples);
    
    /* Record the position offset */
    getProperty(CANdev, motorInfo->puckID, POSITION_OFFSET, &pOffset);
        
    /* TR Test */
    pid->Ki = 0;
    pid->Kd = 0;
    printf("\nTR Test, press ctrl-c to exit..."); fflush(stdout);
    
    while(1)
    {
        pid->yref = pid->y;
        modPos = ((long)pid->y + pOffset) % motorInfo->encCtsPerRev;
        if(modPos < 0) modPos += motorInfo->encCtsPerRev;
        compTorque = aTrq[modPos];
        //printf("\ncompTorque = %d\t", compTorque);
        usleep(1000);
        
    }
    
}

int initTR(CANdev_t *CANdev, motorInfo_t *motorInfo)
{
    long reply;
    //int torque;
    char c;
    int i;

    syslog(LOG_ERR, "Check the status of puck %d", motorInfo->puckID);

    /* Check the status of the puck */
    getProperty(CANdev, motorInfo->puckID, STATUS, &reply);
    syslog(LOG_ERR, "Puck status is %s", statusTxt[reply+1].s);
    switch(reply)
    {
        case STATUS_RESET:
            /* Get the puck out of RESET */
            setProperty(CANdev, motorInfo->puckID, STATUS, STATUS_INITIALIZE);
        
            /* Wait a moment */
            usleep(1000);
            
            /* Check the status of the puck */
            getProperty(CANdev, motorInfo->puckID, STATUS, &reply);
            
            if(reply != STATUS_READY)
            {
                printf("\nSomething is wrong with puck %d!  Try cycling the power.\n");
                return(1);
            }
            break;
        case STATUS_READY:
            /* Ready to go */
            
            break;
        default:
            printf("\nSomething is wrong with puck %d!  Try cycling the power.\n");
            return(1);
    }
    
    /* Send the offsets */
    setProperty(CANdev, motorInfo->puckID, MOTOR_OFFSET, motorInfo->mechOffset);
    setProperty(CANdev, motorInfo->puckID, I_OFFSET_B, motorInfo->iOffsetB);
    setProperty(CANdev, motorInfo->puckID, I_OFFSET_C, motorInfo->iOffsetC);
    setProperty(CANdev, motorInfo->puckID, CTS_PER_REV, motorInfo->encCtsPerRev);

    /* Go live with zero torque */
    setProperty(CANdev, motorInfo->puckID, COMMANDED_TORQUE, 0);
    setProperty(CANdev, motorInfo->puckID, CONTROL_MODE, MODE_IDLE);
    
    
}

int main( int argc, char **argv )
{
    char            c;
    int             err;
    
    CANdev_t        CANdev;
    int             CANfd;
 
    /* Initialize syslogd */
    openlog("LOG_ERR",LOG_CONS | LOG_NDELAY, LOG_USER);
    syslog(LOG_ERR, "syslog initalized");
    
    /* Initialize CAN */
    if(err = initCAN("/dev/dpm_00", &CANdev.CANfd))
    {
        syslog(LOG_ERR, "initCAN returned err=%d", err);
    }
    if(err = pthread_mutex_init(&CANdev.mutex, NULL))
    {
        syslog(LOG_ERR, "pthread_mutex_init returned err=%d", err);
    }
    CANdev.channel = 0;

    /* Show Menu */
    showMenu();
    
    /* Get Choice */
    c = getchar();
    //printf("\nYou pressed ASCII: %d", c);
    c = toupper(c);
    
    /* Handle Menu */
    handleMenu(c, &CANdev);
    
    return(0);
}

void PuckControlThread(void *data)
{
    long actPos;
    CThreadParams_t *CThreadParams;
    
    CThreadParams = (CThreadParams_t*)data;
    
    syslog(LOG_ERR, "Started PuckControlThread");
    while (!shutdown_threads)
    {
        sem_wait(&timer_semaphore);
        //syslog(LOG_ERR, "Timer posted");
        getProperty(CThreadParams->c, CThreadParams->m->puckID, ACTUAL_POSITION, &actPos);
        //syslog(LOG_ERR, "Ctrl Thread Position received");
        /* Update PID */
        CThreadParams->p->y = actPos; 
        PIDcalc(CThreadParams->p);

        /* Set the new torque */
        setProperty_noVerify(CThreadParams->c, CThreadParams->m->puckID, COMMANDED_TORQUE, (int)CThreadParams->p->lastresult + compTorque);

    }

    pthread_exit(NULL);
}



#if 0    
    /* Fill in Forward holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {
        /* Find the last non-HOLE */
        start = i;
        while(Tf[start] == HOLE)
        {
            --start;
            if(start < 0)
                start = motorInfo->encCtsPerRev;
        }
        
        /* Find the next non-HOLE */
        end = i;
        while(Tf[end] == HOLE)
        {
            ++end;
            if(end > motorInfo->encCtsPerRev-1)
                end = 0;
        }
        
        /* Find the length of the hole */
        if(end > start)
            length = end - start;
        else
            length = end + motorInfo->encCtsPerRev - start;
        
        /* Find the increment */
        inc = (float)(Tf[end] - Tf[start]) / (float)length;
        sum = 0;
        
        /* Fill in data */
        for(end = 1; end < length; end++)
        {
            ++start;
            sum += inc;
            if(start == motorInfo->encCtsPerRev)
                start = 0;
            Tf[start] = (int)sum;
        }
    }

    /* Fill in Reverse holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {
        /* Find the last non-HOLE */
        start = i;
        while(Tr[start] == HOLE)
        {
            --start;
            if(start < 0)
                start = motorInfo->encCtsPerRev;
        }
        
        /* Find the next non-HOLE */
        end = i;
        while(Tr[end] == HOLE)
        {
            ++end;
            if(end > motorInfo->encCtsPerRev-1)
                end = 0;
        }
        
        /* Find the length of the hole */
        if(end > start)
            length = end - start;
        else
            length = end + motorInfo->encCtsPerRev - start;
        
        /* Find the increment */
        inc = (float)(Tr[end] - Tr[start]) / (float)length;
        sum = 0;
        
        /* Fill in data */
        for(end = 1; end < length; end++)
        {
            ++start;
            sum += inc;
            if(start == motorInfo->encCtsPerRev)
                start = 0;
            Tr[start] = (int)sum;
        }
    }

#endif
           
#if 0     
    /* Fill in holes */
    for(i = 0; i < motorInfo->encCtsPerRev /* * REVOLUTIONS */; i++)
    {

        /* Fill in forward data */
        count = 0;
        while(Tf[i+count] == HOLE)
        {
            ++count;
            if(i + count >= motorInfo->encCtsPerRev /* * REVOLUTIONS */)
                count -= motorInfo->encCtsPerRev /* * REVOLUTIONS */;
        }
        Tf[i] = Tf[i+count];
                
        /* Fill in reverse data */
        count = 0;
        while(Tr[i+count] == HOLE)
        {
            ++count;
            if(i + count >= motorInfo->encCtsPerRev /* * REVOLUTIONS */)
                count -= motorInfo->encCtsPerRev /* * REVOLUTIONS */;
        }
        Tr[i] = Tr[i+count];
        
    }

   
    /* Average each REVOLUTIONS data points */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        for(rev = 1; rev < REVOLUTIONS; rev++)
        {
            Tf[i] = Tf[i] + Tf[(rev * motorInfo->encCtsPerRev) + i];
            Tr[i] = Tr[i] + Tr[(rev * motorInfo->encCtsPerRev) + i];
        }
        Tf[i] = Tf[i] / REVOLUTIONS;
        Tr[i] = Tr[i] / REVOLUTIONS;
    }

    /* Average forward and reverse torques */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        Tc[i] = (Tf[i] + Tr[i]) / 2;
    }
    
    /* Now run the filter BACKWARDS to eliminate the spatial phase introduced *
     * above.  Likewise, average every SMOOTH_CT data points. */                       
    for(i = motorInfo->encCtsPerRev-1; i >= 0; i--)
    {
        torqueSum = 0;
        for(count = 0; count < SMOOTH_CT; count++)
        {
            modPos = i + count - (SMOOTH_CT-1) / 2;
            modPos %= motorInfo->encCtsPerRev;
            if(modPos < 0) modPos += motorInfo->encCtsPerRev;
            torqueSum += Tc[modPos];
        }
        Tc[i] = torqueSum / SMOOTH_CT;
    }

    /* Apply a gain */
    for(i = 0; i < motorInfo->encCtsPerRev; i++)
    {
        Tc[i] = (int)((float)Tc[i] * TR_GAIN);
    }
#endif

#if(0) /* Sample tr.info file */

PuckID : 5
MotorSN : 10
CtsPerRev : 6600
mechOffset : 233
iOffsetB : 575
iOffsetC : 575
Rev_F : 5
Rev_R : 5
StepSize : 1
SmoothCt : 12
SampleTime : 0.010
WaitTime : 0.0025
Kp : 2
Kd : 10
Ki : 0

#endif