/* main.c */

#define toupper(c)      ( ((c >= 'a') && (c <= 'z')) ? c - ('a' - 'A') : c )
#define HOLE            (0x0000)


#define TR_GAIN         (0.8)
#define GAIN_WINDOW     (1)
#define pi              (3.14159)

#define TEST_TR         (0)
#define COLLECT_TR      (1)
#define BUS             (0)

#include <stdio.h>
#include <syslog.h>
#include <sys/neutrino.h>
#include <sys/syspage.h>
#include <inttypes.h>
#include <malloc.h>
#include <math.h>
#include <semaphore.h>

#include "btcan.h"
#include "SimpleControl.h"

int     REVOLUTIONS_F, REVOLUTIONS_R, SMOOTH_CT;
int     STEP;
double  TR_SAMPLE_TIME;
double  WAIT_TIME;
double  kp, kd, ki;
int     compTorque;
extern sem_t   timer_semaphore;
extern int     shutdown_threads;

typedef struct
{
    int puckID;
    int motorSN;
    int encCtsPerRev;
    int mechOffset;
    int iOffsetB;
    int iOffsetC;
}motorInfo_t;

typedef struct{char s[20];}string20;

typedef struct
{
    motorInfo_t *m;
    //CANdev_t    *c;
    PIDregulator *p;
    SimpleCtl *s;
}CThreadParams_t;

const    string20	statusTxt[]={
        {"STATUS_OFFLINE"},
    	{"STATUS_RESET"},
        {"STATUS_SHUTDOWN"},
        {"STATUS_INITIALIZE"},
        {"STATUS_ERR"},
        {"STATUS_READY"},
        {"STATUS_COMMUTATING"},
    };

void PuckControlThread(void *data);

void showMenu(void){
    printf("\nMENU");
    printf("\n--------");
    printf("\nE)numerate bus status");
    printf("\nM)easure CAN speed");
    printf("\nD)ownload firmware");
    printf("\nO)ld firmware download");
    printf("\nC)ollect TR data");
    printf("\nT)est TR data");
    
    printf("\n\nYour Choice: ");
}

void handleMenu(char c){
    char        fname[32];
    int         node;
    int         err;
    uint64_t    time1, time2;
    float       delta;
    long        reply;
    long         status[MAX_NODES];
    int         i;
    double      cps;
    int         newID;
    
    
    switch(c){
    case 'E':
        printf("\n\nCAN bus enumeration (status)\n");
        getBusStatus(BUS, status);
        for(i = 1; i < MAX_NODES; i++)
        {
            if(i % 2)
                printf("\t\tNode %2d: %s", i, statusTxt[status[i]+1].s);
            else
                printf("\nNode %2d: %s", i, statusTxt[status[i]+1].s);
        }
        printf("\n");
        break;
        
    case 'M':
        printf("\n\nCAN Speed Test...\n");
        cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
        while(1)
        {
            time1 = ClockCycles();
            getProperty(BUS, 5, STAT, &reply);
            time2 = ClockCycles();
            delta = (double)(time2 - time1) / cps;
            printf("\r%f", delta); fflush(stdout);
        }
        
    case 'D':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Get new Node ID */
        if(node > 0)
        {
            printf("\nChange node ID to (enter 0 for no change): ");
            err = scanf("%d", &newID);
            if(err == EOF) /* No input data */
                newID = 0; /* Zero signifies "no change" */
        }
        
        /* Download the firmware */
        //err = downloadFirmware(node, newID, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'O':
        /* Get filename */
        printf("\nFirmware filename: ");
        scanf("%s", fname);
        
        /* Get puck string */
        printf("\nWhich puck(s): ");
        scanf("%d", &node);
        
        /* Download the firmware */
        //err = downloadFirmware_Old( node, fname);
        
        /* Check for errors */
        if(err){
            printf("\n*** Download failed! ***\n");
        }else{
            printf("\n*** Download complete! ***\n");    
        }
        
        /* Wait for user acknowledgement */
        getchar();
        
        break;
        
    case 'C':
        /* Collect TR Data */
        doTR(COLLECT_TR);
        break;
        
    case 'T':
        doTR(TEST_TR);
        break;
        
    default:
    
        break;
    }
}

int doTR(int doCollect)
{
    motorInfo_t motorInfo;
    int         err;
    FILE        *in;
    CThreadParams_t CThreadParams;
    SimpleCtl   sc;
    long        actPos;
    long        reply;
    
    if((in = fopen("tr.info","r")) == NULL)
    {
        printf("\nFailed to open tr.info file!");
        return(1);
    }
    fscanf(in, "%*s : %d", &motorInfo.puckID);
    fscanf(in, "%*s : %d", &motorInfo.motorSN);
    fscanf(in, "%*s : %d", &motorInfo.encCtsPerRev);
    fscanf(in, "%*s : %d %*s : %d %*s : %d", &motorInfo.mechOffset, &motorInfo.iOffsetB, &motorInfo.iOffsetC);
    fscanf(in, "%*s : %d %*s : %d %*s : %d %*s : %d %*s : %lf %*s : %lf", &REVOLUTIONS_F, &REVOLUTIONS_R, &STEP, &SMOOTH_CT, &TR_SAMPLE_TIME, &WAIT_TIME);
    fscanf(in, "%*s : %lf %*s : %lf %*s : %lf", &kp, &kd, &ki);
    
    fclose(in);
    
    printf("CANID      : %d\n"
            "MOTOR_SN   : %d\n"
            "CTS_PER_REV: %d\n"
            "MECH_OFFSET: %d \n"
            "IOFFSET_B  : %d \n"
            "IOFFSET_C  : %d\n"
            "FWD_REVS   : %d \n"
            "REV_REVS   : %d \n"
            "STEP_SIZE  : %d \n"
            "SMOOTH_CT  : %d \n"
            "SAMP_TIME  : %0.3f \n"
            "WAIT_TIME  : %0.3f\n"
            "KP         : %0.3f \n"
            "KD         : %0.3f \n"
            "KI         : %0.3f\n", motorInfo.puckID, motorInfo.motorSN, motorInfo.encCtsPerRev,
            motorInfo.mechOffset, motorInfo.iOffsetB, motorInfo.iOffsetC,
            REVOLUTIONS_F, REVOLUTIONS_R, STEP, SMOOTH_CT, TR_SAMPLE_TIME, WAIT_TIME,
            kp, kd, ki);
   
    SCinit(&sc);
    SCsetpid(&sc, kp, kd, ki, 8191);
    
    CThreadParams.m = &motorInfo;
    CThreadParams.s = &sc;
    
    syslog(LOG_ERR, "Check the status of puck %d", motorInfo.puckID);

    /* Check the status of the puck */
    getProperty(BUS, motorInfo.puckID, STAT, &reply);
    syslog(LOG_ERR, "Puck status is %s", statusTxt[reply+1].s);
    switch(reply)
    {
        case STATUS_RESET:
            /* Get the puck out of RESET */
            setProperty(BUS, motorInfo.puckID, STAT, FALSE, STATUS_READY);
        
            /* Wait a moment */
            usleep(1000);
            
            /* Check the status of the puck */
            getProperty(BUS, motorInfo.puckID, STAT, &reply);
            
            if(reply != STATUS_READY)
            {
                printf("\nSomething is wrong with puck %d!  Try cycling the power.\n", motorInfo.puckID);
                return(1);
            }
            break;
        case STATUS_READY:
            /* Ready to go */
            
            break;
        default:
            printf("\nSomething is wrong with puck %d!  Try cycling the power.\n", motorInfo.puckID);
            return(1);
    }

    compTorque = 0;

    syslog(LOG_ERR, "Starting control threads");
    start_control_threads( 40, 1, PuckControlThread, (void*)&CThreadParams);
    
    SCsetmode(&sc, MODE_IDLE);
    
    /* Start at index pulse */
    initTR(&CThreadParams);
    
    
    usleep(1000000);
    
    /* Collect the data */
    if(doCollect)
        err = collectTR(&CThreadParams);
    else
        err = testTR(&CThreadParams);
    
    return(err);
}

int collectTR(CThreadParams_t *p)
{
    int         *Tc, *Tf, *Tr;
    long        modPos;
    uint64_t    startTime;
    float       delta;
    int         torque;
    long        i;
    int         rev, rev2;
    float       torqueSum;
    long        samples;
    int         count;
    char        fileName[32];
    FILE        *outFile, *trFile;
    double      cps;
    int         stepsPerRev;
    int         idx;
    int         id;
    
    //motorInfo_t *m;
    //CANdev_t    *CANdev;
    //PIDregulator *pid;
    
    //CANdev = CThreadParams->c;
    //pid = CThreadParams->p;
    id = p->m->puckID;
    
    stepsPerRev = p->m->encCtsPerRev / STEP;
    if(stepsPerRev * STEP != p->m->encCtsPerRev)
    {
        printf("\nCtsPerRev not evenly divisible by STEP!");
        return(1);
    }

    syslog(LOG_ERR, "Allocating TR arrays");
    
    /* Allocate TorqueRipple Forward Compensation array space */
    Tf = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tf == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Reverse Compensation array space */
    Tr = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tr == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
       
    /* Allocate TorqueRipple Compensation array space */
    Tc = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tc == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Fill the initial arrays with holes */
    for(i = 0; i < stepsPerRev; i++)
    {
        Tf[i] = Tr[i] = HOLE;
    }
    

    /* Write the data to a file */
    sprintf(fileName, "motor%d.raw", p->m->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "Dir\tRev#\tSamples\tCmdPos\tActPos\tTCmdTrq\n");

    sprintf(fileName, "motor%d.tr", p->m->motorSN);
    if((trFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }

    
    cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
    
    syslog(LOG_ERR, "Collecting data");
    
    /* Collect the raw forward data */
    for(rev = 0; rev < REVOLUTIONS_F; rev++)
    {
        for(idx = 0; idx < stepsPerRev; idx++)
        //for(pid->yref = rev * motorInfo->encCtsPerRev; 
        //    pid->yref < (rev + 1) * motorInfo->encCtsPerRev; 
        //    pid->yref += STEP)
        {
            SCsetcmd(p->s,rev * p->m->encCtsPerRev + idx * STEP);
            samples = 0;
            torqueSum = 0;
            
            /* Calc autoranged position */
            modPos = idx; //(long)pid->yref % motorInfo->encCtsPerRev;
            compTorque = Tf[modPos];
                
            startTime = ClockCycles();
            do
            {
                /* Find the new torque value */
                torque = (int)p->s->pid.lastresult + compTorque;
    
                //printf("\ne=%f,de=%f",pid->e,pid->de);
                /* Calculate elapsed time at this position */
                delta = (float)(ClockCycles() - startTime) / cps;
                if(delta > WAIT_TIME)
                {
                    if(fabs(p->s->pid.de) == 0)
                    {
                        /* Record the torque sum */
                        torqueSum += torque;
                        ++samples;
                    }
                    //else printf("\nWaiting for de to settle");
                }
                //syslog(LOG_ERR, "Delta: %lf, SampleT: %lf", delta, TR_SAMPLE_TIME);
                usleep(50);
            }while(samples < 16);//delta < TR_SAMPLE_TIME); /* Run for TR_SAMPLE_TIME seconds */
            
            /* Save the average torque */
            Tf[modPos] = (int)(torqueSum / samples);
            fprintf(outFile, "F\t%d\t%d\t%ld\t%ld\t%d\n", rev, samples, modPos, (long)p->s->pid.y % p->m->encCtsPerRev, Tf[modPos]);
            printf("F\t%d\t%d\t%ld\t%ld\t%d\n", rev, samples, modPos, (long)p->s->pid.y % p->m->encCtsPerRev, Tf[modPos]);
        }
        
        /* Average every SMOOTH_CT data points of Tf into Tc */
        for(i = 0; i < stepsPerRev; i++)
        {
            torqueSum = 0;
            for(count = 0; count < SMOOTH_CT; count++)
            {
                modPos = i + count - (SMOOTH_CT-1) / 2;
                modPos %= stepsPerRev;
                if(modPos < 0) modPos += stepsPerRev;
                torqueSum += Tf[modPos];
                
            }
            Tc[i] = torqueSum / SMOOTH_CT;
        }
        
        /* Save the smoothed values back into Tf */
        for(i = 0; i < stepsPerRev; i++)
        {
            Tf[i] = Tc[i];
        }
        
        /* Save this cycle */
        rewind(trFile);
        fprintf(trFile, "%d\n", stepsPerRev);
        for(i = 0; i < stepsPerRev; i++)
        {
            fprintf(trFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
        }
        
    }
    
    rev2 = 0;
    //setProperty(BUS, id, ACTUAL_POSITION, (REVOLUTIONS_R * motorInfo->encCtsPerRev)-1 );
    for(rev = REVOLUTIONS_R - 1; rev >= 0; rev--)
    {  
        
        /* Collect the raw reverse data */
        for(idx = stepsPerRev-1; idx >= 0; idx--)
        //for(pid->yref = (rev * motorInfo->encCtsPerRev)-1; 
        //    pid->yref >= (rev - 1) * motorInfo->encCtsPerRev; 
        //    pid->yref -= STEP)
        {
            SCsetcmd(p->s, rev * p->m->encCtsPerRev + idx * STEP);
            samples = 0;
            torqueSum = 0;
            
            /* Calc autoranged position */
            modPos = idx; //(long)pid->yref % motorInfo->encCtsPerRev;
            compTorque = Tr[modPos];
            
            startTime = ClockCycles();
            do
            {
                /* Find the new torque value */
                torque = (int)p->s->pid.lastresult + compTorque;
  
                /* Calculate elapsed time at this position */
                delta = (float)(ClockCycles() - startTime) / cps;
                if(delta > WAIT_TIME)
                {
                    if(fabs(p->s->pid.de) == 0)
                    {
                        /* Record the torque sum */
                        torqueSum += torque;
                        ++samples;
                    }
                    //else printf("\nWaiting for de to settle");
                }
                //syslog(LOG_ERR, "Delta: %lf, SampleT: %lf", delta, TR_SAMPLE_TIME);
                usleep(50);
            }while(samples < 16);//delta < TR_SAMPLE_TIME); /* Run for TR_SAMPLE_TIME seconds */
             
            /* Save the average torque */
            Tr[modPos] = (int)(torqueSum / samples);
            fprintf(outFile, "R\t%d\t%d\t%ld\t%ld\t%d\n", rev2, samples, modPos, (long)p->s->pid.y % p->m->encCtsPerRev, Tr[modPos]);
            printf("R\t%d\t%d\t%ld\t%ld\t%d\n", rev2, samples, modPos, (long)p->s->pid.y % p->m->encCtsPerRev, Tr[modPos]);
                    
        }
        
       /* Average every SMOOTH_CT data points of Tf into Tc */
        for(i = 0; i < stepsPerRev; i++)
        {
            torqueSum = 0;
            for(count = 0; count < SMOOTH_CT; count++)
            {
                modPos = i + count - (SMOOTH_CT-1) / 2;
                modPos %= stepsPerRev;
                if(modPos < 0) modPos += stepsPerRev;
                torqueSum += Tr[modPos];
                
            }
            Tc[i] = torqueSum / SMOOTH_CT;
        }
        
        /* Save the smoothed values back into Tr */
        for(i = 0; i < stepsPerRev; i++)
        {
            Tr[i] = Tc[i];
        }
        
        /* Save this cycle */
        rewind(trFile);
        fprintf(trFile, "%d\n", stepsPerRev);
        for(i = 0; i < stepsPerRev; i++)
        {
            fprintf(trFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
        }
        
        rev2++;
    }
    
    /* Close the RAW data file */
    fclose(outFile);

    /* Close *.tr data file */
    fclose(trFile);
    
    /* Smooth it out */
    /* Average every SMOOTH_CT data points of Tf and Tr, then average Tf and Tr into Tc */
    for(i = 0; i < stepsPerRev; i++)
    {
        torqueSum = 0;
        for(count = 0; count < SMOOTH_CT; count++)
        {
            modPos = i + count - (SMOOTH_CT-1) / 2;
            modPos %= stepsPerRev;
            if(modPos < 0) modPos += stepsPerRev;
            torqueSum += Tf[modPos];
            torqueSum += Tr[modPos];
        }
        Tc[i] = torqueSum / (SMOOTH_CT * 2);
    }
    
    /* Write the data to a file */
    sprintf(fileName, "motor%d.tr", p->m->motorSN);
    if((outFile = fopen(fileName, "w")) == NULL)
    {
        printf("\nUnable to open '%s' for writing!", fileName);
        return(1);
    }
    fprintf(outFile, "%d\n", stepsPerRev);
    
    /* Output the data */
    for(i = 0; i < stepsPerRev; i++)
    {
        /* Save the forward data, reverse data, and smoothed data */
        fprintf(outFile, "%d\t%d\t%d\n", Tf[i], Tr[i], Tc[i]);
    }
    
    fclose(outFile);
    
    printf("\nDone collecting TR data!\n");
    
    /* Clean up */
    free(Tf);
    free(Tr);
    free(Tc);
    return(0);
}

int testTR(CThreadParams_t *p)
{
    int         i;
    char        fileName[32];
    FILE        *in;
    int         *Tf, *Tr, *Tc;
    int         modPos;
    int         stepsPerRev;
    
    printf("\nPlease enter *.tr file name: ");
    scanf("%s", fileName);
    
    if((in = fopen(fileName, "r")) == NULL)
    {
        printf("\nFailed to open torque ripple file!");
        return(1);
    }

    fscanf(in, "%d", &stepsPerRev); 
    

    /* Allocate TorqueRipple Compensation array space */
    Tf = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tf == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Compensation array space */
    Tr = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tr == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    /* Allocate TorqueRipple Compensation array space */
    Tc = (int *)malloc(stepsPerRev * sizeof(int));
    if(Tc == NULL) /* Failed to allocate buffer */
    {
        printf("\nFailed to allocate TR buffer!");
        return(1);
    }
    
    for(i = 0; i < stepsPerRev; i++)
    {
        fscanf(in, "%d %d %d", &Tf[i], &Tr[i], &Tc[i]);
    }
    fclose(in);

    /* TR Test */
    p->s->pid.Kp = 0;
    p->s->pid.Ki = 0;
    p->s->pid.Kd = 0;
    
    printf("\nTR Test, press ctrl-c to exit..."); fflush(stdout);
    do
    {
        modPos = (long)p->s->pid.y % p->m->encCtsPerRev;
        if(modPos < 0) modPos += p->m->encCtsPerRev;
        modPos /= STEP;
        compTorque = Tc[modPos] + TR_SAMPLE_TIME;
        //printf("\ncompTorque = %d\t", compTorque);
        usleep(1000);
        
    }while(1);
    
}

int initTR(CThreadParams_t *p)
{
    long    reply;
    int     id;
    double  yref;
    
    id = p->m->puckID;
   
    syslog(LOG_ERR, "Looking for index pulse");

    /* Spin the motor until MECH >= 0 */
    setProperty(BUS, id, MODE, FALSE, MODE_TORQUE);
    SCsetmode(p->s, MODE_PID);
    
    
    getProperty(BUS, id, MECH, &reply);
    while(reply < 0)
    {
        SCsetcmd(p->s, p->s->pid.yref + 25);
        usleep(1000);
        getProperty(BUS, id, MECH, &reply);
    }
    
    /* Short delay */
    usleep(50000); 
    
    /* Get the motor near its zero (index pulse) position */
    SCsetmode(p->s, MODE_IDLE);
    setProperty(BUS, id, MODE, FALSE, MODE_IDLE);
    
    getProperty(BUS, id, MECH, &reply);
    printf("\nMECH = %ld", reply);
    
    setProperty(BUS, id, AP, FALSE, reply);
    getProperty(BUS, id, AP, &reply);
    printf("\nAP = %ld", reply);
    
    setProperty(BUS, id, MODE, FALSE, MODE_TORQUE);
    SCsetmode(p->s, MODE_PID);
    printf("\nYREF = %f", p->s->pid.yref);
    

    /* Go backwards */
    for(yref = reply; yref > 0; yref-=10)
    {
        SCsetcmd(p->s, yref);
        usleep(1000);        
    }
    SCsetcmd(p->s, 0);
    
    /* Short delay */
    usleep(50000); 
    
    //while(1) usleep(10);
    
}

int main( int argc, char **argv )
{
    char            c;
    int             err;
     
    /* Initialize syslogd */
    openlog("LOG_ERR",LOG_CONS | LOG_NDELAY, LOG_USER);
    syslog(LOG_ERR, "syslog initalized");
    
    /* Initialize CAN */
    if(err = initCAN(BUS))
    {
        syslog(LOG_ERR, "initCAN returned err=%d", err);
    }
    
    /* Show Menu */
    showMenu();
    
    /* Get Choice */
    c = getchar();
    //printf("\nYou pressed ASCII: %d", c);
    c = toupper(c);
    
    /* Handle Menu */
    handleMenu(c);
    
    return(0);
}

void PuckControlThread(void *data)
{
    long actPos;
    double torq;
    CThreadParams_t *CThreadParams;
    
    CThreadParams = (CThreadParams_t*)data;
    
    syslog(LOG_ERR, "Started PuckControlThread");
    while (!shutdown_threads)
    {
        sem_wait(&timer_semaphore);
        //syslog(LOG_ERR, "Timer posted");
        getProperty(BUS, CThreadParams->m->puckID, AP, &actPos);
        //syslog(LOG_ERR, "Ctrl Thread Position received");
        /* Update PID */
        //CThreadParams->p->y = actPos; 
        //PIDcalc(CThreadParams->p);

        /* Set the new torque */
        torq = SCevaluate(CThreadParams->s, (double)actPos, 0.001);
        setProperty(BUS, CThreadParams->m->puckID, TORQ, FALSE, (long)torq + compTorque);

    }

    pthread_exit(NULL);
}
   

#if(0) /* Sample tr.info file */

PuckID : 5
MotorSN : 10
CtsPerRev : 6600
mechOffset : 233
iOffsetB : 575
iOffsetC : 575
Rev_F : 5
Rev_R : 5
StepSize : 1
SmoothCt : 12
SampleTime : 0.010
WaitTime : 0.0025
Kp : 2
Kd : 10
Ki : 0

#endif