#include <unistd.h>
#include <malloc.h>
#include <termios.h>
#include <semaphore.h>
#include <time.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <sys/neutrino.h>
#include <inttypes.h>
#include <sys/syspage.h>

#include <ncurses.h>

#include "windows.h"
//#include "BHand.h"

#define BUFLEN 255
#define NUMBEROFPORTS 2

typedef struct { 
    FILE   *isp, *osp;                   /*stream pointers to the serial port*/ 
    int     ifd, ofd;                   /*file descriptors for the stream pointers*/ 
    //char    obuf[BUFLEN], ibuf[BUFLEN]; /*buffers for handling the streams*/ 
} PORT; 

typedef struct {
    sem_t   sem;
    int     event;
} SEM_T;

DWORD GetTickCount(void)
{
    return((DWORD)(ClockCycles() * 1000.0 / SYSPAGE_ENTRY(qtime)->cycles_per_sec));   
}

BOOL InitPort(PORT *port,char *portlocation)
{ 
    //addstr("InitPort..."); refresh();
    if ( ( port->isp=fopen(portlocation,"r") ) != NULL ) 
    { 
        port->ifd=fileno(port->isp); 
    }else return(1); //printf("\nUnable to open port for input!\n"); 
    
    if ( ( port->osp=fopen(portlocation,"w") ) != NULL ) 
    {
        port->ofd=fileno(port->osp); 
    }else return(2); //printf("\nUnable to open port for output!\n");
        
    return 0; 
}

DWORD WaitForSingleObject(
  HANDLE hHandle,
  DWORD dwMilliseconds
)
{
    struct timespec tm;
    //int val;
    //addstr("WaitForSingleObject..."); refresh();
    
    if(dwMilliseconds == 0) // Just polling sem
    {
        //sem_getvalue((sem_t*)hHandle, &val);
        //addstr("poll"); refresh();
        if(((SEM_T*)hHandle)->event>0)
        {
            return(WAIT_OBJECT_0);
        }
        else
        {
            return(!WAIT_OBJECT_0);
        }
    }
    
    clock_gettime(CLOCK_REALTIME, &tm);
    if(tm.tv_nsec + dwMilliseconds * 1000000 > 1000000000)
    {
        tm.tv_sec += 1;
        tm.tv_nsec += dwMilliseconds * 1000000 - 1000000000;
    }
    else
        tm.tv_nsec += dwMilliseconds * 1000000;
        
    //Sleep(10000);
    if(dwMilliseconds == INFINITE)
    {
        while(sem_wait( &((SEM_T*)hHandle)->sem ) == EINTR){};
        return(WAIT_OBJECT_0);
    }
    else 
    {
        if(sem_timedwait( &((SEM_T*)hHandle)->sem, &tm ) == -1 ) // Timeout?
            return(WAIT_TIMEOUT);
        else // Success
            return(WAIT_OBJECT_0);
    }
    //return(WAIT_TIMEOUT);
}

// Close the com port, free data structure memory
BOOL CloseHandle(
  HANDLE hObject
)
{
    //addstr("CloseHandle..."); refresh();
    fclose(((PORT*)hObject)->isp);
    fclose(((PORT*)hObject)->osp);
    free(hObject);  
    
    return(1); 
}

// Read from the com port
BOOL ReadFile(
  HANDLE hFile,
  LPVOID lpBuffer,
  DWORD nNumberOfBytesToRead,
  LPDWORD lpNumberOfBytesRead,
  LPOVERLAPPED lpOverlapped
)
{
    char *buf = (char*)lpBuffer;
    clock_t ticks;
    int result=0, n; 
    char a[30];
    
    //addstr("ReadFile..."); refresh();
    ticks = clock();
    while(1)
    {
        n = tcischars(((PORT*)hFile)->ifd);
        if(n >= nNumberOfBytesToRead)
            break;
            
        if( (double)(clock() - ticks) / CLOCKS_PER_SEC > 15.0 )
            break;
            
        usleep(10);
    }
        
    if(nNumberOfBytesToRead < n)
        n = nNumberOfBytesToRead;
        
    *lpNumberOfBytesRead = fread(lpBuffer,1,n,((PORT*)hFile)->isp); /*read them and */ 
    //sprintf(a,".%x:%d:%c.",lpBuffer,n,*(char*)lpBuffer);
    //addstr(a); refresh();
    return 1;
}

// Write to the com port
BOOL WriteFile(
  HANDLE hFile,
  LPCVOID lpBuffer,
  DWORD nNumberOfBytesToWrite,
  LPDWORD lpNumberOfBytesWritten,
  LPOVERLAPPED lpOverlapped
)
{
    //addstr("WriteFile..."); refresh();
    int i;
    for(i=0; i<nNumberOfBytesToWrite; i++)
        fprintf(((PORT*)hFile)->osp,"%c",((char*)lpBuffer)[i]); 
        
    fflush(((PORT*)hFile)->osp);                /*needed to force transmission*/
    *lpNumberOfBytesWritten = nNumberOfBytesToWrite;
    
    return 1;
}

// Create a semaphore
HANDLE CreateEvent(
  LPSECURITY_ATTRIBUTES lpEventAttributes,
  BOOL bManualReset,
  BOOL bInitialState,
  LPCTSTR lpName
)
{
    SEM_T *s;
    //addstr("CreateEvent..."); refresh();
    s = (SEM_T *)malloc(1 * sizeof(SEM_T));
    if(s == NULL) 
    {
        //addstr("CreateEvent FAILED"); 
        refresh();
    }
    else
    {
        //addstr("CreateEvent SUCCESS"); 
        refresh();
    }
    sem_init( &(s->sem), 0, bInitialState );
    s->event = bInitialState;
    return((HANDLE)s);
}

// Post a semaphore
BOOL SetEvent(
  HANDLE hEvent
)
{
    //addstr("SetEvent..."); refresh();
    ++((SEM_T*)hEvent)->event;
    sem_post(&((SEM_T*)hEvent)->sem);
    return(1);
}

// Do nothing, semaphores are automatically reset
BOOL ResetEvent(
  HANDLE hEvent
)
{
    //addstr("ResetEvent..."); refresh();
    --((SEM_T*)hEvent)->event;
    //sem_trywait((sem_t*)hEvent );
    return(1);
}

// Post a semaphore (it is automatically reset)
BOOL PulseEvent(
  HANDLE hEvent
)
{
    //addstr("PulseEvent..."); refresh();
    SetEvent(hEvent);
    ResetEvent(hEvent);
    return(1);
}

// Sleep
VOID Sleep(
  DWORD dwMilliseconds
)
{
    //addstr("Sleep..."); refresh();
    usleep(dwMilliseconds * 1000);   
}

// Clear the isp and osp of the port
BOOL PurgeComm(
  HANDLE hFile,
  DWORD dwFlags
)
{
    char c;
    int n;
    
    //addstr("PurgeComm..."); refresh();
    while(1)
    {
        n = tcischars(((PORT*)hFile)->ifd);
        if(n > 0)
            fread(&c,1,1,((PORT*)hFile)->isp); /*read them and */
        else break;
    }
        
    return(1);
}

BOOL SetCommTimeouts(
  HANDLE hFile,
  LPCOMMTIMEOUTS lpCommTimeouts
)
{
    return(1);
}

// Get the serial port parameter structure
BOOL GetCommState(
  HANDLE hFile,
  LPDCB lpDCB
)
{
    return(1);
}

// Set the new baud rate
BOOL SetCommState(
  HANDLE hFile,
  LPDCB lpDCB
)
{
    struct termios tc;
    //addstr("SetCommState..."); refresh();
    // Set the input baud rate
    tcgetattr( ((PORT*)hFile)->ifd, &tc );
    cfsetispeed( &tc, (speed_t)(lpDCB->BaudRate) );
    tcsetattr( ((PORT*)hFile)->ifd, TCSADRAIN, &tc);
    
    // Set the output baud rate
    tcgetattr( ((PORT*)hFile)->ofd, &tc );
    cfsetospeed( &tc, (speed_t)(lpDCB->BaudRate) );
    tcsetattr( ((PORT*)hFile)->ofd, TCSADRAIN, &tc);  
    
    return(1);  
}

// Open the serial port
HANDLE CreateFile(
  LPCTSTR lpFileName,
  DWORD dwDesiredAccess,
  DWORD dwShareMode,
  LPSECURITY_ATTRIBUTES lpSecurityAttributes,
  DWORD dwCreationDisposition,
  DWORD dwFlagsAndAttributes,
  HANDLE hTemplateFile
)
{
    PORT        *serport;
    char portName[] = "/dev/ser0";
    int err;
    char a[10];
    
    portName[8] = lpFileName[3];
    //addstr("CreateFile..."); //addstr(portName); refresh();
    /* Initialize the serial port */
    serport = (PORT *)calloc(1, sizeof(PORT));
    err = InitPort(serport, portName);
    //addstr("InitPort="); sprintf(a,"%d",err); //addstr(a); refresh();
    
    return((HANDLE)serport);
}



HANDLE CreateThread(
  LPSECURITY_ATTRIBUTES lpThreadAttributes,
  SIZE_T dwStackSize,
  LPTHREAD_START_ROUTINE lpStartAddress,
  LPVOID lpParameter,
  DWORD dwCreationFlags,
  LPDWORD lpThreadId
)
{
    //addstr("CreateThread..."); refresh();
    pthread_t *thread;
    pthread_attr_t attr;

    struct sched_param  sched;
    int x; char a[10];

    thread = (pthread_t*)malloc(1*sizeof(pthread_t));
    //if(thread == NULL){//addstr("malloc failed..."); refresh();}
    
    pthread_attr_init(&attr);
    pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
    sched.sched_priority = 10;
    pthread_attr_setschedparam(&attr, &sched);
    sched_setscheduler(0, SCHED_FIFO, &sched);
    
    x = sched_getscheduler( 0 );
    //addstr("Scheduler=");
    sprintf(a,"%d",x); //addstr(a); refresh();
    
    pthread_attr_destroy( &attr );
    pthread_attr_init( &attr );
    pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED );
    pthread_attr_setschedpolicy(&attr, SCHED_FIFO );
    
    //pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED );
    pthread_create( thread, &attr, (void *(*)(void *))lpStartAddress, lpParameter );
    //(void *(*)(void *))&lpStartAddress
    
    return((HANDLE)thread); 
}

BOOL SetThreadPriority(
  HANDLE hThread,
  int nPriority
)
{
    struct sched_param thread_param;
    int policy;
    
    //addstr("SetThreadPriority..."); refresh();
    pthread_getschedparam(*(pthread_t*)hThread, &policy, &thread_param );
    thread_param.sched_priority = nPriority;
    pthread_setschedparam(*(pthread_t*)hThread, policy, &thread_param );
    
    return(1);
}