#ifndef _WINDOWS_H
#define _WINDOWS_H

#include <string.h>

#define VOID            void
#define CONST           const
#define _stdcall
#define WINAPI 

typedef int             BOOL;
typedef unsigned char   BYTE;
typedef unsigned short  WORD;

typedef void            *PVOID;
typedef PVOID           HANDLE;
typedef unsigned long   DWORD;
typedef DWORD           *LPDWORD;
typedef VOID            *LPVOID;

typedef const char      *LPCTSTR;
typedef unsigned long   *SIZE_T;
typedef LPVOID          LPOVERLAPPED;
typedef DWORD           (WINAPI *LPTHREAD_START_ROUTINE)(LPVOID);
typedef CONST void      *LPCVOID;

#define PURGE_TXABORT       0x0001  // Kill the pending/current writes to the comm port.
#define PURGE_RXABORT       0x0002  // Kill the pending/current reads to the comm port.
#define PURGE_TXCLEAR       0x0004  // Kill the transmit queue if there.
#define PURGE_RXCLEAR       0x0008  // Kill the typeahead buffer if there.

#define NOPARITY            0
#define ODDPARITY           1
#define EVENPARITY          2
#define MARKPARITY          3
#define SPACEPARITY         4

#define ONESTOPBIT          0
#define ONE5STOPBITS        1
#define TWOSTOPBITS         2

#define GENERIC_READ      (0x80000000L)
#define GENERIC_WRITE     (0x40000000L)
#define GENERIC_EXECUTE   (0x20000000L)
#define GENERIC_ALL       (0x10000000L)

#define CREATE_NEW          1
#define CREATE_ALWAYS       2
#define OPEN_EXISTING       3
#define OPEN_ALWAYS         4
#define TRUNCATE_EXISTING   5

#define INVALID_HANDLE_VALUE (HANDLE)-1

#define INFINITE        0xFFFFFFFF
#define WAIT_FAILED     ((DWORD)-1)
#define WAIT_OBJECT_0   ((DWORD)0)
#define WAIT_ABANDONED  ((DWORD)0x00000080L)
#define WAIT_TIMEOUT    ((DWORD)0x00000102L)



#define THREAD_PRIORITY_TIME_CRITICAL 11

#define max(a,b) (((a) > (b)) ? (a) : (b))


typedef struct _COMMTIMEOUTS {
  DWORD ReadIntervalTimeout;
  DWORD ReadTotalTimeoutMultiplier;
  DWORD ReadTotalTimeoutConstant;
  DWORD WriteTotalTimeoutMultiplier;
  DWORD WriteTotalTimeoutConstant;
} COMMTIMEOUTS, *LPCOMMTIMEOUTS;

typedef struct _DCB {
  DWORD DCBlength;
  DWORD BaudRate;
  DWORD fBinary  :1;
  DWORD fParity  :1;
  DWORD fOutxCtsFlow  :1;
  DWORD fOutxDsrFlow  :1;
  DWORD fDtrControl  :2;
  DWORD fDsrSensitivity  :1;
  DWORD fTXContinueOnXoff  :1;
  DWORD fOutX  :1;
  DWORD fInX  :1;
  DWORD fErrorChar  :1;
  DWORD fNull  :1;
  DWORD fRtsControl  :2;
  DWORD fAbortOnError  :1;
  DWORD fDummy2  :17;
  WORD wReserved;
  WORD XonLim;
  WORD XoffLim;
  BYTE ByteSize;
  BYTE Parity;
  BYTE StopBits;
  char XonChar;
  char XoffChar;
  char ErrorChar;
  char EofChar;
  char EvtChar;
  WORD wReserved1;
} DCB, *LPDCB;

typedef struct _SECURITY_ATTRIBUTES { 
    DWORD  nLength;                 // set this to sizeof( SECURITY_ATTRIBUTES );
    LPVOID lpSecurityDescriptor;    // "complex" structure.
    BOOL   bInheritHandle;          // whether child process inherits handle.
} SECURITY_ATTRIBUTES, *LPSECURITY_ATTRIBUTES;

DWORD GetTickCount(void);

DWORD WaitForSingleObject(
  HANDLE hHandle,
  DWORD dwMilliseconds
);

BOOL CloseHandle(
  HANDLE hObject
);

BOOL SetEvent(
  HANDLE hEvent
);

BOOL PurgeComm(
  HANDLE hFile,
  DWORD dwFlags
);

BOOL ReadFile(
  HANDLE hFile,
  LPVOID lpBuffer,
  DWORD nNumberOfBytesToRead,
  LPDWORD lpNumberOfBytesRead,
  LPOVERLAPPED lpOverlapped
);

BOOL WriteFile(
  HANDLE hFile,
  LPCVOID lpBuffer,
  DWORD nNumberOfBytesToWrite,
  LPDWORD lpNumberOfBytesWritten,
  LPOVERLAPPED lpOverlapped
);

BOOL ResetEvent(
  HANDLE hEvent
);

BOOL PulseEvent(
  HANDLE hEvent
);

VOID Sleep(
  DWORD dwMilliseconds
);

BOOL SetCommTimeouts(
  HANDLE hFile,
  LPCOMMTIMEOUTS lpCommTimeouts
);

BOOL GetCommState(
  HANDLE hFile,
  LPDCB lpDCB
);

BOOL SetCommState(
  HANDLE hFile,
  LPDCB lpDCB
);

HANDLE CreateFile(
  LPCTSTR lpFileName,
  DWORD dwDesiredAccess,
  DWORD dwShareMode,
  LPSECURITY_ATTRIBUTES lpSecurityAttributes,
  DWORD dwCreationDisposition,
  DWORD dwFlagsAndAttributes,
  HANDLE hTemplateFile
);

HANDLE CreateEvent(
  LPSECURITY_ATTRIBUTES lpEventAttributes,
  BOOL bManualReset,
  BOOL bInitialState,
  LPCTSTR lpName
);

HANDLE CreateThread(
  LPSECURITY_ATTRIBUTES lpThreadAttributes,
  SIZE_T dwStackSize,
  LPTHREAD_START_ROUTINE lpStartAddress,
  LPVOID lpParameter,
  DWORD dwCreationFlags,
  LPDWORD lpThreadId
);

BOOL SetThreadPriority(
  HANDLE hThread,
  int nPriority
);



#endif /* _WINDOWS_H */